@@ -39,7 +39,7 @@
  *  @param  proto  the desired protocol
  *  @return length of the encoded APDU
  */
-static size_t sc_apdu_get_length(const sc_apdu_t *apdu, unsigned int proto)
+size_t sc_apdu_get_length(const sc_apdu_t *apdu, unsigned int proto)
 {
 	size_t ret = 4;
 
@@ -77,10 +77,10 @@ static size_t sc_apdu_get_length(const sc_apdu_t *apdu, unsigned int proto)
  *  @param  apdu    APDU to be encoded as an octet string
  *  @param  proto   protocol version to be used
  *  @param  out     output buffer of size outlen.
- *  @param  outlen  size of hte output buffer
+ *  @param  outlen  size of the output buffer
  *  @return SC_SUCCESS on success and an error code otherwise
  */
-static int sc_apdu2bytes(sc_context_t *ctx, const sc_apdu_t *apdu,
+int sc_apdu2bytes(sc_context_t *ctx, const sc_apdu_t *apdu,
 	unsigned int proto, u8 *out, size_t outlen)
 {
 	u8     *p = out;
@@ -244,7 +244,7 @@ int sc_apdu_set_resp(sc_context_t *ctx, sc_apdu_t *apdu, const u8 *buf,
  *                        +-------------------------------+
  *                                                        |
  *                                                        v
- *                                               card->reader->ops->tranmit
+ *                                               card->reader->ops->transmit
  */
 
 /** basic consistency check of the sc_apdu_t object
@@ -291,10 +291,6 @@ sc_check_apdu(sc_card_t *card, const sc_apdu_t *apdu)
 		/* data is expected       */
 		if (apdu->resplen == 0 || apdu->resp == NULL)
 			goto error;
-		/* return buffer to small */
-		if ((apdu->le == 0 && apdu->resplen < SC_MAX_APDU_BUFFER_SIZE-2)
-				|| (apdu->resplen < apdu->le))
-			goto error;
 		break;
 	case SC_APDU_CASE_3_SHORT:
 		/* data is sent */
@@ -314,10 +310,6 @@ sc_check_apdu(sc_card_t *card, const sc_apdu_t *apdu)
 		/* data is expected       */
 		if (apdu->resplen == 0 || apdu->resp == NULL)
 			goto error;
-		/* return buffer to small */
-		if ((apdu->le == 0 && apdu->resplen < SC_MAX_APDU_BUFFER_SIZE-2)
-				|| (apdu->resplen < apdu->le))
-			goto error;
 		/* inconsistent datalen   */
 		if (apdu->datalen != apdu->lc)
 			goto error;
@@ -350,7 +342,7 @@ sc_detect_apdu_cse(const sc_card_t *card, sc_apdu_t *apdu)
 	if (apdu->cse == SC_APDU_CASE_2 || apdu->cse == SC_APDU_CASE_3 ||
 	    apdu->cse == SC_APDU_CASE_4) {
 		int btype = apdu->cse & SC_APDU_SHORT_MASK;
-		/* if either Lc or Le is bigger than the maximun for
+		/* if either Lc or Le is bigger than the maximum for
 		 * short APDUs and the card supports extended APDUs
 		 * use extended APDUs (unless Lc is greater than
 		 * 255 and command chaining is activated) */
@@ -443,7 +435,7 @@ sc_get_response(struct sc_card *card, struct sc_apdu *apdu, size_t olen)
 	if (!card->ops->get_response)
 		LOG_TEST_RET(ctx, SC_ERROR_NOT_SUPPORTED, "no GET RESPONSE command");
 
-	/* call GET RESPONSE until we have read all data requested or until the card retuns 0x9000,
+	/* call GET RESPONSE until we have read all data requested or until the card returns 0x9000,
 	 * whatever happens first. */
 
 	/* if there are already data in response append a new data to the end of the buffer */
@@ -468,7 +460,7 @@ sc_get_response(struct sc_card *card, struct sc_apdu *apdu, size_t olen)
 		if (rv < 0)   {
 #ifdef ENABLE_SM
 			if (resp_len)   {
-				sc_log(ctx, "SM response data %s", sc_dump_hex(resp, resp_len));
+				sc_log_hex(ctx, "SM response data", resp, resp_len);
 				sc_sm_update_apdu_response(card, resp, resp_len, rv, apdu);
 			}
 #endif
@@ -524,7 +516,7 @@ sc_transmit(sc_card_t *card, sc_apdu_t *apdu)
 	LOG_TEST_RET(ctx, r, "transmit APDU failed");
 
 	/* ok, the APDU was successfully transmitted. Now we have two special cases:
-	 * 1. the card returned 0x6Cxx: in this case APDU will be re-trasmitted with Le set to SW2
+	 * 1. the card returned 0x6Cxx: in this case APDU will be re-transmitted with Le set to SW2
 	 * (possible only if response buffer size is larger than new Le = SW2)
 	 */
 	if (apdu->sw1 == 0x6C && (apdu->flags & SC_APDU_FLAGS_NO_RETRY_WL) == 0)
@@ -627,9 +619,18 @@ int sc_transmit_apdu(sc_card_t *card, sc_apdu_t *apdu)
 			len -= plen;
 			buf += plen;
 		}
-	} else
+	} else {
 		/* transmit single APDU */
 		r = sc_transmit(card, apdu);
+	}
+
+	if (r == SC_ERROR_CARD_RESET || r == SC_ERROR_READER_REATTACHED) {
+		sc_invalidate_cache(card);
+		/* give card driver a chance to react on resets */
+		if (card->ops->card_reader_lock_obtained)
+			card->ops->card_reader_lock_obtained(card, 1);
+	}
+
 	/* all done => release lock */
 	if (sc_unlock(card) != SC_SUCCESS)
 		sc_log(card->ctx, "sc_unlock failed");
