@@ -236,7 +236,7 @@ openssl_enc(const EVP_CIPHER * cipher, const unsigned char *key, const unsigned
 	r = SC_SUCCESS;
 out:
 	if (ctx)
-	    EVP_CIPHER_CTX_free(ctx);
+		EVP_CIPHER_CTX_free(ctx);
 	return r;
 }
 
@@ -376,10 +376,10 @@ openssl_dig(const EVP_MD * digest, const unsigned char *input, size_t length,
 
 	ctx = EVP_MD_CTX_create();
 	if (ctx == NULL) {
-	    r = SC_ERROR_OUT_OF_MEMORY;
-	    goto err;
+		r = SC_ERROR_OUT_OF_MEMORY;
+		goto err;
 	}
-	    
+		
 	EVP_MD_CTX_init(ctx);
 	EVP_DigestInit_ex(ctx, digest, NULL);
 	if (!EVP_DigestUpdate(ctx, input, length)) {
@@ -1157,6 +1157,7 @@ epass2003_init(struct sc_card *card)
 	unsigned char data[SC_MAX_APDU_BUFFER_SIZE] = { 0 };
 	size_t datalen = SC_MAX_APDU_BUFFER_SIZE;
 	epass2003_exdata *exdata = NULL;
+	void *old_drv_data = card->drv_data;
 
 	LOG_FUNC_CALLED(card->ctx);
 
@@ -1171,8 +1172,11 @@ epass2003_init(struct sc_card *card)
 	exdata->sm = SM_SCP01;
 
 	/* decide FIPS/Non-FIPS mode */
-	if (SC_SUCCESS != get_data(card, 0x86, data, datalen))
+	if (SC_SUCCESS != get_data(card, 0x86, data, datalen)) {
+		free(exdata);
+		card->drv_data = old_drv_data;
 		return SC_ERROR_INVALID_CARD;
+	}
 
 	if (0x01 == data[2])
 		exdata->smtype = KEY_TYPE_AES;
@@ -1365,7 +1369,7 @@ epass2003_select_fid(struct sc_card *card, unsigned int id_hi, unsigned int id_l
 		sc_file_t ** file_out)
 {
 	int r;
-	sc_file_t *file = 0;
+	sc_file_t *file = NULL;
 	sc_path_t path;
 
 	memset(&path, 0, sizeof(path));
@@ -1392,8 +1396,11 @@ epass2003_select_fid(struct sc_card *card, unsigned int id_hi, unsigned int id_l
 		}
 	}
 
-	if (file_out)
+	if (file_out) {
 		*file_out = file;
+	} else {
+		sc_file_free(file);
+	}
 
 	LOG_FUNC_RETURN(card->ctx, SC_SUCCESS);
 }
@@ -1586,7 +1593,7 @@ epass2003_set_security_env(struct sc_card *card, const sc_security_env_t * env,
 	exdata = (epass2003_exdata *)card->drv_data;
 
 	sc_format_apdu(card, &apdu, SC_APDU_CASE_3_SHORT, 0x22, 0x41, 0);
-    
+
 	p = sbuf;
 	*p++ = 0x80;		/* algorithm reference */
 	*p++ = 0x01;
@@ -1618,6 +1625,11 @@ epass2003_set_security_env(struct sc_card *card, const sc_security_env_t * env,
 			sbuf[2] = 0x92;
 			exdata->ecAlgFlags = SC_ALGORITHM_ECDSA_HASH_SHA256;
 		}
+		else if (env->algorithm_flags & SC_ALGORITHM_ECDSA_HASH_NONE)
+		{
+			sbuf[2] = 0x92;
+			exdata->ecAlgFlags = SC_ALGORITHM_ECDSA_HASH_NONE;
+		}
 		else
 		{
 			sc_log(card->ctx, "%0x Alg Not Support! ", env->algorithm_flags);
@@ -1714,6 +1726,13 @@ static int epass2003_decipher(struct sc_card *card, const u8 * data, size_t data
 			apdu.lc = 0x20;
 			apdu.datalen = 0x20;
 		}
+		else if (exdata->ecAlgFlags & SC_ALGORITHM_ECDSA_HASH_NONE)
+		{
+			sc_format_apdu(card, &apdu, SC_APDU_CASE_3,0x2A, 0x9E, 0x9A);
+			apdu.data = data;
+			apdu.lc = 0x20;
+			apdu.datalen = 0x20;
+		}
 		else
 		{
 			return SC_ERROR_NOT_SUPPORTED;
@@ -1808,7 +1827,7 @@ epass2003_process_fci(struct sc_card *card, sc_file_t * file, const u8 * buf, si
 		if (taglen == 2)
 			file->size = (file->size << 8) + tag[1];
 		sc_log(ctx, "  bytes in file: %"SC_FORMAT_LEN_SIZE_T"u",
-		       file->size);
+			   file->size);
 	}
 
 	if (tag == NULL) {
@@ -2003,9 +2022,9 @@ epass2003_construct_fci(struct sc_card *card, const sc_file_t * file,
 	}
 	else if (file->type == SC_FILE_TYPE_INTERNAL_EF) {
 		if (file->ef_structure == SC_CARDCTL_OBERTHUR_KEY_RSA_CRT ||
-		    file->ef_structure == SC_CARDCTL_OBERTHUR_KEY_RSA_PUBLIC||
-		    file->ef_structure == SC_CARDCTL_OBERTHUR_KEY_EC_CRT||
-		    file->ef_structure == SC_CARDCTL_OBERTHUR_KEY_EC_PUBLIC) {
+			file->ef_structure == SC_CARDCTL_OBERTHUR_KEY_RSA_PUBLIC||
+			file->ef_structure == SC_CARDCTL_OBERTHUR_KEY_EC_CRT||
+			file->ef_structure == SC_CARDCTL_OBERTHUR_KEY_EC_PUBLIC) {
 			buf[0] = (file->size >> 8) & 0xFF;
 			buf[1] = file->size & 0xFF;
 			sc_asn1_put_tag(0x85, buf, 2, p, *outlen - (p - out), &p);
@@ -2417,7 +2436,7 @@ epass2003_gen_key(struct sc_card *card, sc_epass2003_gen_key_data * data)
 	r = sc_transmit_apdu_t(card, &apdu);
 	LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
 	r = sc_check_sw(card, apdu.sw1, apdu.sw2);
-	LOG_TEST_RET(card->ctx, r, "generate keypair failed");
+	LOG_TEST_RET(card->ctx, r, "generate key pair failed");
 
 	/* read public key */
 	sc_format_apdu(card, &apdu, SC_APDU_CASE_3_SHORT, 0xb4, 0x02, 0x00);
@@ -2425,6 +2444,7 @@ epass2003_gen_key(struct sc_card *card, sc_epass2003_gen_key_data * data)
 	{
 		apdu.p1 = 0x00;
 	}
+
 	apdu.cla = 0x80;
 	apdu.lc = apdu.datalen = 2;
 	apdu.data = &sbuf[5];
@@ -2440,12 +2460,57 @@ epass2003_gen_key(struct sc_card *card, sc_epass2003_gen_key_data * data)
 	if (len < apdu.resplen)
 		LOG_FUNC_RETURN(card->ctx, SC_ERROR_INVALID_ARGUMENTS);
 
-	data->modulus = (u8 *) malloc(len);
-	if (!data->modulus)
-		LOG_FUNC_RETURN(card->ctx, SC_ERROR_OUT_OF_MEMORY);
+	if(256 == len)
+	{
+		int xCoordinateLen = rbuf[1];
+		int yCoordinateLen = rbuf[2+xCoordinateLen+1];
+		unsigned char * tmp =(u8 *)malloc(xCoordinateLen + yCoordinateLen);
+		if(!tmp)
+		{
+			LOG_FUNC_RETURN(card->ctx, SC_ERROR_OUT_OF_MEMORY);
+		}
 
-	memcpy(data->modulus, rbuf, len);
+		if(0x58 == rbuf[0])
+		{
+			memcpy(tmp, &rbuf[2], xCoordinateLen);
+		}
+		else{
+			free(tmp);
+			LOG_FUNC_RETURN(card->ctx, SC_ERROR_OBJECT_NOT_VALID);
+		}
+		if(0x59 == rbuf[2+xCoordinateLen])
+		{
+			memcpy(tmp + xCoordinateLen, &rbuf[2+xCoordinateLen+2], yCoordinateLen);
+		}
+		else{
+			free(tmp);
+			LOG_FUNC_RETURN(card->ctx, SC_ERROR_OBJECT_NOT_VALID);
+		}
 
+		data->modulus = (u8 *) malloc(xCoordinateLen + yCoordinateLen);
+		if (!data->modulus)
+		{
+			free(tmp);
+			LOG_FUNC_RETURN(card->ctx, SC_ERROR_OUT_OF_MEMORY);
+		}
+		else
+		{
+			memcpy(data->modulus, tmp, xCoordinateLen+yCoordinateLen);
+			free(tmp);
+		}
+	}
+	else
+	{
+		data->modulus = (u8 *) malloc(len);
+		if (!data->modulus)
+		{
+			LOG_FUNC_RETURN(card->ctx, SC_ERROR_OUT_OF_MEMORY);
+		}
+		else
+		{	
+			memcpy(data->modulus, rbuf, len);
+		}
+	}
 	LOG_FUNC_RETURN(card->ctx, SC_SUCCESS);
 }
 
@@ -2453,12 +2518,48 @@ epass2003_gen_key(struct sc_card *card, sc_epass2003_gen_key_data * data)
 static int
 epass2003_erase_card(struct sc_card *card)
 {
+	static const unsigned char install_magic_pin[26] = {
+		/* compare install_secret_key */
+		0x06,0x01,0x10,0x16, 0x16,0x16,0x00,0x0f, 0xff,0x66,
+		0x31,0x32,0x33,0x34, 0x35,0x36,0x37,0x38,
+		0x31,0x32,0x33,0x34, 0x35,0x36,0x37,0x38,
+	};
+	static const unsigned char magic_pin[16] = "1234567812345678";
+	static const unsigned char mf_path[2] = { 0x3f, 0x00 };
+	sc_apdu_t apdu;
 	int r;
 
 	LOG_FUNC_CALLED(card->ctx);
 	sc_invalidate_cache(card);
 
-	r = sc_delete_file(card, sc_get_mf_path());
+	/* install magic pin */
+	sc_format_apdu(card, &apdu, 0x03, 0xe3, 0x00, 0x00);
+	apdu.cla = 0x80;
+	apdu.data = install_magic_pin;
+	apdu.datalen = apdu.lc = sizeof(install_magic_pin);
+	r = sc_transmit_apdu(card, &apdu);
+	LOG_TEST_RET(card->ctx, r, "APDU install magic pin failed");
+	r = sc_check_sw(card, apdu.sw1, apdu.sw2);
+	LOG_TEST_RET(card->ctx, r, "install magic pin failed");
+
+	/* verify magic pin */
+	sc_format_apdu(card, &apdu, 0x03, 0x20, 0x00, 0x01);
+	apdu.cla = 0;
+	apdu.data = magic_pin;
+	apdu.datalen = apdu.lc = sizeof(magic_pin);
+	r = sc_transmit_apdu(card, &apdu);
+	LOG_TEST_RET(card->ctx, r, "APDU verify magic pin failed");
+	r = sc_check_sw(card, apdu.sw1, apdu.sw2);
+	LOG_TEST_RET(card->ctx, r, "verify magic pin failed");
+
+	/* delete MF */
+	sc_format_apdu(card, &apdu, 0x03, 0xe4, 0x00, 0x00);
+	apdu.cla = 0;
+	apdu.data = mf_path;
+	apdu.datalen = apdu.lc = sizeof(mf_path);
+	r = sc_transmit_apdu(card, &apdu);
+	LOG_TEST_RET(card->ctx, r, "APDU delete MF failed");
+	r = sc_check_sw(card, apdu.sw1, apdu.sw2);
 	LOG_TEST_RET(card->ctx, r, "delete MF failed");
 
 	LOG_FUNC_RETURN(card->ctx, r);
@@ -2671,6 +2772,10 @@ epass2003_pin_cmd(struct sc_card *card, struct sc_pin_cmd_data *data, int *tries
 	internal_sanitize_pin_info(&data->pin2, 1);
 	data->flags |= SC_PIN_CMD_NEED_PADDING;
 	kid = data->pin_reference;
+
+	if(NULL == (unsigned char *)data->pin1.data || 0 == data->pin1.len)
+		LOG_FUNC_RETURN(card->ctx, SC_ERROR_PIN_CODE_INCORRECT);
+
 	/* get pin retries */
 	if (data->cmd == SC_PIN_CMD_GET_INFO) {
 
@@ -2685,8 +2790,7 @@ epass2003_pin_cmd(struct sc_card *card, struct sc_pin_cmd_data *data, int *tries
 
 			data->pin1.max_tries = maxtries;
 		}
-//remove below code, because the old implement only return PIN retries, now modify the code and return PIN status
-//		return r;
+		LOG_TEST_RET(card->ctx, r, "verify pin failed");
 	}
 	else if (data->cmd == SC_PIN_CMD_UNBLOCK) { /* verify */
 		r = external_key_auth(card, (kid + 1), (unsigned char *)data->pin1.data,
@@ -2701,18 +2805,19 @@ epass2003_pin_cmd(struct sc_card *card, struct sc_pin_cmd_data *data, int *tries
 	else {
 		r = external_key_auth(card, kid, (unsigned char *)data->pin1.data,
 				data->pin1.len);
-		get_external_key_retries(card, 0x80 | kid, &retries);
+		LOG_TEST_RET(card->ctx, r, "verify pin failed");
+
+		r = get_external_key_retries(card, 0x80 | kid, &retries);
 		if (retries < pin_low)
 			sc_log(card->ctx, "Verification failed (remaining tries: %d)", retries);
 
+		LOG_TEST_RET(card->ctx, r, "verify pin failed");
 	}
-	LOG_TEST_RET(card->ctx, r, "verify pin failed");
 
 	if (r == SC_SUCCESS)
 	{
 		data->pin1.logged_in = SC_PIN_STATE_LOGGED_IN;
 	}
-
 	return r;
 }
 
