@@ -28,7 +28,7 @@
 #include "internal.h"
 #include "cardctl.h"
 
-static struct sc_atr_table jcop_atrs[] = {
+static const struct sc_atr_table jcop_atrs[] = {
 	{ "3B:E6:00:FF:81:31:FE:45:4A:43:4F:50:33:31:06", NULL, NULL, SC_CARD_TYPE_JCOP_GENERIC, 0, NULL },
 	{ NULL, NULL, NULL, 0, 0, NULL }
 };
@@ -539,7 +539,9 @@ static int jcop_create_file(sc_card_t *card, sc_file_t *file) {
 	  
 	  entry = sc_file_get_acl_entry(file, ops[i]);
 	  r = acl_to_ac_nibble(entry);
-	  sec_attr_data[i/2] |= r << ((i % 2) ? 0 : 4);
+	  if (r >= 0) {
+		sec_attr_data[i/2] |= r << ((i % 2) ? 0 : 4);
+	  }
      }
 
      sc_file_set_sec_attr(file, sec_attr_data, 3);
@@ -610,7 +612,7 @@ static int jcop_set_security_env(sc_card_t *card,
 
         assert(card != NULL && env != NULL);
 	if (se_num) 
-	     SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_NORMAL, SC_ERROR_INVALID_ARGUMENTS);
+	     LOG_FUNC_RETURN(card->ctx, SC_ERROR_INVALID_ARGUMENTS);
 	if (drvdata->selected == SELECT_MF || 
 	    drvdata->selected == SELECT_EFDIR) {
 	     drvdata->invalid_senv=1;
@@ -624,11 +626,11 @@ static int jcop_set_security_env(sc_card_t *card,
                 tmp.flags &= ~SC_SEC_ENV_ALG_PRESENT;
                 tmp.flags |= SC_SEC_ENV_ALG_REF_PRESENT;
                 if (tmp.algorithm != SC_ALGORITHM_RSA) {
-                        sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "Only RSA algorithm supported.\n");
+                        sc_log(card->ctx,  "Only RSA algorithm supported.\n");
                         return SC_ERROR_NOT_SUPPORTED;
                 }
                 if (!(env->algorithm_flags & SC_ALGORITHM_RSA_PAD_PKCS1)){
-                        sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "Card requires RSA padding\n");
+                        sc_log(card->ctx,  "Card requires RSA padding\n");
                         return SC_ERROR_NOT_SUPPORTED;
                 }
                 tmp.algorithm_ref = 0x02;
@@ -681,13 +683,13 @@ static int jcop_set_security_env(sc_card_t *card,
         apdu.resplen = 0;
 	r = sc_transmit_apdu(card, &apdu);
 	if (r) {
-	     sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL,
+	     sc_log(card->ctx, 
 		"%s: APDU transmit failed", sc_strerror(r));
 	     return r;
 	}
 	r = sc_check_sw(card, apdu.sw1, apdu.sw2);
 	if (r) {
-	     sc_debug(card->ctx,  SC_LOG_DEBUG_NORMAL,
+	     sc_log(card->ctx,  
 		"%s: Card returned error", sc_strerror(r));
 	     return r;
 	}
@@ -733,7 +735,7 @@ static int jcop_compute_signature(sc_card_t *card,
 
         apdu.data = sbuf;
         r = sc_transmit_apdu(card, &apdu);
-        SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
+        LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
         if (apdu.sw1 == 0x90 && apdu.sw2 == 0x00) {
                 int len = apdu.resplen > outlen ? outlen : apdu.resplen;
 
@@ -756,7 +758,7 @@ static int jcop_decipher(sc_card_t *card,
 	struct jcop_private_data *drvdata=DRVDATA(card);
 
         assert(card != NULL && crgram != NULL && out != NULL);
-        SC_FUNC_CALLED(card->ctx, SC_LOG_DEBUG_NORMAL);
+        LOG_FUNC_CALLED(card->ctx);
         if (crgram_len > 256)
                 SC_FUNC_RETURN(card->ctx, SC_LOG_DEBUG_VERBOSE, SC_ERROR_INVALID_ARGUMENTS);
 	if (drvdata->invalid_senv)
@@ -784,7 +786,7 @@ static int jcop_decipher(sc_card_t *card,
 	
         apdu.data = sbuf;
         r = sc_transmit_apdu(card, &apdu);
-        SC_TEST_RET(card->ctx, SC_LOG_DEBUG_NORMAL, r, "APDU transmit failed");
+        LOG_TEST_RET(card->ctx, r, "APDU transmit failed");
         if (apdu.sw1 == 0x90 && apdu.sw2 == 0x00) {
                 int len = apdu.resplen > outlen ? outlen : apdu.resplen;
 
@@ -811,7 +813,7 @@ static int jcop_generate_key(sc_card_t *card, struct sc_cardctl_jcop_genkey *a)
      if (a->exponent == 0x10001) {
 	  is_f4=1;
      } else if (a->exponent != 3) {
-	  sc_debug(card->ctx,  SC_LOG_DEBUG_NORMAL,
+	  sc_log(card->ctx,  
 		"%s: Invalid exponent", sc_strerror(SC_ERROR_NOT_SUPPORTED));
 	  return SC_ERROR_NOT_SUPPORTED;
      }
@@ -841,13 +843,13 @@ static int jcop_generate_key(sc_card_t *card, struct sc_cardctl_jcop_genkey *a)
      apdu.resplen = 0;
      r = sc_transmit_apdu(card, &apdu);
      if (r) {
-	  sc_debug(card->ctx,  SC_LOG_DEBUG_NORMAL,
+	  sc_log(card->ctx,  
 		"%s: APDU transmit failed", sc_strerror(r));
 	  return r;
      }
      r = sc_check_sw(card, apdu.sw1, apdu.sw2);
      if (r) {
-	  sc_debug(card->ctx,  SC_LOG_DEBUG_NORMAL,
+	  sc_log(card->ctx,  
 	  	"%s: Card returned error", sc_strerror(r));
 	  return r;
      }
@@ -860,13 +862,13 @@ static int jcop_generate_key(sc_card_t *card, struct sc_cardctl_jcop_genkey *a)
      
      r = sc_transmit_apdu(card, &apdu);
      if (r) {
-	  sc_debug(card->ctx,  SC_LOG_DEBUG_NORMAL,
+	  sc_log(card->ctx,  
 		"%s: APDU transmit failed", sc_strerror(r));
 	  return r;
      }
      r = sc_check_sw(card, apdu.sw1, apdu.sw2);
      if (r) {
-	  sc_debug(card->ctx,  SC_LOG_DEBUG_NORMAL,
+	  sc_log(card->ctx,  
 		"%s: Card returned error", sc_strerror(r));
 	  return r;
      }
