@@ -88,7 +88,7 @@ static int westcos_get_default_key(sc_card_t * card,
 				   struct sc_cardctl_default_key *data)
 {
 	const char *default_key;
-	sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL,
+	sc_log(card->ctx, 
 		 "westcos_get_default_key:data->method=%d, data->key_ref=%d\n",
 		 data->method, data->key_ref);
 	if (data->method != SC_AC_AUT || data->key_ref != 0)
@@ -221,6 +221,23 @@ static int westcos_init(sc_card_t * card)
 	
 	priv_data = (priv_data_t *) card->drv_data;
 
+	default_key =
+	    scconf_get_str(card->ctx->conf_blocks[0], "westcos_default_key",
+			   DEFAULT_TRANSPORT_KEY);
+	if (default_key) {
+		priv_data = (priv_data_t *) (card->drv_data);
+		priv_data->default_key.key_reference = 0;
+		priv_data->default_key.key_len =
+			sizeof(priv_data->default_key.key_value);
+		r = sc_hex_to_bin(default_key, priv_data->default_key.key_value,
+				&(priv_data->default_key.key_len));
+		if (r) {
+			free (priv_data);
+			card->drv_data = NULL;
+			return (r);
+		}
+	}
+
 	if (card->type & JAVACARD) {
 		priv_data->flags |= JAVACARD;
 	}
@@ -250,19 +267,6 @@ static int westcos_init(sc_card_t * card)
 	_sc_card_add_rsa_alg(card, 1400, flags, exponent);
 	_sc_card_add_rsa_alg(card, 1536, flags, exponent);
 	_sc_card_add_rsa_alg(card, 2048, flags, exponent);
-	default_key =
-	    scconf_get_str(card->ctx->conf_blocks[0], "westcos_default_key",
-			   DEFAULT_TRANSPORT_KEY);
-	if (default_key) {
-		priv_data = (priv_data_t *) (card->drv_data);
-		priv_data->default_key.key_reference = 0;
-		priv_data->default_key.key_len =
-			sizeof(priv_data->default_key.key_value);
-		r = sc_hex_to_bin(default_key, priv_data->default_key.key_value,
-				&(priv_data->default_key.key_len));
-		if (r)
-			return (r);
-	}
 	return 0;
 }
 
@@ -295,17 +299,17 @@ static int westcos_process_fci(sc_card_t * card, sc_file_t * file,
 	sc_context_t *ctx = card->ctx;
 	size_t taglen, len = buflen;
 	const u8 *tag = NULL, *p = buf;
-	sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "processing FCI bytes\n");
+	sc_log(card->ctx,  "processing FCI bytes\n");
 	tag = sc_asn1_find_tag(ctx, p, len, 0x83, &taglen);
 	if (tag != NULL && taglen == 2) {
 		file->id = (tag[0] << 8) | tag[1];
-		sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL,
+		sc_log(card->ctx, 
 			"  file identifier: 0x%02X%02X\n", tag[0], tag[1]);
 	}
 	tag = sc_asn1_find_tag(ctx, p, len, 0x80, &taglen);
 	if (tag != NULL && taglen >= 2) {
 		int bytes = (tag[0] << 8) + tag[1];
-		sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL,
+		sc_log(card->ctx, 
 			"  bytes in file: %d\n", bytes);
 		file->size = bytes;
 	}
@@ -313,7 +317,7 @@ static int westcos_process_fci(sc_card_t * card, sc_file_t * file,
 		tag = sc_asn1_find_tag(ctx, p, len, 0x81, &taglen);
 		if (tag != NULL && taglen >= 2) {
 			int bytes = (tag[0] << 8) + tag[1];
-			sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL,
+			sc_log(card->ctx, 
 				"  bytes in file: %d\n", bytes);
 			file->size = bytes;
 		}
@@ -324,7 +328,7 @@ static int westcos_process_fci(sc_card_t * card, sc_file_t * file,
 			unsigned char byte = tag[0];
 			const char *type;
 			file->shareable = 0;
-			sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL,
+			sc_log(card->ctx, 
 				"  shareable: %s\n",
 				 (file->shareable) ? "yes" : "no");
 			file->ef_structure = SC_FILE_EF_UNKNOWN;
@@ -351,9 +355,9 @@ static int westcos_process_fci(sc_card_t * card, sc_file_t * file,
 			default:
 				type = "unknown";
 			}
-			sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL,
+			sc_log(card->ctx, 
 				"  type: %s\n", type);
-			sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL,
+			sc_log(card->ctx, 
 				"  EF structure: %d\n", file->ef_structure);
 		}
 	}
@@ -507,7 +511,7 @@ static int westcos_create_file(sc_card_t *card, struct sc_file *file)
 	int buflen;
 	if (card == NULL)
 		return SC_ERROR_INVALID_ARGUMENTS;
-	sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "westcos_create_file\n");
+	sc_log(card->ctx,  "westcos_create_file\n");
 	memset(buf, 0, sizeof(buf));
 
 	/* transport key */
@@ -593,7 +597,7 @@ static int westcos_create_file(sc_card_t *card, struct sc_file *file)
 		p1 = (file->id) / 256;
 		p2 = (file->id) % 256;
 	}
-	sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL,
+	sc_log(card->ctx, 
 		 "create file %s, id %X size %"SC_FORMAT_LEN_SIZE_T"u\n",
 		 file->path.value, file->id, file->size);
 	sc_format_apdu(card, &apdu, SC_APDU_CASE_3_SHORT, 0xE0, p1, p2);
@@ -614,7 +618,7 @@ static int westcos_delete_file(sc_card_t * card, const sc_path_t * path_in)
 	sc_apdu_t apdu;
 	if (card == NULL || path_in == NULL || path_in->len < 2)
 		return SC_ERROR_INVALID_ARGUMENTS;
-	sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "westcos_delete_file\n");
+	sc_log(card->ctx,  "westcos_delete_file\n");
 	if (path_in->len > 2) {
 		r = sc_select_file(card, path_in, NULL);
 		if (r)
@@ -639,7 +643,7 @@ static int westcos_list_files(sc_card_t * card, u8 * buf, size_t buflen)
 	sc_apdu_t apdu;
 	if (card == NULL)
 		return SC_ERROR_INVALID_ARGUMENTS;
-	sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "westcos_list_files\n");
+	sc_log(card->ctx,  "westcos_list_files\n");
 	sc_format_apdu(card, &apdu, SC_APDU_CASE_2_SHORT, 0x34, 0x00, 0x00);
 	apdu.cla = 0x80;
 	apdu.le = buflen;
@@ -688,7 +692,7 @@ static int westcos_pin_cmd(sc_card_t * card, struct sc_pin_cmd_data *data,
 	int pad = 0, use_pin_pad = 0, ins, p1 = 0;
 	if (card == NULL)
 		return SC_ERROR_INVALID_ARGUMENTS;
-	sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL,
+	sc_log(card->ctx, 
 		 "westcos_pin_cmd:data->pin_type=%X, data->cmd=%X\n",
 		 data->pin_type, data->cmd);
 	if (tries_left)
@@ -859,7 +863,7 @@ static int westcos_card_ctl(sc_card_t * card, unsigned long cmd, void *ptr)
 	priv_data_t *priv_data = NULL;
 	if (card == NULL)
 		return SC_ERROR_INVALID_ARGUMENTS;
-	sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL,
+	sc_log(card->ctx, 
 		"westcos_card_ctl cmd = %lX\n", cmd);
 	priv_data = (priv_data_t *) card->drv_data;
 	switch (cmd) {
@@ -1044,7 +1048,7 @@ static int westcos_set_security_env(sc_card_t *card,
 	priv_data_t *priv_data = NULL;
 	if (card == NULL)
 		return SC_ERROR_INVALID_ARGUMENTS;
-	sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL,
+	sc_log(card->ctx, 
 		"westcos_set_security_env\n");
 	priv_data = (priv_data_t *) card->drv_data;
 	priv_data->env = *env;
@@ -1082,7 +1086,7 @@ static int westcos_restore_security_env(sc_card_t *card, int se_num)
 {
 	if (card == NULL)
 		return SC_ERROR_INVALID_ARGUMENTS;
-	sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL,
+	sc_log(card->ctx, 
 		"westcos_restore_security_env\n");
 	return 0;
 }
@@ -1105,7 +1109,7 @@ static int westcos_sign_decipher(int mode, sc_card_t *card,
 	if (card == NULL)
 		return SC_ERROR_INVALID_ARGUMENTS;
 
-	sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL,
+	sc_log(card->ctx, 
 		 "westcos_sign_decipher outlen=%"SC_FORMAT_LEN_SIZE_T"u\n",
 		 outlen);
 
@@ -1161,7 +1165,7 @@ static int westcos_sign_decipher(int mode, sc_card_t *card,
 		alire = min(((keyfile->size) - idx), sizeof(buf));
 		if (alire <= 0)
 			break;
-		sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL,
+		sc_log(card->ctx, 
 			"idx = %d, alire=%d\n", idx, alire);
 		r = sc_read_binary(card, idx, buf, alire, 0);
 		if (r < 0)
@@ -1171,7 +1175,7 @@ static int westcos_sign_decipher(int mode, sc_card_t *card,
 	} while (1);
 	BIO_set_mem_eof_return(mem, -1);
 	if (!d2i_RSAPrivateKey_bio(mem, &rsa)) {
-		sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL,
+		sc_log(card->ctx, 
 			"RSA key invalid, %lu\n", ERR_get_error());
 		r = SC_ERROR_UNKNOWN;
 		goto out;
@@ -1181,7 +1185,7 @@ static int westcos_sign_decipher(int mode, sc_card_t *card,
 	RSA_set_method(rsa, RSA_PKCS1_OpenSSL());
 
 	if ((size_t)RSA_size(rsa) > outlen) {
-		sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "Buffer too small\n");
+		sc_log(card->ctx,  "Buffer too small\n");
 		r = SC_ERROR_OUT_OF_MEMORY;
 		goto out;
 	}
@@ -1194,7 +1198,7 @@ static int westcos_sign_decipher(int mode, sc_card_t *card,
 			print_openssl_error();
 
 #endif
-			sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL,
+			sc_log(card->ctx, 
 				"Decipher error %lu\n", ERR_get_error());
 			r = SC_ERROR_UNKNOWN;
 			goto out;
@@ -1210,7 +1214,7 @@ static int westcos_sign_decipher(int mode, sc_card_t *card,
 			print_openssl_error();
 
 #endif
-			sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL,
+			sc_log(card->ctx, 
 				"Signature error %lu\n", ERR_get_error());
 			r = SC_ERROR_UNKNOWN;
 			goto out;
@@ -1219,7 +1223,7 @@ static int westcos_sign_decipher(int mode, sc_card_t *card,
 
 #else
 	if (RSA_sign(nid, data, data_len, out, &outlen, rsa) != 1) {
-		sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL,
+		sc_log(card->ctx, 
 			"RSA_sign error %d \n", ERR_get_error());
 		r = SC_ERROR_UNKNOWN;
 		goto out;
