@@ -28,15 +28,13 @@
 #include <unistd.h>
 #endif
 #include <string.h>
+#include <limits.h>
 
+#include "reader-tr03119.h"
 #include "internal.h"
 #include "asn1.h"
 #include "common/compat_strlcpy.h"
 
-/*
-#define INVALIDATE_CARD_CACHE_IN_UNLOCK
-*/
-
 #ifdef ENABLE_SM
 static int sc_card_sm_load(sc_card_t *card, const char *path, const char *module);
 static int sc_card_sm_unload(sc_card_t *card);
@@ -55,7 +53,9 @@ int sc_check_sw(sc_card_t *card, unsigned int sw1, unsigned int sw2)
 void sc_format_apdu(sc_card_t *card, sc_apdu_t *apdu,
 		    int cse, int ins, int p1, int p2)
 {
-	assert(card != NULL && apdu != NULL);
+	if (card == NULL || apdu == NULL) {
+		return;
+	}
 	memset(apdu, 0, sizeof(*apdu));
 	apdu->cla = (u8) card->cla;
 	apdu->cse = cse;
@@ -64,6 +64,65 @@ void sc_format_apdu(sc_card_t *card, sc_apdu_t *apdu,
 	apdu->p2 = (u8) p2;
 }
 
+void sc_format_apdu_cse_lc_le(struct sc_apdu *apdu)
+{
+	/* TODO calculating the APDU case, Lc and Le should actually only be
+	 * done in sc_apdu2bytes, but to gradually change OpenSC we start here. */
+	/* Let sc_detect_apdu_cse set short or extended  and test for chaining */
+
+	if (!apdu)
+		return;
+	if (apdu->datalen > SC_MAX_APDU_DATA_SIZE
+			|| apdu->resplen > SC_MAX_APDU_RESP_SIZE) {
+		/* extended length  or data chaining and/or get response */
+		if (apdu->datalen <= SC_MAX_EXT_APDU_DATA_SIZE)
+			apdu->lc = apdu->datalen;
+		if (apdu->resplen <= SC_MAX_EXT_APDU_RESP_SIZE)
+			apdu->le = apdu->resplen;
+		if (apdu->resplen && !apdu->datalen)
+			apdu->cse = SC_APDU_CASE_2;
+		if (!apdu->resplen && apdu->datalen)
+			apdu->cse = SC_APDU_CASE_3;
+		if (apdu->resplen && apdu->datalen)
+			apdu->cse = SC_APDU_CASE_4;
+	} else {
+		/* short length */
+		if (apdu->datalen <= SC_MAX_APDU_DATA_SIZE)
+			apdu->lc = apdu->datalen;
+		if (apdu->resplen <= SC_MAX_APDU_RESP_SIZE)
+			apdu->le = apdu->resplen;
+		if (!apdu->resplen && !apdu->datalen)
+			apdu->cse = SC_APDU_CASE_1;
+		if (apdu->resplen && !apdu->datalen)
+			apdu->cse = SC_APDU_CASE_2_SHORT;
+		if (!apdu->resplen && apdu->datalen)
+			apdu->cse = SC_APDU_CASE_3_SHORT;
+		if (apdu->resplen && apdu->datalen)
+			apdu->cse = SC_APDU_CASE_4_SHORT;
+	}
+}
+
+void sc_format_apdu_ex(struct sc_apdu *apdu,
+		u8 cla, u8 ins, u8 p1, u8 p2,
+		const u8 *data, size_t datalen,
+		u8 *resp, size_t resplen)
+{
+	if (!apdu) {
+		return;
+	}
+
+	memset(apdu, 0, sizeof(*apdu));
+	apdu->cla = cla;
+	apdu->ins = ins;
+	apdu->p1 = p1;
+	apdu->p2 = p2;
+	apdu->resp = resp;
+	apdu->resplen = resplen;
+	apdu->data = data;
+	apdu->datalen = datalen;
+	sc_format_apdu_cse_lc_le(apdu);
+}
+
 static sc_card_t * sc_card_new(sc_context_t *ctx)
 {
 	sc_card_t *card;
@@ -98,9 +157,6 @@ static void sc_card_free(sc_card_t *card)
 	sc_free_apps(card);
 	sc_free_ef_atr(card);
 
-	if (card->ef_dir != NULL)
-		sc_file_free(card->ef_dir);
-
 	free(card->ops);
 
 	if (card->algorithms != NULL)   {
@@ -120,11 +176,8 @@ static void sc_card_free(sc_card_t *card)
 		card->algorithm_count = 0;
 	}
 
-	if (card->cache.current_ef)
-		sc_file_free(card->cache.current_ef);
-
-	if (card->cache.current_df)
-		sc_file_free(card->cache.current_df);
+	sc_file_free(card->cache.current_ef);
+	sc_file_free(card->cache.current_df);
 
 	if (card->mutex != NULL) {
 		int r = sc_mutex_destroy(card->ctx, card->mutex);
@@ -138,10 +191,12 @@ static void sc_card_free(sc_card_t *card)
 size_t sc_get_max_recv_size(const sc_card_t *card)
 {
 	size_t max_recv_size;
-	assert(card != NULL && card->reader != NULL);
+	if (card == NULL || card->reader == NULL) {
+		return 0;
+	}
 	max_recv_size = card->max_recv_size;
 
-	/* initialize max_recv_size to a meaningfull value */
+	/* initialize max_recv_size to a meaningful value */
 	if (card->caps & SC_CARD_CAP_APDU_EXT) {
 		if (!max_recv_size)
 			max_recv_size = 65536;
@@ -162,12 +217,15 @@ size_t sc_get_max_send_size(const sc_card_t *card)
 {
 	size_t max_send_size;
 
-	assert(card != NULL && card->reader != NULL);
+	if (card == NULL || card->reader == NULL) {
+		return 0;
+	}
 
 	max_send_size = card->max_send_size;
 
-	/* initialize max_send_size to a meaningfull value */
-	if (card->caps & SC_CARD_CAP_APDU_EXT) {
+	/* initialize max_send_size to a meaningful value */
+	if (card->caps & SC_CARD_CAP_APDU_EXT
+			&& card->reader->active_protocol != SC_PROTO_T0) {
 		if (!max_send_size)
 			max_send_size = 65535;
 	} else {
@@ -208,7 +266,11 @@ int sc_connect_card(sc_reader_t *reader, sc_card_t **card_out)
 	card->reader = reader;
 	card->ctx = ctx;
 
+	if (reader->flags & SC_READER_ENABLE_ESCAPE)
+		sc_detect_escape_cmds(reader);
+
 	memcpy(&card->atr, &reader->atr, sizeof(card->atr));
+	memcpy(&card->uid, &reader->uid, sizeof(card->uid));
 
 	_sc_parse_atr(reader);
 
@@ -257,8 +319,22 @@ int sc_connect_card(sc_reader_t *reader, sc_card_t **card_out)
 		}
 	}
 	else {
+		sc_card_t uninitialized = *card;
 		sc_log(ctx, "matching built-in ATRs");
 		for (i = 0; ctx->card_drivers[i] != NULL; i++) {
+			/* FIXME If we had a clean API description, we'd probably get a
+			 * cleaner implementation of the driver's match_card and init,
+			 * which should normally *not* modify the card object if
+			 * unsuccessful. However, after years of relentless hacking, reality
+			 * is different: The card object is changed in virtually every card
+			 * driver so in order to prevent unwanted interaction, we reset the
+			 * card object here and hope that the card driver at least doesn't
+			 * allocate any internal resources that need to be freed. If we
+			 * had more time, we should refactor the existing code to not
+			 * modify sc_card_t until complete success (possibly by combining
+			 * `match_card()` and `init()`) */
+			*card = uninitialized;
+
 			struct sc_card_driver *drv = ctx->card_drivers[i];
 			const struct sc_card_operations *ops = drv->ops;
 
@@ -299,12 +375,14 @@ int sc_connect_card(sc_reader_t *reader, sc_card_t **card_out)
 	if (card->name == NULL)
 		card->name = card->driver->name;
 
-	/* initialize max_send_size/max_recv_size to a meaningfull value */
+	/* initialize max_send_size/max_recv_size to a meaningful value */
 	card->max_recv_size = sc_get_max_recv_size(card);
 	card->max_send_size = sc_get_max_send_size(card);
 
-	sc_log(ctx, "card info name:'%s', type:%i, flags:0x%X, max_send/recv_size:%i/%i",
-		card->name, card->type, card->flags, card->max_send_size, card->max_recv_size);
+	sc_log(ctx,
+	       "card info name:'%s', type:%i, flags:0x%lX, max_send/recv_size:%"SC_FORMAT_LEN_SIZE_T"u/%"SC_FORMAT_LEN_SIZE_T"u",
+	       card->name, card->type, card->flags, card->max_send_size,
+	       card->max_recv_size);
 
 #ifdef ENABLE_SM
         /* Check, if secure messaging module present. */
@@ -335,7 +413,8 @@ int sc_disconnect_card(sc_card_t *card)
 	ctx = card->ctx;
 	LOG_FUNC_CALLED(ctx);
 
-	assert(card->lock_count == 0);
+	if (card->lock_count != 0)
+		return SC_ERROR_NOT_ALLOWED;
 	if (card->ops->finish) {
 		int r = card->ops->finish(card);
 		if (r)
@@ -371,9 +450,7 @@ int sc_reset(sc_card_t *card, int do_cold_reset)
 		return r;
 
 	r = card->reader->ops->reset(card->reader, do_cold_reset);
-	/* invalidate cache */
-	memset(&card->cache, 0, sizeof(card->cache));
-	card->cache.valid = 0;
+	sc_invalidate_cache(card);
 
 	r2 = sc_mutex_unlock(card->ctx, card->mutex);
 	if (r2 != SC_SUCCESS) {
@@ -387,6 +464,8 @@ int sc_reset(sc_card_t *card, int do_cold_reset)
 int sc_lock(sc_card_t *card)
 {
 	int r = 0, r2 = 0;
+	int was_reset = 0;
+	int reader_lock_obtained  = 0;
 
 	if (card == NULL)
 		return SC_ERROR_INVALID_ARGUMENTS;
@@ -399,29 +478,39 @@ int sc_lock(sc_card_t *card)
 	if (card->lock_count == 0) {
 		if (card->reader->ops->lock != NULL) {
 			r = card->reader->ops->lock(card->reader);
-			if (r == SC_ERROR_CARD_RESET || r == SC_ERROR_READER_REATTACHED) {
-				/* invalidate cache */
-				memset(&card->cache, 0, sizeof(card->cache));
-				card->cache.valid = 0;
-#ifdef ENABLE_SM
-				if (card->sm_ctx.ops.open)
-					card->sm_ctx.ops.open(card);
-#endif
+			while (r == SC_ERROR_CARD_RESET || r == SC_ERROR_READER_REATTACHED) {
+				sc_invalidate_cache(card);
+				if (was_reset++ > 4) /* TODO retry a few times */
+					break;
 				r = card->reader->ops->lock(card->reader);
 			}
+			if (r == 0)
+				reader_lock_obtained = 1;
 		}
 		if (r == 0)
 			card->cache.valid = 1;
 	}
 	if (r == 0)
 		card->lock_count++;
+
+	if (r == 0 && was_reset > 0) {
+#ifdef ENABLE_SM
+		if (card->sm_ctx.ops.open)
+			card->sm_ctx.ops.open(card);
+#endif
+	}
+
 	r2 = sc_mutex_unlock(card->ctx, card->mutex);
 	if (r2 != SC_SUCCESS) {
-		sc_log(card->ctx, "unable to release lock");
+		sc_log(card->ctx, "unable to release card->mutex lock");
 		r = r != SC_SUCCESS ? r : r2;
 	}
 
-	return r;
+	/* give card driver a chance to do something when reader lock first obtained */
+	if (r == 0 && reader_lock_obtained == 1  && card->ops->card_reader_lock_obtained)
+		r = card->ops->card_reader_lock_obtained(card, was_reset);
+
+	LOG_FUNC_RETURN(card->ctx, r);
 }
 
 int sc_unlock(sc_card_t *card)
@@ -437,14 +526,16 @@ int sc_unlock(sc_card_t *card)
 	if (r != SC_SUCCESS)
 		return r;
 
-	assert(card->lock_count >= 1);
+	if (card->lock_count < 1) {
+		return SC_ERROR_INVALID_ARGUMENTS;
+	}
 	if (--card->lock_count == 0) {
-#ifdef INVALIDATE_CARD_CACHE_IN_UNLOCK
-		/* invalidate cache */
-		memset(&card->cache, 0, sizeof(card->cache));
-		card->cache.valid = 0;
-		sc_log(card->ctx, "cache invalidated");
-#endif
+		if (card->flags & SC_CARD_FLAG_KEEP_ALIVE) {
+			/* Multiple processes accessing the card will most likely render
+			 * the card cache useless. To not have a bad cache, we explicitly
+			 * invalidate it. */
+			sc_invalidate_cache(card);
+		}
 		/* release reader lock */
 		if (card->reader->ops->unlock != NULL)
 			r = card->reader->ops->unlock(card->reader);
@@ -462,7 +553,9 @@ int sc_list_files(sc_card_t *card, u8 *buf, size_t buflen)
 {
 	int r;
 
-	assert(card != NULL);
+	if (card == NULL) {
+		return SC_ERROR_INVALID_ARGUMENTS;
+	}
 	LOG_FUNC_CALLED(card->ctx);
 
 	if (card->ops->list_files == NULL)
@@ -478,14 +571,18 @@ int sc_create_file(sc_card_t *card, sc_file_t *file)
 	char pbuf[SC_MAX_PATH_STRING_SIZE];
 	const sc_path_t *in_path;
 
-	assert(card != NULL && file != NULL);
+	if (card == NULL || file == NULL) {
+		return SC_ERROR_INVALID_ARGUMENTS;
+	}
 
 	in_path = &file->path;
 	r = sc_path_print(pbuf, sizeof(pbuf), in_path);
 	if (r != SC_SUCCESS)
 		pbuf[0] = '\0';
 
-	sc_log(card->ctx, "called; type=%d, path=%s, id=%04i, size=%u",  in_path->type, pbuf, file->id, file->size);
+	sc_log(card->ctx,
+	       "called; type=%d, path=%s, id=%04i, size=%"SC_FORMAT_LEN_SIZE_T"u",
+	       in_path->type, pbuf, file->id, file->size);
 	/* ISO 7816-4: "Number of data bytes in the file, including structural information if any"
 	 * can not be bigger than two bytes */
 	if (file->size > 0xFFFF)
@@ -503,7 +600,9 @@ int sc_delete_file(sc_card_t *card, const sc_path_t *path)
 	int r;
 	char pbuf[SC_MAX_PATH_STRING_SIZE];
 
-	assert(card != NULL);
+	if (card == NULL) {
+		return SC_ERROR_INVALID_ARGUMENTS;
+	}
 
 	r = sc_path_print(pbuf, sizeof(pbuf), path);
 	if (r != SC_SUCCESS)
@@ -521,12 +620,16 @@ int sc_read_binary(sc_card_t *card, unsigned int idx,
 		   unsigned char *buf, size_t count, unsigned long flags)
 {
 	size_t max_le = sc_get_max_recv_size(card);
+	size_t todo = count;
 	int r;
 
-	assert(card != NULL && card->ops != NULL && buf != NULL);
-	sc_log(card->ctx, "called; %d bytes at index %d", count, idx);
+	if (card == NULL || card->ops == NULL || buf == NULL) {
+		return SC_ERROR_INVALID_ARGUMENTS;
+	}
+	sc_log(card->ctx, "called; %"SC_FORMAT_LEN_SIZE_T"u bytes at index %d",
+	       count, idx);
 	if (count == 0)
-		return 0;
+		LOG_FUNC_RETURN(card->ctx, SC_SUCCESS);
 
 #ifdef ENABLE_SM
 	if (card->sm_ctx.ops.read_binary)   {
@@ -535,91 +638,102 @@ int sc_read_binary(sc_card_t *card, unsigned int idx,
 			LOG_FUNC_RETURN(card->ctx, r);
 	}
 #endif
+
 	if (card->ops->read_binary == NULL)
 		LOG_FUNC_RETURN(card->ctx, SC_ERROR_NOT_SUPPORTED);
 
-	if (count > max_le) {
-		int bytes_read = 0;
-		unsigned char *p = buf;
+	/* lock the card now to avoid deselection of the file */
+	r = sc_lock(card);
+	LOG_TEST_RET(card->ctx, r, "sc_lock() failed");
 
-		r = sc_lock(card);
-		LOG_TEST_RET(card->ctx, r, "sc_lock() failed");
-		while (count > 0) {
-			size_t n = count > max_le ? max_le : count;
-			r = sc_read_binary(card, idx, p, n, flags);
-			if (r < 0) {
-				sc_unlock(card);
-				LOG_TEST_RET(card->ctx, r, "sc_read_binary() failed");
-			}
-			p += r;
-			idx += r;
-			bytes_read += r;
-			count -= r;
-			if (r == 0) {
-				sc_unlock(card);
-				LOG_FUNC_RETURN(card->ctx, bytes_read);
-			}
+	while (todo > 0) {
+		size_t chunk = todo > max_le ? max_le : todo;
+
+		r = card->ops->read_binary(card, idx, buf, chunk, flags);
+		if (r == 0 || r == SC_ERROR_FILE_END_REACHED)
+			break;
+		if ((idx > SIZE_MAX - (size_t) r)
+				|| (size_t) r > todo) {
+			/* `idx + r` or `todo - r` would overflow */
+			r = SC_ERROR_OFFSET_TOO_LARGE;
 		}
-		sc_unlock(card);
-		LOG_FUNC_RETURN(card->ctx, bytes_read);
+		if (r < 0) {
+			sc_unlock(card);
+			LOG_FUNC_RETURN(card->ctx, r);
+		}
+
+		todo -= (size_t) r;
+		buf  += (size_t) r;
+		idx  += (size_t) r;
 	}
-	r = card->ops->read_binary(card, idx, buf, count, flags);
-	LOG_FUNC_RETURN(card->ctx, r);
+
+	sc_unlock(card);
+
+	LOG_FUNC_RETURN(card->ctx, count - todo);
 }
 
 int sc_write_binary(sc_card_t *card, unsigned int idx,
 		    const u8 *buf, size_t count, unsigned long flags)
 {
 	size_t max_lc = sc_get_max_send_size(card);
+	size_t todo = count;
 	int r;
 
-	assert(card != NULL && card->ops != NULL && buf != NULL);
-	sc_log(card->ctx, "called; %d bytes at index %d", count, idx);
+	if (card == NULL || card->ops == NULL || buf == NULL) {
+		return SC_ERROR_INVALID_ARGUMENTS;
+	}
+	sc_log(card->ctx, "called; %"SC_FORMAT_LEN_SIZE_T"u bytes at index %d",
+	       count, idx);
 	if (count == 0)
-		LOG_FUNC_RETURN(card->ctx, 0);
+		LOG_FUNC_RETURN(card->ctx, SC_SUCCESS);
+
 	if (card->ops->write_binary == NULL)
 		LOG_FUNC_RETURN(card->ctx, SC_ERROR_NOT_SUPPORTED);
 
-	if (count > max_lc) {
-		int bytes_written = 0;
-		const u8 *p = buf;
+	/* lock the card now to avoid deselection of the file */
+	r = sc_lock(card);
+	LOG_TEST_RET(card->ctx, r, "sc_lock() failed");
 
-		r = sc_lock(card);
-		LOG_TEST_RET(card->ctx, r, "sc_lock() failed");
-		while (count > 0) {
-			size_t n = count > max_lc? max_lc : count;
-			r = sc_write_binary(card, idx, p, n, flags);
-			if (r < 0) {
-				sc_unlock(card);
-				LOG_TEST_RET(card->ctx, r, "sc_write_binary() failed");
-			}
-			p += r;
-			idx += r;
-			bytes_written += r;
-			count -= r;
-			if (r == 0) {
-				sc_unlock(card);
-				LOG_FUNC_RETURN(card->ctx, bytes_written);
-			}
+	while (todo > 0) {
+		size_t chunk = todo > max_lc ? max_lc : todo;
+
+		r = card->ops->write_binary(card, idx, buf, chunk, flags);
+		if (r == 0 || r == SC_ERROR_FILE_END_REACHED)
+			break;
+		if ((idx > SIZE_MAX - (size_t) r)
+				|| (size_t) r > todo) {
+			/* `idx + r` or `todo - r` would overflow */
+			r = SC_ERROR_OFFSET_TOO_LARGE;
 		}
-		sc_unlock(card);
-		LOG_FUNC_RETURN(card->ctx, bytes_written);
+		if (r < 0) {
+			sc_unlock(card);
+			LOG_FUNC_RETURN(card->ctx, r);
+		}
+
+		todo -= (size_t) r;
+		buf  += (size_t) r;
+		idx  += (size_t) r;
 	}
 
-	r = card->ops->write_binary(card, idx, buf, count, flags);
-	LOG_FUNC_RETURN(card->ctx, r);
+	sc_unlock(card);
+
+	LOG_FUNC_RETURN(card->ctx, count - todo);
 }
 
 int sc_update_binary(sc_card_t *card, unsigned int idx,
 		     const u8 *buf, size_t count, unsigned long flags)
 {
 	size_t max_lc = sc_get_max_send_size(card);
+	size_t todo = count;
 	int r;
 
-	assert(card != NULL && card->ops != NULL && buf != NULL);
-	sc_log(card->ctx, "called; %d bytes at index %d", count, idx);
+	if (card == NULL || card->ops == NULL || buf == NULL) {
+		return SC_ERROR_INVALID_ARGUMENTS;
+	}
+	sc_log(card->ctx, "called; %"SC_FORMAT_LEN_SIZE_T"u bytes at index %d",
+	       count, idx);
 	if (count == 0)
-		return 0;
+		LOG_FUNC_RETURN(card->ctx, SC_SUCCESS);
 
 #ifdef ENABLE_SM
 	if (card->sm_ctx.ops.update_binary)   {
@@ -632,49 +746,79 @@ int sc_update_binary(sc_card_t *card, unsigned int idx,
 	if (card->ops->update_binary == NULL)
 		LOG_FUNC_RETURN(card->ctx, SC_ERROR_NOT_SUPPORTED);
 
-	if (count > max_lc) {
-		int bytes_written = 0;
-		const u8 *p = buf;
+	/* lock the card now to avoid deselection of the file */
+	r = sc_lock(card);
+	LOG_TEST_RET(card->ctx, r, "sc_lock() failed");
 
-		r = sc_lock(card);
-		LOG_TEST_RET(card->ctx, r, "sc_lock() failed");
-		while (count > 0) {
-			size_t n = count > max_lc? max_lc : count;
-			r = sc_update_binary(card, idx, p, n, flags);
-			if (r < 0) {
-				sc_unlock(card);
-				LOG_TEST_RET(card->ctx, r, "sc_update_binary() failed");
-			}
-			p += r;
-			idx += r;
-			bytes_written += r;
-			count -= r;
-			if (r == 0) {
-				sc_unlock(card);
-				LOG_FUNC_RETURN(card->ctx, bytes_written);
-			}
+	while (todo > 0) {
+		size_t chunk = todo > max_lc ? max_lc : todo;
+
+		r = card->ops->update_binary(card, idx, buf, chunk, flags);
+		if (r == 0 || r == SC_ERROR_FILE_END_REACHED)
+			break;
+		if ((idx > SIZE_MAX - (size_t) r)
+				|| (size_t) r > todo) {
+			/* `idx + r` or `todo - r` would overflow */
+			r = SC_ERROR_OFFSET_TOO_LARGE;
 		}
-		sc_unlock(card);
-		LOG_FUNC_RETURN(card->ctx, bytes_written);
+		if (r < 0) {
+			sc_unlock(card);
+			LOG_FUNC_RETURN(card->ctx, r);
+		}
+
+		todo -= (size_t) r;
+		buf  += (size_t) r;
+		idx  += (size_t) r;
 	}
 
-	r = card->ops->update_binary(card, idx, buf, count, flags);
-	LOG_FUNC_RETURN(card->ctx, r);
+	sc_unlock(card);
+
+	LOG_FUNC_RETURN(card->ctx, count - todo);
 }
 
 
-int sc_erase_binary(struct sc_card *card, unsigned int offs, size_t count,  unsigned long flags)
+int sc_erase_binary(struct sc_card *card, unsigned int idx, size_t count,  unsigned long flags)
 {
 	int r;
+	size_t todo = count;
 
-	assert(card != NULL && card->ops != NULL);
-	sc_log(card->ctx, "called; erase %d bytes from offset %d", count, offs);
+	if (card == NULL || card->ops == NULL) {
+		return SC_ERROR_INVALID_ARGUMENTS;
+	}
+	sc_log(card->ctx,
+	       "called; erase %"SC_FORMAT_LEN_SIZE_T"u bytes from offset %d",
+	       count, idx);
+	if (count == 0)
+		LOG_FUNC_RETURN(card->ctx, SC_SUCCESS);
 
 	if (card->ops->erase_binary == NULL)
 		LOG_FUNC_RETURN(card->ctx, SC_ERROR_NOT_SUPPORTED);
 
-	r = card->ops->erase_binary(card, offs, count, flags);
-	LOG_FUNC_RETURN(card->ctx, r);
+	/* lock the card now to avoid deselection of the file */
+	r = sc_lock(card);
+	LOG_TEST_RET(card->ctx, r, "sc_lock() failed");
+
+	while (todo > 0) {
+		r = card->ops->erase_binary(card, idx, todo, flags);
+		if (r == 0 || r == SC_ERROR_FILE_END_REACHED)
+			break;
+		if ((idx > SIZE_MAX - (size_t) r)
+				|| (size_t) r > todo) {
+			/* `idx + r` or `todo - r` would overflow */
+			r = SC_ERROR_OFFSET_TOO_LARGE;
+		}
+		if (r < 0) {
+			sc_unlock(card);
+			LOG_FUNC_RETURN(card->ctx, r);
+		}
+
+		todo -= (size_t) r;
+		idx  += (size_t) r;
+	}
+
+	sc_unlock(card);
+
+	LOG_FUNC_RETURN(card->ctx, count - todo);
 }
 
 
@@ -683,7 +827,9 @@ int sc_select_file(sc_card_t *card, const sc_path_t *in_path,  sc_file_t **file)
 	int r;
 	char pbuf[SC_MAX_PATH_STRING_SIZE];
 
-	assert(card != NULL && in_path != NULL);
+	if (card == NULL || in_path == NULL) {
+		return SC_ERROR_INVALID_ARGUMENTS;
+	}
 
 	r = sc_path_print(pbuf, sizeof(pbuf), in_path);
 	if (r != SC_SUCCESS)
@@ -763,14 +909,38 @@ int sc_get_challenge(sc_card_t *card, u8 *rnd, size_t len)
 {
 	int r;
 
-	assert(card != NULL);
+	if (len == 0)
+		return SC_SUCCESS;
+
+	if (card == NULL || rnd == NULL) {
+		return SC_ERROR_INVALID_ARGUMENTS;
+	}
+
 	LOG_FUNC_CALLED(card->ctx);
 
-	if (card->ops->get_challenge == NULL)
+	if (card->ops == NULL || card->ops->get_challenge == NULL)
 		LOG_FUNC_RETURN(card->ctx, SC_ERROR_NOT_SUPPORTED);
-	r = card->ops->get_challenge(card, rnd, len);
 
-	LOG_FUNC_RETURN(card->ctx, r);
+	r = sc_lock(card);
+	if (r != SC_SUCCESS)
+		LOG_FUNC_RETURN(card->ctx, r);
+
+	while (len > 0) {
+		r = card->ops->get_challenge(card, rnd, len);
+		if (r == 0)
+			r = SC_ERROR_INVALID_DATA;
+		if (r < 0) {
+			sc_unlock(card);
+			LOG_FUNC_RETURN(card->ctx, r);
+		}
+
+		rnd += (size_t) r;
+		len -= (size_t) r;
+	}
+
+	sc_unlock(card);
+
+	LOG_FUNC_RETURN(card->ctx, SC_SUCCESS);
 }
 
 int sc_read_record(sc_card_t *card, unsigned int rec_nr, u8 *buf,
@@ -778,12 +948,18 @@ int sc_read_record(sc_card_t *card, unsigned int rec_nr, u8 *buf,
 {
 	int r;
 
-	assert(card != NULL);
+	if (card == NULL) {
+		return SC_ERROR_INVALID_ARGUMENTS;
+	}
 	LOG_FUNC_CALLED(card->ctx);
 
 	if (card->ops->read_record == NULL)
 		LOG_FUNC_RETURN(card->ctx, SC_ERROR_NOT_SUPPORTED);
+
 	r = card->ops->read_record(card, rec_nr, buf, count, flags);
+	if (r == SC_SUCCESS) {
+		r = count;
+	}
 
 	LOG_FUNC_RETURN(card->ctx, r);
 }
@@ -793,12 +969,18 @@ int sc_write_record(sc_card_t *card, unsigned int rec_nr, const u8 * buf,
 {
 	int r;
 
-	assert(card != NULL);
+	if (card == NULL) {
+		return SC_ERROR_INVALID_ARGUMENTS;
+	}
 	LOG_FUNC_CALLED(card->ctx);
 
 	if (card->ops->write_record == NULL)
 		LOG_FUNC_RETURN(card->ctx, SC_ERROR_NOT_SUPPORTED);
+
 	r = card->ops->write_record(card, rec_nr, buf, count, flags);
+	if (r == SC_SUCCESS) {
+		r = count;
+	}
 
 	LOG_FUNC_RETURN(card->ctx, r);
 }
@@ -808,12 +990,18 @@ int sc_append_record(sc_card_t *card, const u8 * buf, size_t count,
 {
 	int r;
 
-	assert(card != NULL);
+	if (card == NULL) {
+		return SC_ERROR_INVALID_ARGUMENTS;
+	}
 	LOG_FUNC_CALLED(card->ctx);
 
 	if (card->ops->append_record == NULL)
 		LOG_FUNC_RETURN(card->ctx, SC_ERROR_NOT_SUPPORTED);
+
 	r = card->ops->append_record(card, buf, count, flags);
+	if (r == SC_SUCCESS) {
+		r = count;
+	}
 
 	LOG_FUNC_RETURN(card->ctx, r);
 }
@@ -823,12 +1011,18 @@ int sc_update_record(sc_card_t *card, unsigned int rec_nr, const u8 * buf,
 {
 	int r;
 
-	assert(card != NULL);
+	if (card == NULL) {
+		return SC_ERROR_INVALID_ARGUMENTS;
+	}
 	LOG_FUNC_CALLED(card->ctx);
 
 	if (card->ops->update_record == NULL)
 		LOG_FUNC_RETURN(card->ctx, SC_ERROR_NOT_SUPPORTED);
+
 	r = card->ops->update_record(card, rec_nr, buf, count, flags);
+	if (r == SC_SUCCESS) {
+		r = count;
+	}
 
 	LOG_FUNC_RETURN(card->ctx, r);
 }
@@ -837,11 +1031,14 @@ int sc_delete_record(sc_card_t *card, unsigned int rec_nr)
 {
 	int r;
 
-	assert(card != NULL);
+	if (card == NULL) {
+		return SC_ERROR_INVALID_ARGUMENTS;
+	}
 	LOG_FUNC_CALLED(card->ctx);
 
 	if (card->ops->delete_record == NULL)
 		LOG_FUNC_RETURN(card->ctx, SC_ERROR_NOT_SUPPORTED);
+
 	r = card->ops->delete_record(card, rec_nr);
 
 	LOG_FUNC_RETURN(card->ctx, r);
@@ -852,7 +1049,9 @@ sc_card_ctl(sc_card_t *card, unsigned long cmd, void *args)
 {
 	int r = SC_ERROR_NOT_SUPPORTED;
 
-	assert(card != NULL);
+	if (card == NULL) {
+		return SC_ERROR_INVALID_ARGUMENTS;
+	}
 	LOG_FUNC_CALLED(card->ctx);
 
 	if (card->ops->card_ctl != NULL)
@@ -870,13 +1069,11 @@ int _sc_card_add_algorithm(sc_card_t *card, const sc_algorithm_info_t *info)
 {
 	sc_algorithm_info_t *p;
 
-	assert(info != NULL);
+	if (info == NULL) {
+		return SC_ERROR_INVALID_ARGUMENTS;
+	}
 	p = (sc_algorithm_info_t *) realloc(card->algorithms, (card->algorithm_count + 1) * sizeof(*info));
 	if (!p) {
-		if (card->algorithms)
-			free(card->algorithms);
-		card->algorithms = NULL;
-		card->algorithm_count = 0;
 		return SC_ERROR_OUT_OF_MEMORY;
 	}
 	card->algorithms = p;
@@ -886,6 +1083,19 @@ int _sc_card_add_algorithm(sc_card_t *card, const sc_algorithm_info_t *info)
 	return SC_SUCCESS;
 }
 
+int _sc_card_add_symmetric_alg(sc_card_t *card, unsigned int algorithm,
+			       unsigned int key_length, unsigned long flags)
+{
+	sc_algorithm_info_t info;
+
+	memset(&info, 0, sizeof(info));
+	info.algorithm = algorithm;
+	info.key_length = key_length;
+	info.flags = flags;
+
+	return _sc_card_add_algorithm(card, &info);
+}
+
 int  _sc_card_add_ec_alg(sc_card_t *card, unsigned int key_length,
 			unsigned long flags, unsigned long ext_flags,
 			struct sc_object_id *curve_oid)
@@ -906,7 +1116,7 @@ int  _sc_card_add_ec_alg(sc_card_t *card, unsigned int key_length,
 	return _sc_card_add_algorithm(card, &info);
 }
 
-static sc_algorithm_info_t * sc_card_find_alg(sc_card_t *card,
+sc_algorithm_info_t * sc_card_find_alg(sc_card_t *card,
 		unsigned int algorithm, unsigned int key_length, void *param)
 {
 	int i;
@@ -960,7 +1170,7 @@ sc_algorithm_info_t * sc_card_find_gostr3410_alg(sc_card_t *card,
 	return sc_card_find_alg(card, SC_ALGORITHM_GOSTR3410, key_length, NULL);
 }
 
-static int match_atr_table(sc_context_t *ctx, struct sc_atr_table *table, struct sc_atr *atr)
+static int match_atr_table(sc_context_t *ctx, const struct sc_atr_table *table, struct sc_atr *atr)
 {
 	u8 *card_atr_bin;
 	size_t card_atr_bin_len;
@@ -975,7 +1185,7 @@ static int match_atr_table(sc_context_t *ctx, struct sc_atr_table *table, struct
 	sc_bin_to_hex(card_atr_bin, card_atr_bin_len, card_atr_hex, sizeof(card_atr_hex), ':');
 	card_atr_hex_len = strlen(card_atr_hex);
 
-	sc_log(ctx, "ATR     : %s", card_atr_hex);
+	sc_debug(ctx, SC_LOG_DEBUG_MATCH, "ATR     : %s", card_atr_hex);
 
 	for (i = 0; table[i].atr != NULL; i++) {
 		const char *tatr = table[i].atr;
@@ -986,14 +1196,14 @@ static int match_atr_table(sc_context_t *ctx, struct sc_atr_table *table, struct
 		size_t fix_hex_len = card_atr_hex_len;
 		size_t fix_bin_len = card_atr_bin_len;
 
-		sc_log(ctx, "ATR try : %s", tatr);
+		sc_debug(ctx, SC_LOG_DEBUG_MATCH, "ATR try : %s", tatr);
 
 		if (tatr_len != fix_hex_len) {
-			sc_log(ctx, "ignored - wrong length");
+			sc_debug(ctx, SC_LOG_DEBUG_MATCH, "ignored - wrong length");
 			continue;
 		}
 		if (matr != NULL) {
-			sc_log(ctx, "ATR mask: %s", matr);
+			sc_debug(ctx, SC_LOG_DEBUG_MATCH, "ATR mask: %s", matr);
 
 			matr_len = strlen(matr);
 			if (tatr_len != matr_len)
@@ -1003,7 +1213,7 @@ static int match_atr_table(sc_context_t *ctx, struct sc_atr_table *table, struct
 			mbin_len = sizeof(mbin);
 			sc_hex_to_bin(matr, mbin, &mbin_len);
 			if (mbin_len != fix_bin_len) {
-				sc_log(ctx, "length of atr and atr mask do not match - ignored: %s - %s", tatr, matr);
+				sc_debug(ctx, SC_LOG_DEBUG_MATCH, "length of atr and atr mask do not match - ignored: %s - %s", tatr, matr);
 				continue;
 			}
 			for (s = 0; s < tbin_len; s++) {
@@ -1023,7 +1233,7 @@ static int match_atr_table(sc_context_t *ctx, struct sc_atr_table *table, struct
 	return -1;
 }
 
-int _sc_match_atr(sc_card_t *card, struct sc_atr_table *table, int *type_out)
+int _sc_match_atr(sc_card_t *card, const struct sc_atr_table *table, int *type_out)
 {
 	int res;
 
@@ -1154,10 +1364,20 @@ scconf_block *sc_get_conf_block(sc_context_t *ctx, const char *name1, const char
 	return conf_block;
 }
 
-void sc_print_cache(struct sc_card *card)   {
+void sc_invalidate_cache(struct sc_card *card)
+{
+	if (card) {
+		memset(&card->cache, 0, sizeof(card->cache));
+		card->cache.valid = 0;
+	}
+}
+
+void sc_print_cache(struct sc_card *card)
+{
 	struct sc_context *ctx = NULL;
 
-	assert(card != NULL);
+	if (card == NULL)
+		return;
 	ctx = card->ctx;
 
 	if (!card->cache.valid || (!card->cache.current_ef && !card->cache.current_df))   {
@@ -1170,9 +1390,11 @@ void sc_print_cache(struct sc_card *card)   {
 				sc_print_path(&card->cache.current_ef->path));
 
 	if (card->cache.current_df)
-		sc_log(ctx, "current_df(type=%i, aid_len=%i) %s", card->cache.current_df->path.type,
-				card->cache.current_df->path.aid.len,
-				sc_print_path(&card->cache.current_df->path));
+		sc_log(ctx,
+		       "current_df(type=%i, aid_len=%"SC_FORMAT_LEN_SIZE_T"u) %s",
+		       card->cache.current_df->path.type,
+		       card->cache.current_df->path.aid.len,
+		       sc_print_path(&card->cache.current_df->path));
 }
 
 int sc_copy_ec_params(struct sc_ec_parameters *dst, struct sc_ec_parameters *src)
@@ -1228,29 +1450,39 @@ sc_card_sm_load(struct sc_card *card, const char *module_path, const char *in_mo
 	char *module = NULL;
 #ifdef _WIN32
 	char temp_path[PATH_MAX];
-	int temp_len;
+	size_t temp_len;
 	const char path_delim = '\\';
+	char expanded_val[PATH_MAX];
+	DWORD expanded_len;
 #else
 	const char path_delim = '/';
 #endif
 
-	assert(card != NULL);
+	if (card == NULL) {
+		return SC_ERROR_INVALID_ARGUMENTS;
+	}
 	ctx = card->ctx;
-	SC_FUNC_CALLED(ctx, SC_LOG_DEBUG_NORMAL);
+	LOG_FUNC_CALLED(ctx);
 	if (!in_module)
 		return sc_card_sm_unload(card);
 
 #ifdef _WIN32
-	if (!module_path)   {
-		temp_len = PATH_MAX;
+	if (!module_path || strlen(module_path) == 0)   {
+		temp_len = PATH_MAX-1;
 		rv = sc_ctx_win32_get_config_value(NULL, "SmDir", "Software\\OpenSC Project\\OpenSC",
 				temp_path, &temp_len);
-		if (rv == SC_SUCCESS)
+		if (rv == SC_SUCCESS) {
+			temp_path[temp_len] = '\0';
 			module_path = temp_path;
+		}
 	}
+	expanded_len = PATH_MAX;
+	expanded_len = ExpandEnvironmentStringsA(module_path, expanded_val, expanded_len);
+	if (0 < expanded_len && expanded_len < sizeof expanded_val)
+		module_path = expanded_val;
 #endif
-	sc_debug(ctx, SC_LOG_DEBUG_NORMAL, "SM module '%s' located in '%s'", in_module, module_path);
-	if (module_path)   {
+	sc_log(ctx, "SM module '%s' located in '%s'", in_module, module_path);
+	if (module_path && strlen(module_path) > 0)   {
 		int sz = strlen(in_module) + strlen(module_path) + 3;
 		module = malloc(sz);
 		if (module)
@@ -1327,8 +1559,7 @@ sc_card_sm_check(struct sc_card *card)
 	scconf_block *atrblock = NULL, *sm_conf_block = NULL;
 	int rv, ii;
 
-	SC_FUNC_CALLED(ctx, SC_LOG_DEBUG_NORMAL);
-	sc_log(ctx, "card->sm_ctx.ops.open %p", card->sm_ctx.ops.open);
+	LOG_FUNC_CALLED(ctx);
 
 	/* get the name of card specific SM configuration section */
 	atrblock = _sc_match_atr_block(ctx, card->driver, &card->atr);
@@ -1343,12 +1574,12 @@ sc_card_sm_check(struct sc_card *card)
         for (ii = 0; ctx->conf_blocks[ii]; ii++) {
 		scconf_block **blocks;
 
-                blocks = scconf_find_blocks(ctx->conf, ctx->conf_blocks[ii], "secure_messaging", sm);
+		blocks = scconf_find_blocks(ctx->conf, ctx->conf_blocks[ii], "secure_messaging", sm);
 		if (blocks) {
 			sm_conf_block = blocks[0];
 			free(blocks);
 		}
-                if (sm_conf_block != NULL)
+		if (sm_conf_block != NULL)
 			break;
 	}
 
@@ -1356,8 +1587,8 @@ sc_card_sm_check(struct sc_card *card)
 		LOG_TEST_RET(ctx, SC_ERROR_INCONSISTENT_CONFIGURATION, "SM configuration block not preset");
 
 	/* check if an external SM module has to be used */
-	module_path = scconf_get_str(sm_conf_block, "module_path", NULL);
-	module_name = scconf_get_str(sm_conf_block, "module_name", NULL);
+	module_path = scconf_get_str(sm_conf_block, "module_path", DEFAULT_SM_MODULE_PATH);
+	module_name = scconf_get_str(sm_conf_block, "module_name", DEFAULT_SM_MODULE);
 	sc_log(ctx, "SM module '%s' in  '%s'", module_name, module_path);
 	if (!module_name)
 		LOG_TEST_RET(ctx, SC_ERROR_INCONSISTENT_CONFIGURATION, "Invalid SM configuration: module not defined");
@@ -1369,18 +1600,15 @@ sc_card_sm_check(struct sc_card *card)
 	strlcpy(card->sm_ctx.config_section, sm, sizeof(card->sm_ctx.config_section));
 
 	/* allocate resources for the external SM module */
-	sc_log(ctx, "'module_init' handler %p", card->sm_ctx.module.ops.module_init);
 	if (card->sm_ctx.module.ops.module_init)   {
 		module_data = scconf_get_str(sm_conf_block, "module_data", NULL);
-		sc_log(ctx, "module_data '%s'", module_data);
 
 		rv = card->sm_ctx.module.ops.module_init(ctx, module_data);
-		SC_TEST_RET(ctx, SC_LOG_DEBUG_NORMAL, rv, "Cannot initialize SM module");
+		LOG_TEST_RET(ctx, rv, "Cannot initialize SM module");
 	}
 
 	/* initialize SM session in the case of 'APDU TRANSMIT' SM mode */
 	sm_mode = scconf_get_str(sm_conf_block, "mode", NULL);
-	sc_log(ctx, "SM mode '%s'; 'open' handler %p", sm_mode, card->sm_ctx.ops.open);
 	if (sm_mode && !strcasecmp("Transmit", sm_mode))   {
 		if (!card->sm_ctx.ops.open || !card->sm_ctx.ops.get_sm_apdu || !card->sm_ctx.ops.free_sm_apdu)
 			LOG_TEST_RET(ctx, SC_ERROR_NOT_SUPPORTED, "'Transmit' SM asked but not supported by card driver");
