@@ -65,12 +65,13 @@ parse_dir_record(sc_card_t *card, u8 ** buf, size_t *buflen, int rec_nr)
 	sc_app_info_t *app = NULL;
 	struct sc_aid aid;
 	u8 label[128], path[128], ddo[128];
-	size_t label_len = sizeof(label), path_len = sizeof(path), ddo_len = sizeof(ddo);
+	size_t label_len = sizeof(label) - 1, path_len = sizeof(path), ddo_len = sizeof(ddo);
 	int r;
 
 	LOG_FUNC_CALLED(ctx);
 	aid.len = sizeof(aid.value);
 
+	memset(label, 0, sizeof(label));
 	sc_copy_asn1_entry(c_asn1_dirrecord, asn1_dirrecord);
 	sc_copy_asn1_entry(c_asn1_dir, asn1_dir);
 	sc_format_asn1_entry(asn1_dir + 0, asn1_dirrecord, NULL, 0);
@@ -117,6 +118,7 @@ parse_dir_record(sc_card_t *card, u8 ** buf, size_t *buflen, int rec_nr)
 	if (asn1_dirrecord[2].flags & SC_ASN1_PRESENT && path_len > 0) {
 		/* application path present: ignore AID */
 		if (path_len > SC_MAX_PATH_SIZE) {
+			free(app->label);
 			free(app);
 			LOG_TEST_RET(ctx, SC_ERROR_INVALID_ASN1_OBJECT, "Application path is too long.");
 		}
@@ -134,6 +136,7 @@ parse_dir_record(sc_card_t *card, u8 ** buf, size_t *buflen, int rec_nr)
 	if (asn1_dirrecord[3].flags & SC_ASN1_PRESENT) {
 		app->ddo.value = malloc(ddo_len);
 		if (app->ddo.value == NULL) {
+			free(app->label);
 			free(app);
 			LOG_TEST_RET(ctx, SC_ERROR_OUT_OF_MEMORY, "Cannot allocate DDO value");
 		}
@@ -159,31 +162,35 @@ int sc_enum_apps(sc_card_t *card)
 	int ef_structure;
 	size_t file_size, jj;
 	int r, ii, idx;
+	struct sc_file *ef_dir = NULL;
 
 	LOG_FUNC_CALLED(ctx);
-	if (card->app_count < 0)
-		card->app_count = 0;
+
+	sc_free_apps(card);
+	card->app_count = 0;
 
 	sc_format_path("3F002F00", &path);
-	sc_file_free(card->ef_dir);
-	card->ef_dir = NULL;
-	r = sc_select_file(card, &path, &card->ef_dir);
+	r = sc_select_file(card, &path, &ef_dir);
+	if (r < 0)
+		sc_file_free(ef_dir);
 	LOG_TEST_RET(ctx, r, "Cannot select EF.DIR file");
 
-	if (card->ef_dir->type != SC_FILE_TYPE_WORKING_EF) {
-		sc_file_free(card->ef_dir);
-		card->ef_dir = NULL;
+	if (ef_dir->type != SC_FILE_TYPE_WORKING_EF) {
+		sc_file_free(ef_dir);
 		LOG_TEST_RET(ctx, SC_ERROR_INVALID_CARD, "EF(DIR) is not a working EF.");
 	}
 
-	ef_structure = card->ef_dir->ef_structure;
+	ef_structure = ef_dir->ef_structure;
+	file_size = ef_dir->size;
+	sc_file_free(ef_dir);
 	if (ef_structure == SC_FILE_EF_TRANSPARENT) {
 		u8 *buf = NULL, *p;
 		size_t bufsize;
 
-		file_size = card->ef_dir->size;
 		if (file_size == 0)
 			LOG_FUNC_RETURN(ctx, 0);
+		if (file_size > MAX_FILE_SIZE)
+			LOG_FUNC_RETURN(ctx, SC_ERROR_INVALID_DATA);
 
 		buf = malloc(file_size);
 		if (buf == NULL)
@@ -214,7 +221,7 @@ int sc_enum_apps(sc_card_t *card)
 		size_t rec_size;
 
 		/* Arbitrary set '16' as maximal number of records to check out:
-		 * to avoid endless loop because of some uncomplete cards/drivers */
+		 * to avoid endless loop because of some incomplete cards/drivers */
 		for (rec_nr = 1; rec_nr < 16; rec_nr++) {
 			r = sc_read_record(card, rec_nr, buf, sizeof(buf), SC_RECORD_BY_REC_NR);
 			if (r == SC_ERROR_RECORD_NOT_FOUND)
@@ -259,10 +266,8 @@ void sc_free_apps(sc_card_t *card)
 	int	i;
 
 	for (i = 0; i < card->app_count; i++) {
-		if (card->app[i]->label)
-			free(card->app[i]->label);
-		if (card->app[i]->ddo.value)
-			free(card->app[i]->ddo.value);
+		free(card->app[i]->label);
+		free(card->app[i]->ddo.value);
 		free(card->app[i]);
 	}
 	card->app_count = -1;
@@ -309,6 +314,8 @@ static int update_transparent(sc_card_t *card, sc_file_t *file)
 				free(buf);
 			return r;
 		}
+		if (!rec_size)
+			continue;
 		tmp = (u8 *) realloc(buf, buf_size + rec_size);
 		if (!tmp) {
 			if (rec)
