@@ -83,7 +83,7 @@ iasecc_sdo_convert_acl(struct sc_card *card, struct iasecc_sdo *sdo,
 		{SC_AC_OP_READ,		IASECC_ACL_GET_DATA},
 		{0x00, 0x00}
 	};
-	unsigned char mask = 0x80, op_mask;
+	unsigned char mask = 0x80, op_mask = 0;
 	int ii;
 
 	LOG_FUNC_CALLED(ctx);
@@ -94,10 +94,10 @@ iasecc_sdo_convert_acl(struct sc_card *card, struct iasecc_sdo *sdo,
 			break;
 		}
 	}
-	if (ops[ii].mask == 0)
+	if (op_mask == 0)
 		LOG_FUNC_RETURN(ctx, SC_ERROR_UNKNOWN_DATA_RECEIVED);
 
-	sc_log(ctx, "OP:%i, mask:0x%X", op, ops[ii].mask);
+	sc_log(ctx, "OP:%i, mask:0x%X", op, op_mask);
 	sc_log(ctx, "AMB:%X, scbs:%s", sdo->docp.amb, sc_dump_hex(sdo->docp.scbs, IASECC_MAX_SCBS));
 	sc_log(ctx, "docp.acls_contact:%s", sc_dump_hex(sdo->docp.acls_contact.value, sdo->docp.acls_contact.size));
 
@@ -344,7 +344,7 @@ iasecc_se_parse(struct sc_card *card, unsigned char *data, size_t data_len, stru
 		LOG_TEST_RET(ctx, size_size, "parse error: invalid SDO SE data size");
 
 		if (data_len != size + size_size + 3)
-			LOG_TEST_RET(ctx, SC_ERROR_INVALID_DATA, "parse error: invalide SDO SE data size");
+			LOG_TEST_RET(ctx, SC_ERROR_INVALID_DATA, "parse error: invalid SDO SE data size");
 
 		data += 3 + size_size;
 		data_len = size;
@@ -364,7 +364,7 @@ iasecc_se_parse(struct sc_card *card, unsigned char *data, size_t data_len, stru
 	LOG_TEST_RET(ctx, size_size, "parse error: invalid size data");
 
 	if (data_len != size + size_size + 1)
-		LOG_TEST_RET(ctx, SC_ERROR_INVALID_DATA, "parse error: invalide SE data size");
+		LOG_TEST_RET(ctx, SC_ERROR_INVALID_DATA, "parse error: invalid SE data size");
 
 	offs = 1 + size_size;
 	for (; offs < data_len;)   {
@@ -375,7 +375,7 @@ iasecc_se_parse(struct sc_card *card, unsigned char *data, size_t data_len, stru
 	}
 
 	if (offs != data_len)
-		LOG_TEST_RET(ctx, SC_ERROR_INVALID_DATA, "parse error: not totaly parsed");
+		LOG_TEST_RET(ctx, SC_ERROR_INVALID_DATA, "parse error: not totally parsed");
 
 	LOG_FUNC_RETURN(ctx, SC_SUCCESS);
 }
@@ -558,8 +558,10 @@ iasecc_parse_keyset(struct sc_card *card, unsigned char *data, size_t data_len,
 
 		if (tlv.tag == IASECC_SDO_KEYSET_TAG_COMPULSORY)
 			keyset->compulsory = tlv;
-		else
+		else {
+			free(tlv.value);
 			LOG_TEST_RET(ctx, SC_ERROR_UNKNOWN_DATA_RECEIVED, "parse error: non KeySet SDO tag");
+		}
 
 		offs += rv;
 	}
@@ -583,7 +585,7 @@ iasecc_parse_docp(struct sc_card *card, unsigned char *data, size_t data_len, st
 		LOG_TEST_RET(ctx, rv, "iasecc_parse_get_tlv() get and parse TLV error");
 
 		sc_log(ctx,
-		       "iasecc_parse_docp() parse_get_tlv retuned %i; tag %X; size %"SC_FORMAT_LEN_SIZE_T"u",
+		       "iasecc_parse_docp() parse_get_tlv returned %i; tag %X; size %"SC_FORMAT_LEN_SIZE_T"u",
 		       rv, tlv.tag, tlv.size);
 
 		if (tlv.tag == IASECC_DOCP_TAG_ACLS)   {
@@ -606,7 +608,7 @@ iasecc_parse_docp(struct sc_card *card, unsigned char *data, size_t data_len, st
 		else if (tlv.tag == IASECC_DOCP_TAG_ISSUER_DATA)   {
 			sdo->docp.issuer_data = tlv;
 		}
-		else if (tlv.tag == IASECC_DOCP_TAG_NON_REPUDATION)   {
+		else if (tlv.tag == IASECC_DOCP_TAG_NON_REPUDIATION)   {
 			sdo->docp.non_repudiation = tlv;
 		}
 		else if (tlv.tag == IASECC_DOCP_TAG_USAGE_REMAINING)   {
@@ -619,6 +621,7 @@ iasecc_parse_docp(struct sc_card *card, unsigned char *data, size_t data_len, st
 			sdo->docp.tries_remaining = tlv;
 		}
 		else   {
+			free(tlv.value);
 			LOG_TEST_RET(ctx, SC_ERROR_UNKNOWN_DATA_RECEIVED, "iasecc_parse_get_tlv() parse error: non DOCP tag");
 		}
 
@@ -655,7 +658,7 @@ iasecc_sdo_parse_data(struct sc_card *card, unsigned char *data, struct iasecc_s
 		free(tlv.value);
 		LOG_TEST_RET(ctx, rv, "parse error: cannot parse DOCP");
 	}
-	else if (tlv.tag == IASECC_DOCP_TAG_NON_REPUDATION)   {
+	else if (tlv.tag == IASECC_DOCP_TAG_NON_REPUDIATION)   {
 		sdo->docp.non_repudiation = tlv;
 	}
 	else if (tlv.tag == IASECC_DOCP_TAG_USAGE_REMAINING)   {
@@ -668,43 +671,48 @@ iasecc_sdo_parse_data(struct sc_card *card, unsigned char *data, struct iasecc_s
 		sdo->docp.tries_remaining = tlv;
 	}
 	else if (tlv.tag == IASECC_SDO_CHV_TAG)   {
-		if (sdo->sdo_class != IASECC_SDO_CLASS_CHV)
+		if (sdo->sdo_class != IASECC_SDO_CLASS_CHV) {
+			free(tlv.value);
 			LOG_TEST_RET(ctx, SC_ERROR_INVALID_DATA, "parse error: IASECC_SDO_CHV_TAG tag in non User CHV SDO");
+		}
 
 		rv = iasecc_parse_chv(card, tlv.value, tlv.size, &sdo->data.chv);
-		LOG_TEST_RET(ctx, rv, "parse error: cannot parse SDO CHV data");
-
 		free(tlv.value);
+		LOG_TEST_RET(ctx, rv, "parse error: cannot parse SDO CHV data");
 	}
 	else if (tlv.tag == IASECC_SDO_PUBKEY_TAG)   {
-		if (sdo->sdo_class != IASECC_SDO_CLASS_RSA_PUBLIC)
+		if (sdo->sdo_class != IASECC_SDO_CLASS_RSA_PUBLIC) {
+			free(tlv.value);
 			LOG_TEST_RET(ctx, SC_ERROR_INVALID_DATA, "parse error: SDO_PUBLIC_KEY tag in non PUBLIC_KEY SDO");
+		}
 
 		rv = iasecc_parse_pubkey(card, tlv.value, tlv.size, &sdo->data.pub_key);
-		LOG_TEST_RET(ctx, rv, "parse error: cannot parse SDO PUBLIC KEY data");
-
 		free(tlv.value);
+		LOG_TEST_RET(ctx, rv, "parse error: cannot parse SDO PUBLIC KEY data");
 	}
 	else if (tlv.tag == IASECC_SDO_PRVKEY_TAG)   {
-		if (sdo->sdo_class != IASECC_SDO_CLASS_RSA_PRIVATE)
+		if (sdo->sdo_class != IASECC_SDO_CLASS_RSA_PRIVATE) {
+			free(tlv.value);
 			LOG_TEST_RET(ctx, SC_ERROR_INVALID_DATA, "parse error: SDO_PRIVATE_KEY tag in non PRIVATE_KEY SDO");
+		}
 
 		rv = iasecc_parse_prvkey(card, tlv.value, tlv.size, &sdo->data.prv_key);
-		LOG_TEST_RET(ctx, rv, "parse error: cannot parse SDO PRIVATE KEY data");
-
 		free(tlv.value);
+		LOG_TEST_RET(ctx, rv, "parse error: cannot parse SDO PRIVATE KEY data");
 	}
 	else if (tlv.tag == IASECC_SDO_KEYSET_TAG)   {
-		if (sdo->sdo_class != IASECC_SDO_CLASS_KEYSET)
+		if (sdo->sdo_class != IASECC_SDO_CLASS_KEYSET) {
+			free(tlv.value);
 			LOG_TEST_RET(ctx, SC_ERROR_INVALID_DATA, "parse error: SDO_KEYSET tag in non KEYSET SDO");
+		}
 
 		rv = iasecc_parse_keyset(card, tlv.value, tlv.size, &sdo->data.keyset);
-		LOG_TEST_RET(ctx, rv, "parse error: cannot parse SDO KEYSET data");
-
 		free(tlv.value);
+		LOG_TEST_RET(ctx, rv, "parse error: cannot parse SDO KEYSET data");
 	}
 	else   {
 		sc_log(ctx, "iasecc_sdo_parse_data() non supported tag 0x%X", tlv.tag);
+		free(tlv.value);
 		LOG_FUNC_RETURN(ctx, SC_ERROR_NOT_SUPPORTED);
 	}
 
@@ -745,7 +753,7 @@ iasecc_sdo_parse(struct sc_card *card, unsigned char *data, size_t data_len, str
 	LOG_TEST_RET(ctx, size_size, "parse error: invalid size data");
 
 	if (data_len != size + size_size + 3)
-		LOG_TEST_RET(ctx, SC_ERROR_INVALID_DATA, "parse error: invalide SDO data size");
+		LOG_TEST_RET(ctx, SC_ERROR_INVALID_DATA, "parse error: invalid SDO data size");
 
 	sc_log(ctx,
 	       "sz %"SC_FORMAT_LEN_SIZE_T"u, sz_size %"SC_FORMAT_LEN_SIZE_T"u",
@@ -760,7 +768,7 @@ iasecc_sdo_parse(struct sc_card *card, unsigned char *data, size_t data_len, str
 	}
 
 	if (offs != data_len)
-		LOG_TEST_RET(ctx, SC_ERROR_INVALID_DATA, "parse error: not totaly parsed");
+		LOG_TEST_RET(ctx, SC_ERROR_INVALID_DATA, "parse error: not totally parsed");
 
 	sc_log(ctx,
 	       "docp.acls_contact.size %"SC_FORMAT_LEN_SIZE_T"u, docp.size.size %"SC_FORMAT_LEN_SIZE_T"u",
@@ -803,7 +811,7 @@ iasecc_sdo_allocate_and_parse(struct sc_card *card, unsigned char *data, size_t
 	LOG_TEST_RET(ctx, size_size, "parse error: invalid size data");
 
 	if (data_len != size + size_size + 3)
-		LOG_TEST_RET(ctx, SC_ERROR_INVALID_DATA, "parse error: invalide SDO data size");
+		LOG_TEST_RET(ctx, SC_ERROR_INVALID_DATA, "parse error: invalid SDO data size");
 
 	sc_log(ctx,
 	       "sz %"SC_FORMAT_LEN_SIZE_T"u, sz_size %"SC_FORMAT_LEN_SIZE_T"u",
@@ -818,7 +826,7 @@ iasecc_sdo_allocate_and_parse(struct sc_card *card, unsigned char *data, size_t
 	}
 
 	if (offs != data_len)
-		LOG_TEST_RET(ctx, SC_ERROR_INVALID_DATA, "parse error: not totaly parsed");
+		LOG_TEST_RET(ctx, SC_ERROR_INVALID_DATA, "parse error: not totally parsed");
 
 	sc_log(ctx,
 	       "docp.acls_contact.size %"SC_FORMAT_LEN_SIZE_T"u; docp.size.size %"SC_FORMAT_LEN_SIZE_T"u",
@@ -833,7 +841,7 @@ iasecc_update_blob(struct sc_context *ctx, struct iasecc_extended_tlv *tlv,
 		unsigned char **blob, size_t *blob_size)
 {
 	unsigned char *pp = NULL;
-	int offs = 0, sz = tlv->size + 2;
+	int offs = 0, sz;
 
 	if (tlv->size == 0)
 		LOG_FUNC_RETURN(ctx, SC_SUCCESS);
@@ -878,7 +886,7 @@ static int
 iasecc_encode_docp(struct sc_context *ctx, struct iasecc_sdo_docp *docp, unsigned char **out, size_t *out_len)
 {
 	struct iasecc_extended_tlv tlv, tlv_st;
-	unsigned char *st_blob, *tmp_blob, *docp_blob;
+	unsigned char *st_blob = NULL, *tmp_blob = NULL, *docp_blob = NULL;
 	size_t blob_size;
 	int rv;
 
@@ -889,63 +897,64 @@ iasecc_encode_docp(struct sc_context *ctx, struct iasecc_sdo_docp *docp, unsigne
 	memset(&tlv, 0, sizeof(tlv));
 	memset(&tlv_st, 0, sizeof(tlv_st));
 
-	st_blob = NULL;
 	blob_size = 0;
 	rv = iasecc_update_blob(ctx, &docp->acls_contact, &st_blob, &blob_size);
-	LOG_TEST_RET(ctx, rv, "ECC: cannot add contact ACLs to blob");
+	LOG_TEST_GOTO_ERR(ctx, rv, "ECC: cannot add contact ACLs to blob");
 
 	rv = iasecc_update_blob(ctx, &docp->acls_contactless, &st_blob, &blob_size);
-	LOG_TEST_RET(ctx, rv, "ECC: cannot add contactless ACLs to blob");
+	LOG_TEST_GOTO_ERR(ctx, rv, "ECC: cannot add contactless ACLs to blob");
 
 	tlv.tag = IASECC_DOCP_TAG_ACLS;
 	tlv.size = blob_size;
 	tlv.value = st_blob;
 
-	tmp_blob = NULL;
 	blob_size = 0;
 	rv = iasecc_update_blob(ctx, &tlv, &tmp_blob, &blob_size);
-	LOG_TEST_RET(ctx, rv, "ECC: cannot add ACLs template to blob");
+	LOG_TEST_GOTO_ERR(ctx, rv, "ECC: cannot add ACLs template to blob");
 
 	rv = iasecc_update_blob(ctx, &docp->name, &tmp_blob, &blob_size);
-	LOG_TEST_RET(ctx, rv, "ECC: cannot add NAME to blob");
+	LOG_TEST_GOTO_ERR(ctx, rv, "ECC: cannot add NAME to blob");
 
 	rv = iasecc_update_blob(ctx, &docp->tries_maximum, &tmp_blob, &blob_size);
-	LOG_TEST_RET(ctx, rv, "ECC: cannot add TRIES MAXIMUM to blob");
+	LOG_TEST_GOTO_ERR(ctx, rv, "ECC: cannot add TRIES MAXIMUM to blob");
 
 	rv = iasecc_update_blob(ctx, &docp->tries_remaining, &tmp_blob, &blob_size);
-	LOG_TEST_RET(ctx, rv, "ECC: cannot add TRIES REMAINING to blob");
+	LOG_TEST_GOTO_ERR(ctx, rv, "ECC: cannot add TRIES REMAINING to blob");
 
 	rv = iasecc_update_blob(ctx, &docp->usage_maximum, &tmp_blob, &blob_size);
-	LOG_TEST_RET(ctx, rv, "ECC: cannot add USAGE MAXIMUM to blob");
+	LOG_TEST_GOTO_ERR(ctx, rv, "ECC: cannot add USAGE MAXIMUM to blob");
 
 	rv = iasecc_update_blob(ctx, &docp->usage_remaining, &tmp_blob, &blob_size);
-	LOG_TEST_RET(ctx, rv, "ECC: cannot add USAGE REMAINING to blob");
+	LOG_TEST_GOTO_ERR(ctx, rv, "ECC: cannot add USAGE REMAINING to blob");
 
 	rv = iasecc_update_blob(ctx, &docp->non_repudiation, &tmp_blob, &blob_size);
-	LOG_TEST_RET(ctx, rv, "ECC: cannot add NON REPUDATION to blob");
+	LOG_TEST_GOTO_ERR(ctx, rv, "ECC: cannot add NON REPUDIATION to blob");
 
 	rv = iasecc_update_blob(ctx, &docp->size, &tmp_blob, &blob_size);
-	LOG_TEST_RET(ctx, rv, "ECC: cannot add SIZE to blob");
+	LOG_TEST_GOTO_ERR(ctx, rv, "ECC: cannot add SIZE to blob");
 
 	rv = iasecc_update_blob(ctx, &docp->issuer_data, &tmp_blob, &blob_size);
-	LOG_TEST_RET(ctx, rv, "ECC: cannot add IDATA to blob");
+	LOG_TEST_GOTO_ERR(ctx, rv, "ECC: cannot add IDATA to blob");
 
 	tlv.tag = IASECC_DOCP_TAG;
 	tlv.size = blob_size;
 	tlv.value = tmp_blob;
 
-	docp_blob = NULL;
 	blob_size = 0;
 	rv = iasecc_update_blob(ctx, &tlv, &docp_blob, &blob_size);
-	LOG_TEST_RET(ctx, rv, "ECC: cannot add ACLs to blob");
-
-	free(tmp_blob);
+	LOG_TEST_GOTO_ERR(ctx, rv, "ECC: cannot add ACLs to blob");
 
 	if (out && out_len)   {
 		*out = docp_blob;
 		*out_len = blob_size;
+		docp_blob = NULL;
 	}
 
+err:
+	free(docp_blob);
+	free(tmp_blob);
+	free(st_blob);
+
 	LOG_FUNC_RETURN(ctx, SC_SUCCESS);
 }
 
@@ -1212,7 +1221,7 @@ iasecc_sdo_parse_card_answer(struct sc_context *ctx, unsigned char *data, size_t
 
 		if (*(data + offs) == IASECC_CARD_ANSWER_TAG_DATA )   {
 			if (size > sizeof(out->data))
-				LOG_TEST_RET(ctx, SC_ERROR_BUFFER_TOO_SMALL, "iasecc_sm_decode_answer() unbelivable !!!");
+				LOG_TEST_RET(ctx, SC_ERROR_BUFFER_TOO_SMALL, "iasecc_sm_decode_answer() unbelievable !!!");
 
 			memcpy(out->data, data + offs + size_size + 1, size);
 			out->data_len = size;
