@@ -105,37 +105,33 @@ typedef struct _pgp_manuf_map {
 } pgp_manuf_map_t;
 
 static const pgp_manuf_map_t manuf_map[] = {
-	{ 0x0001, "PPC Card Systems"   },
-	{ 0x0002, "Prism"              },
-	{ 0x0003, "OpenFortress"       },
-	{ 0x0004, "Wewid AB"           },
-	{ 0x0005, "ZeitControl"        },
-	{ 0x0006, "Yubico"             },
-	{ 0x0007, "OpenKMS"            },
-	{ 0x0008, "LogoEmail"          },
-	{ 0x0009, "Fidesmo"            },
-	{ 0x000A, "Dangerous Things"   },
-	{ 0x002A, "Magrathea"          },
-	{ 0x0042, "GnuPG e.V."         },
-	{ 0x1337, "Warsaw Hackerspace" },
-	{ 0x2342, "warpzone"           },
-	{ 0x63AF, "Trustica"           },
-	{ 0xBD0E, "Paranoidlabs"       },
-	{ 0xF517, "FSIJ"               },
-	{ 0x0000, "test card"          },
-	{ 0xffff, "test card"          },
+	{ 0x0001, "PPC Card Systems"		},
+	{ 0x0002, "Prism"			},
+	{ 0x0003, "OpenFortress"		},
+	{ 0x0004, "Wewid AB"			},
+	{ 0x0005, "ZeitControl"			},
+	{ 0x0006, "Yubico"			},
+	{ 0x0007, "OpenKMS"			},
+	{ 0x0008, "LogoEmail"			},
+	{ 0x0009, "Fidesmo"			},
+	{ 0x000A, "Dangerous Things"		},
+	{ 0x000B, "Feitian Technologies"	},
+	{ 0x002A, "Magrathea"			},
+	{ 0x0042, "GnuPG e.V."			},
+	{ 0x1337, "Warsaw Hackerspace"		},
+	{ 0x2342, "warpzone"			},
+	{ 0x4354, "Confidential Technologies"	},
+	{ 0x5443, "TIF-IT e.V."			},
+	{ 0x63AF, "Trustica"			},
+	{ 0xBA53, "c-base e.V."			},
+	{ 0xBD0E, "Paranoidlabs"		},
+	{ 0xF517, "FSIJ"			},
+	{ 0xF5EC, "F-Secure"			},
+	{ 0x0000, "test card"			},
+	{ 0xffff, "test card"			},
 	{ 0, NULL }
 };
 
-
-static void
-set_string(char **strp, const char *value)
-{
-	if (*strp)
-		free(*strp);
-	*strp = value? strdup(value) : NULL;
-}
-
 /*
  * This function pretty much follows what find_tlv in the GNUpg
  * code does.
@@ -263,7 +259,7 @@ sc_pkcs15emu_openpgp_init(sc_pkcs15_card_t *p15card)
 		return SC_ERROR_OBJECT_NOT_VALID;
 	}
 
-	/* XXX: check if "halfkeys" can be stored with gpg2. If not, add keypairs in one loop */
+	/* XXX: check if "halfkeys" can be stored with gpg2. If not, add key pairs in one loop */
 	for (i = 0; i < 3; i++) {
 		sc_pkcs15_prkey_info_t prkey_info;
 		sc_pkcs15_object_t     prkey_obj;
