@@ -620,8 +620,8 @@ static int sc_pkcs15emu_piv_init(sc_pkcs15_card_t *p15card)
 	/* could read this off card if needed */
 
 	/* CSP does not like a - in the name */
-	p15card->tokeninfo->label = strdup("PIV_II");
-	p15card->tokeninfo->manufacturer_id = strdup(MANU_ID);
+	set_string(&p15card->tokeninfo->label, "PIV_II");
+	set_string(&p15card->tokeninfo->manufacturer_id, MANU_ID);
 
 	/*
 	 * get serial number 
@@ -633,10 +633,10 @@ static int sc_pkcs15emu_piv_init(sc_pkcs15_card_t *p15card)
 	r = sc_card_ctl(card, SC_CARDCTL_GET_SERIALNR, &serial);
 	if (r < 0) {
 		sc_log(card->ctx, "sc_card_ctl rc=%d",r);
-		p15card->tokeninfo->serial_number = strdup("00000000");
+		set_string(&p15card->tokeninfo->serial_number, "00000000");
 	} else {
 		sc_bin_to_hex(serial.value, serial.len, buf, sizeof(buf), 0);
-		p15card->tokeninfo->serial_number = strdup(buf);
+		set_string(&p15card->tokeninfo->serial_number, buf);
 	}
 	/* US gov issued PIVs have CHUID with a FASCN that does not start with 9999 */
 	if (serial.len == 25 && !(serial.value[0] == 0xD4 && serial.value[1] == 0xE7 && serial.value[2] == 0x39 && (serial.value[3] | 0x7F) == 0xFF)) {
@@ -663,7 +663,7 @@ static int sc_pkcs15emu_piv_init(sc_pkcs15_card_t *p15card)
 		strncpy(obj_info.app_label, objects[i].label, SC_PKCS15_MAX_LABEL_SIZE - 1);
 		r = sc_format_oid(&obj_info.app_oid, objects[i].aoid);
 		if (r != SC_SUCCESS)
-			return r;
+			LOG_FUNC_RETURN(card->ctx, r);
 
 		if (objects[i].auth_id)
 			sc_pkcs15_format_id(objects[i].auth_id, &obj_obj.auth_id);
@@ -757,7 +757,7 @@ static int sc_pkcs15emu_piv_init(sc_pkcs15_card_t *p15card)
 		}
 		/* following will find the cached cert in cert_info */
 		r =  sc_pkcs15_read_certificate(p15card, &cert_info, &cert_out);
-		if (r < 0 || cert_out->key == NULL) {
+		if (r < 0 || cert_out == NULL || cert_out->key == NULL) {
 			sc_log(card->ctx,  "Failed to read/parse the certificate r=%d",r);
 			if (cert_out != NULL)
 				sc_pkcs15_free_certificate(cert_out);
@@ -777,8 +777,9 @@ static int sc_pkcs15emu_piv_init(sc_pkcs15_card_t *p15card)
 				if (!token_name) {
 					sc_pkcs15_free_certificate(cert_out);
 					free(cn_name);
-					SC_FUNC_RETURN(card->ctx,
-						SC_ERROR_OUT_OF_MEMORY, r);
+					r = SC_ERROR_OUT_OF_MEMORY;
+					LOG_TEST_GOTO_ERR(card->ctx, r,
+						"Failed to allocate memory");
 				}
 				memcpy(token_name, cn_name, cn_len);
 				free(cn_name);
@@ -967,12 +968,9 @@ sc_log(card->ctx,  "DEE Adding pin %d label=%s",i, label);
 		}
 
 		r = sc_pkcs15emu_add_pin_obj(p15card, &pin_obj, &pin_info);
-		if (r < 0)
-			LOG_FUNC_RETURN(card->ctx, r);
+		LOG_TEST_GOTO_ERR(card->ctx, r, "Failed to add PIN");
 	}
 
-
-
 	/* set public keys */
 	/* We may only need this during initialization when genkey
 	 * gets the pubkey, but it can not be read from the card 
@@ -1043,7 +1041,7 @@ sc_log(card->ctx,  "DEE Adding pin %d label=%s",i, label);
 
 			/* Lets also try another method. */
 			r = sc_pkcs15_encode_pubkey_as_spki(card->ctx, p15_key, &pubkey_info.direct.spki.value, &pubkey_info.direct.spki.len);
-        		LOG_TEST_RET(card->ctx, r, "SPKI encode public key error");
+			LOG_TEST_GOTO_ERR(card->ctx, r, "SPKI encode public key error");
 			
 			/* Only get here if no cert, and the the above found the
 			 * pub key file (actually the SPKI version). This only 
@@ -1077,10 +1075,12 @@ sc_log(card->ctx,  "DEE Adding pin %d label=%s",i, label);
 			p15_key = NULL;
 		}
 		else if (ckis[i].pubkey_from_cert)   {
-			r = sc_pkcs15_encode_pubkey_as_spki(card->ctx, ckis[i].pubkey_from_cert, &pubkey_info.direct.spki.value, &pubkey_info.direct.spki.len);
-        		LOG_TEST_RET(card->ctx, r, "SPKI encode public key error");
+			r = sc_pkcs15_encode_pubkey_as_spki(card->ctx, ckis[i].pubkey_from_cert,
+				&pubkey_info.direct.spki.value, &pubkey_info.direct.spki.len);
+			LOG_TEST_GOTO_ERR(card->ctx, r, "SPKI encode public key error");
 
 			pubkey_obj.emulated = ckis[i].pubkey_from_cert;
+			ckis[i].pubkey_from_cert = NULL;
 		}
 
 		sc_log(card->ctx, "adding pubkey for %d keyalg=%d",i, ckis[i].key_alg);
@@ -1094,9 +1094,9 @@ sc_log(card->ctx,  "DEE Adding pin %d label=%s",i, label);
 				pubkey_info.modulus_length = ckis[i].pubkey_len;
 				strncpy(pubkey_obj.label, pubkeys[i].label, SC_PKCS15_MAX_LABEL_SIZE - 1);
 
+				/* should not fail */
 				r = sc_pkcs15emu_add_rsa_pubkey(p15card, &pubkey_obj, &pubkey_info);
-				if (r < 0)
-					LOG_FUNC_RETURN(card->ctx, r); /* should not fail */
+				LOG_TEST_GOTO_ERR(card->ctx, r, "Failed to add RSA pubkey");
 
 				ckis[i].pubkey_found = 1;
 				break;
@@ -1104,15 +1104,16 @@ sc_log(card->ctx,  "DEE Adding pin %d label=%s",i, label);
 				if (ckis[i].cert_keyUsage_present) {
 					pubkey_info.usage = ckis[i].pub_usage;
 				} else {
-				    pubkey_info.usage = pubkeys[i].usage_ec;
+					pubkey_info.usage = pubkeys[i].usage_ec;
 				}
 
 				pubkey_info.field_length = ckis[i].pubkey_len; 
 				strncpy(pubkey_obj.label, pubkeys[i].label, SC_PKCS15_MAX_LABEL_SIZE - 1);
 
+				/* should not fail */
 				r = sc_pkcs15emu_add_ec_pubkey(p15card, &pubkey_obj, &pubkey_info);
-				if (r < 0) 
-					LOG_FUNC_RETURN(card->ctx, r); /* should not fail */
+				LOG_TEST_GOTO_ERR(card->ctx, r, "Failed to add EC pubkey");
+
 				ckis[i].pubkey_found = 1;
 				break;
 			default:
@@ -1196,13 +1197,17 @@ sc_log(card->ctx,  "DEE Adding pin %d label=%s",i, label);
 				r = 0; /* we just skip this one */
 		}
 		sc_log(card->ctx, "USAGE: cert_keyUsage_present:%d usage:0x%8.8x", ckis[i].cert_keyUsage_present ,prkey_info.usage);
-		if (r < 0)
-			LOG_FUNC_RETURN(card->ctx, r);
+		LOG_TEST_GOTO_ERR(card->ctx, r, "Failed to add Private key");
 	}
 
 	p15card->ops.get_guid = piv_get_guid;
 
 	LOG_FUNC_RETURN(card->ctx, SC_SUCCESS);
+err:
+	for (i = 0; i < PIV_NUM_CERTS_AND_KEYS; i++) {
+		sc_pkcs15_free_pubkey(ckis[i].pubkey_from_cert);
+	}
+	LOG_FUNC_RETURN(card->ctx, r);
 }
 
 int sc_pkcs15emu_piv_init_ex(sc_pkcs15_card_t *p15card,
