@@ -34,21 +34,21 @@
 #include <unistd.h>
 #endif
 
-#include "internal.h"
-#include "asn1.h"
-#include "pkcs15.h"
-
 #ifdef ENABLE_OPENSSL
+#include <openssl/bn.h>
 #include <openssl/rsa.h>
 #include <openssl/dsa.h>
 #include <openssl/evp.h>
 #include <openssl/err.h>
-#if OPENSSL_VERSION_NUMBER >= 0x10000000L
 #ifndef OPENSSL_NO_EC
 #include <openssl/ec.h>
 #endif
 #endif
-#endif
+
+#include "internal.h"
+#include "asn1.h"
+#include "pkcs15.h"
+
 
 #define C_ASN1_PKINFO_ATTR_SIZE 3
 static const struct sc_asn1_entry c_asn1_pkinfo[C_ASN1_PKINFO_ATTR_SIZE] = {
@@ -198,8 +198,8 @@ sc_pkcs15_decode_pubkey_direct_value(struct sc_pkcs15_card *p15card, struct sc_p
 		LOG_TEST_RET(ctx, rv, "Failed to decode 'SPKI' direct value");
 
 		rv = sc_pkcs15_encode_pubkey(ctx, pubkey, &info->direct.raw.value, &info->direct.raw.len);
-		LOG_TEST_RET(ctx, rv, "Failed to endode 'RAW' direct value");
 		sc_pkcs15_free_pubkey(pubkey);
+		LOG_TEST_RET(ctx, rv, "Failed to encode 'RAW' direct value");
 	}
 
 	LOG_FUNC_RETURN(ctx, SC_SUCCESS);
@@ -211,11 +211,10 @@ int sc_pkcs15_decode_pukdf_entry(struct sc_pkcs15_card *p15card,
 		const u8 ** buf, size_t *buflen)
 {
 	struct sc_context *ctx = p15card->card->ctx;
-	struct sc_pkcs15_pubkey_info info;
+	struct sc_pkcs15_pubkey_info *info;
 	int r, gostr3410_params[3];
 	struct sc_pkcs15_keyinfo_gostparams *keyinfo_gostparams;
-	size_t usage_len = sizeof(info.usage);
-	size_t af_len = sizeof(info.access_flags);
+	size_t usage_len, af_len;
 	struct sc_pkcs15_der *der = &obj->content;
 	struct sc_asn1_entry asn1_com_key_attr[C_ASN1_COM_KEY_ATTR_SIZE];
 	struct sc_asn1_entry asn1_com_pubkey_attr[C_ASN1_COM_PUBKEY_ATTR_SIZE];
@@ -240,6 +239,14 @@ int sc_pkcs15_decode_pukdf_entry(struct sc_pkcs15_card *p15card,
 	struct sc_asn1_pkcs15_object gostr3410key_obj =  { obj, asn1_com_key_attr,
 			asn1_com_pubkey_attr, asn1_gostr3410_type_attr };
 
+	info = calloc(1, sizeof *info);
+	if (info == NULL) {
+		r = SC_ERROR_OUT_OF_MEMORY;
+		goto err;
+	}
+	usage_len = sizeof(info->usage);
+	af_len = sizeof(info->access_flags);
+
 	sc_copy_asn1_entry(c_asn1_pubkey, asn1_pubkey);
 	sc_copy_asn1_entry(c_asn1_pubkey_choice, asn1_pubkey_choice);
 	sc_copy_asn1_entry(c_asn1_rsa_type_attr, asn1_rsa_type_attr);
@@ -255,7 +262,7 @@ int sc_pkcs15_decode_pukdf_entry(struct sc_pkcs15_card *p15card,
 	sc_copy_asn1_entry(c_asn1_com_pubkey_attr, asn1_com_pubkey_attr);
 	sc_copy_asn1_entry(c_asn1_com_key_attr, asn1_com_key_attr);
 
-	sc_format_asn1_entry(asn1_com_pubkey_attr + 0, &info.subject.value, &info.subject.len, 0);
+	sc_format_asn1_entry(asn1_com_pubkey_attr + 0, &info->subject.value, &info->subject.len, 0);
 
 	sc_format_asn1_entry(asn1_pubkey_choice + 0, &rsakey_obj, NULL, 0);
 	sc_format_asn1_entry(asn1_pubkey_choice + 1, &dsakey_obj, NULL, 0);
@@ -264,61 +271,62 @@ int sc_pkcs15_decode_pukdf_entry(struct sc_pkcs15_card *p15card,
 
 	sc_format_asn1_entry(asn1_rsa_type_attr + 0, asn1_rsakey_attr, NULL, 0);
 
-	sc_format_asn1_entry(asn1_rsakey_value_choice + 0, &info.path, NULL, 0);
+	sc_format_asn1_entry(asn1_rsakey_value_choice + 0, &info->path, NULL, 0);
 	sc_format_asn1_entry(asn1_rsakey_value_choice + 1, &der->value, &der->len, 0);
 
 	sc_format_asn1_entry(asn1_rsakey_attr + 0, asn1_rsakey_value_choice, NULL, 0);
-	sc_format_asn1_entry(asn1_rsakey_attr + 1, &info.modulus_length, NULL, 0);
+	sc_format_asn1_entry(asn1_rsakey_attr + 1, &info->modulus_length, NULL, 0);
 
 	sc_format_asn1_entry(asn1_ec_type_attr + 0, asn1_eckey_attr, NULL, 0);
 
-	sc_format_asn1_entry(asn1_eckey_value_choice + 0, &info.path, NULL, 0);
+	sc_format_asn1_entry(asn1_eckey_value_choice + 0, &info->path, NULL, 0);
 	sc_format_asn1_entry(asn1_eckey_value_choice + 1, &der->value, &der->len, 0);
 
 	sc_format_asn1_entry(asn1_eckey_attr + 0, asn1_eckey_value_choice, NULL, 0);
 
 	sc_format_asn1_entry(asn1_dsa_type_attr + 0, asn1_dsakey_attr, NULL, 0);
 
-	sc_format_asn1_entry(asn1_dsakey_attr + 0, &info.path, NULL, 0);
+	sc_format_asn1_entry(asn1_dsakey_attr + 0, &info->path, NULL, 0);
 
 	sc_format_asn1_entry(asn1_gostr3410_type_attr + 0, asn1_gostr3410key_attr, NULL, 0);
 
-	sc_format_asn1_entry(asn1_gostr3410key_attr + 0, &info.path, NULL, 0);
+	sc_format_asn1_entry(asn1_gostr3410key_attr + 0, &info->path, NULL, 0);
 	sc_format_asn1_entry(asn1_gostr3410key_attr + 1, &gostr3410_params[0], NULL, 0);
 	sc_format_asn1_entry(asn1_gostr3410key_attr + 2, &gostr3410_params[1], NULL, 0);
 	sc_format_asn1_entry(asn1_gostr3410key_attr + 3, &gostr3410_params[2], NULL, 0);
 
-	sc_format_asn1_entry(asn1_com_key_attr + 0, &info.id, NULL, 0);
-	sc_format_asn1_entry(asn1_com_key_attr + 1, &info.usage, &usage_len, 0);
-	sc_format_asn1_entry(asn1_com_key_attr + 2, &info.native, NULL, 0);
-	sc_format_asn1_entry(asn1_com_key_attr + 3, &info.access_flags, &af_len, 0);
-	sc_format_asn1_entry(asn1_com_key_attr + 4, &info.key_reference, NULL, 0);
+	sc_format_asn1_entry(asn1_com_key_attr + 0, &info->id, NULL, 0);
+	sc_format_asn1_entry(asn1_com_key_attr + 1, &info->usage, &usage_len, 0);
+	sc_format_asn1_entry(asn1_com_key_attr + 2, &info->native, NULL, 0);
+	sc_format_asn1_entry(asn1_com_key_attr + 3, &info->access_flags, &af_len, 0);
+	sc_format_asn1_entry(asn1_com_key_attr + 4, &info->key_reference, NULL, 0);
 
 	sc_format_asn1_entry(asn1_pubkey + 0, asn1_pubkey_choice, NULL, 0);
 
 	/* Fill in defaults */
-	memset(&info, 0, sizeof(info));
-	info.key_reference = -1;
-	info.native = 1;
+	info->key_reference = -1;
+	info->native = 1;
 	memset(gostr3410_params, 0, sizeof(gostr3410_params));
 
 	r = sc_asn1_decode(ctx, asn1_pubkey, *buf, *buflen, buf, buflen);
 	if (r == SC_ERROR_ASN1_END_OF_CONTENTS)
-		return r;
-	LOG_TEST_RET(ctx, r, "ASN.1 decoding failed");
+		goto err;
+	LOG_TEST_GOTO_ERR(ctx, r, "ASN.1 decoding failed");
 	if (asn1_pubkey_choice[0].flags & SC_ASN1_PRESENT) {
 		obj->type = SC_PKCS15_TYPE_PUBKEY_RSA;
 	} else if (asn1_pubkey_choice[2].flags & SC_ASN1_PRESENT) {
 		obj->type = SC_PKCS15_TYPE_PUBKEY_GOSTR3410;
-		assert(info.modulus_length == 0);
-		info.modulus_length = SC_PKCS15_GOSTR3410_KEYSIZE;
-		assert(info.params.len == 0);
-		info.params.len = sizeof(struct sc_pkcs15_keyinfo_gostparams);
-		info.params.data = malloc(info.params.len);
-		if (info.params.data == NULL)
-			LOG_FUNC_RETURN(ctx, SC_ERROR_OUT_OF_MEMORY);
-		assert(sizeof(*keyinfo_gostparams) == info.params.len);
-		keyinfo_gostparams = info.params.data;
+		assert(info->modulus_length == 0);
+		info->modulus_length = SC_PKCS15_GOSTR3410_KEYSIZE;
+		assert(info->params.len == 0);
+		info->params.len = sizeof(struct sc_pkcs15_keyinfo_gostparams);
+		info->params.data = malloc(info->params.len);
+		if (info->params.data == NULL) {
+			r = SC_ERROR_OUT_OF_MEMORY;
+			goto err;
+		}
+		assert(sizeof(*keyinfo_gostparams) == info->params.len);
+		keyinfo_gostparams = info->params.data;
 		keyinfo_gostparams->gostr3410 = (unsigned int)gostr3410_params[0];
 		keyinfo_gostparams->gostr3411 = (unsigned int)gostr3410_params[1];
 		keyinfo_gostparams->gost28147 = (unsigned int)gostr3410_params[2];
@@ -330,35 +338,43 @@ int sc_pkcs15_decode_pukdf_entry(struct sc_pkcs15_card *p15card,
 		obj->type = SC_PKCS15_TYPE_PUBKEY_DSA;
 	}
 
-	if (!p15card->app || !p15card->app->ddo.aid.len)   {
-		r = sc_pkcs15_make_absolute_path(&p15card->file_app->path, &info.path);
+	if (!p15card->app || !p15card->app->ddo.aid.len) {
+		if (!p15card->file_app) {
+			r = SC_ERROR_INTERNAL;
+			goto err;
+		}
+		r = sc_pkcs15_make_absolute_path(&p15card->file_app->path, &info->path);
 		if (r < 0) {
-			sc_pkcs15_free_key_params(&info.params);
-			return r;
+			goto err;
 		}
 	}
 	else   {
-		info.path.aid = p15card->app->ddo.aid;
+		info->path.aid = p15card->app->ddo.aid;
 	}
-	sc_log(ctx, "PubKey path '%s'", sc_print_path(&info.path));
+	sc_log(ctx, "PubKey path '%s'", sc_print_path(&info->path));
 
 	/* OpenSC 0.11.4 and older encoded "keyReference" as a negative
 	   value. Fixed in 0.11.5 we need to add a hack, so old cards
 	   continue to work. */
-	if (info.key_reference < -1)
-		info.key_reference += 256;
+	if (info->key_reference < -1)
+		info->key_reference += 256;
 
-	obj->data = malloc(sizeof(info));
-	if (obj->data == NULL) {
-		sc_pkcs15_free_key_params(&info.params);
-		LOG_FUNC_RETURN(ctx, SC_ERROR_OUT_OF_MEMORY);
-	}
-	memcpy(obj->data, &info, sizeof(info));
+	obj->data = info;
+	info = NULL;
 
 	r = sc_pkcs15_decode_pubkey_direct_value(p15card, obj);
-	LOG_TEST_RET(ctx, r, "Decode public key direct value failed");
+	if (r < 0) {
+		info = obj->data;
+		obj->data = NULL;
+	}
+	LOG_TEST_GOTO_ERR(ctx, r, "Decode public key direct value failed");
 
-	return 0;
+err:
+	if (r < 0) {
+		sc_pkcs15_free_pubkey_info(info);
+	}
+
+	LOG_FUNC_RETURN(ctx, r);
 }
 
 
@@ -724,8 +740,6 @@ sc_pkcs15_decode_pubkey_ec(sc_context_t *ctx,
 	if (*ecpoint_data != 0x04)
 		LOG_TEST_RET(ctx, SC_ERROR_NOT_SUPPORTED, "Supported only uncompressed EC pointQ value");
 
-	sc_log(ctx, "decode-EC key=%p, buf=%p, buflen=%d", key, buf, buflen);
-
 	key->ecpointQ.len = ecpoint_len;
 	key->ecpointQ.value = ecpoint_data;
 
@@ -744,17 +758,13 @@ sc_pkcs15_encode_pubkey_ec(sc_context_t *ctx, struct sc_pkcs15_pubkey_ec *key,
 		u8 **buf, size_t *buflen)
 {
 	struct sc_asn1_entry asn1_ec_pointQ[C_ASN1_EC_POINTQ_SIZE];
-	int r;
 
 	LOG_FUNC_CALLED(ctx);
 	sc_copy_asn1_entry(c_asn1_ec_pointQ, asn1_ec_pointQ);
 	sc_format_asn1_entry(asn1_ec_pointQ + 0, key->ecpointQ.value, &key->ecpointQ.len, 1);
 
-	r = sc_asn1_encode(ctx, asn1_ec_pointQ, buf, buflen);
-	LOG_TEST_RET(ctx, r, "ASN.1 encoding failed");
-
-	sc_log(ctx, "EC key->ecpointQ=%p:%d *buf=%p:%d", key->ecpointQ.value, key->ecpointQ.len, *buf, *buflen);
-	LOG_FUNC_RETURN(ctx, SC_SUCCESS);
+	LOG_FUNC_RETURN(ctx,
+			sc_asn1_encode(ctx, asn1_ec_pointQ, buf, buflen));
 }
 
 
@@ -841,6 +851,8 @@ sc_pkcs15_encode_pubkey_as_spki(sc_context_t *ctx, struct sc_pkcs15_pubkey *pubk
 			}
 			memcpy(ec_params->der.value, pubkey->u.ec.params.der.value, pubkey->u.ec.params.der.len);
 			ec_params->der.len = pubkey->u.ec.params.der.len;
+			/* This could have been already allocated: avoid memory leak */
+			sc_asn1_clear_algorithm_id(pubkey->alg_id);
 			pubkey->alg_id->params = ec_params;
 		}
 		break;
@@ -898,14 +910,19 @@ int
 sc_pkcs15_read_pubkey(struct sc_pkcs15_card *p15card, const struct sc_pkcs15_object *obj,
 		struct sc_pkcs15_pubkey **out)
 {
-	struct sc_context *ctx = p15card->card->ctx;
+	struct sc_context *ctx;
 	const struct sc_pkcs15_pubkey_info *info = NULL;
 	struct sc_pkcs15_pubkey *pubkey = NULL;
 	unsigned char *data = NULL;
 	size_t	len;
 	int	algorithm, r;
 
-	assert(p15card != NULL && obj != NULL && out != NULL);
+	if (p15card == NULL || p15card->card == NULL || p15card->card->ops == NULL
+			|| obj == NULL || out == NULL) {
+		return SC_ERROR_INVALID_ARGUMENTS;
+	}
+	ctx = p15card->card->ctx;
+
 	LOG_FUNC_CALLED(ctx);
 	sc_log(ctx, "Public key type 0x%X", obj->type);
 
@@ -929,7 +946,6 @@ sc_pkcs15_read_pubkey(struct sc_pkcs15_card *p15card, const struct sc_pkcs15_obj
 
 	pubkey = calloc(1, sizeof(struct sc_pkcs15_pubkey));
 	if (pubkey == NULL) {
-		free(data);
 		LOG_FUNC_RETURN(ctx, SC_ERROR_OUT_OF_MEMORY);
 	}
 	pubkey->algorithm = algorithm;
@@ -980,10 +996,11 @@ sc_pkcs15_read_pubkey(struct sc_pkcs15_card *p15card, const struct sc_pkcs15_obj
 	}
 
 err:
-	if (r)
+	if (r) {
 		sc_pkcs15_free_pubkey(pubkey);
-	else
+	} else
 		*out = pubkey;
+	free(data);
 
 	LOG_FUNC_RETURN(ctx, r);
 }
@@ -992,7 +1009,9 @@ err:
 static int
 sc_pkcs15_dup_bignum (struct sc_pkcs15_bignum *dst, struct sc_pkcs15_bignum *src)
 {
-	assert(dst && src);
+	if (!dst || !src) {
+		return SC_ERROR_INVALID_ARGUMENTS;
+	}
 
 	if (src->data && src->len)   {
 		dst->data = calloc(1, src->len);
@@ -1013,7 +1032,9 @@ sc_pkcs15_pubkey_from_prvkey(struct sc_context *ctx, struct sc_pkcs15_prkey *prv
 	struct sc_pkcs15_pubkey *pubkey = NULL;
 	int rv = SC_SUCCESS;
 
-	assert(prvkey && out);
+	if (!prvkey || !out) {
+		return SC_ERROR_INVALID_ARGUMENTS;
+	}
 
 	*out = NULL;
 	pubkey = calloc(1, sizeof(struct sc_pkcs15_pubkey));
@@ -1069,7 +1090,11 @@ sc_pkcs15_dup_pubkey(struct sc_context *ctx, struct sc_pkcs15_pubkey *key, struc
 	u8* alg;
 	size_t alglen;
 
-	assert(key && out);
+	LOG_FUNC_CALLED(ctx);
+
+	if (!key || !out) {
+		return SC_ERROR_INVALID_ARGUMENTS;
+	}
 
 	*out = NULL;
 	pubkey = calloc(1, sizeof(struct sc_pkcs15_pubkey));
@@ -1082,6 +1107,10 @@ sc_pkcs15_dup_pubkey(struct sc_context *ctx, struct sc_pkcs15_pubkey *key, struc
 		rv = sc_asn1_encode_algorithm_id(ctx, &alg, &alglen,key->alg_id, 0);
 		if (rv == SC_SUCCESS) {
 			pubkey->alg_id = (struct sc_algorithm_id *)calloc(1, sizeof(struct sc_algorithm_id));
+			if (pubkey->alg_id == NULL) {
+				free(pubkey);
+				LOG_FUNC_RETURN(ctx, SC_ERROR_OUT_OF_MEMORY);
+			}
 			rv = sc_asn1_decode_algorithm_id(ctx, alg, alglen, pubkey->alg_id, 0);
 			free(alg);
 		}
@@ -1121,9 +1150,15 @@ sc_pkcs15_dup_pubkey(struct sc_context *ctx, struct sc_pkcs15_pubkey *key, struc
 		memcpy(pubkey->u.ec.params.der.value, key->u.ec.params.der.value, key->u.ec.params.der.len);
 		pubkey->u.ec.params.der.len = key->u.ec.params.der.len;
 
-		pubkey->u.ec.params.named_curve = strdup(key->u.ec.params.named_curve);
-		if (!pubkey->u.ec.params.named_curve)
-			rv = SC_ERROR_OUT_OF_MEMORY;
+		if (key->u.ec.params.named_curve){
+			pubkey->u.ec.params.named_curve = strdup(key->u.ec.params.named_curve);
+			if (!pubkey->u.ec.params.named_curve)
+				rv = SC_ERROR_OUT_OF_MEMORY;
+		}
+		else {
+			sc_log(ctx, "named_curve parameter missing");
+			rv = SC_ERROR_NOT_SUPPORTED;
+		}
 
 		break;
 	default:
@@ -1144,7 +1179,9 @@ sc_pkcs15_dup_pubkey(struct sc_context *ctx, struct sc_pkcs15_pubkey *key, struc
 void
 sc_pkcs15_erase_pubkey(struct sc_pkcs15_pubkey *key)
 {
-	assert(key != NULL);
+	if (key == NULL) {
+		return;
+	}
 	if (key->alg_id) {
 		sc_asn1_clear_algorithm_id(key->alg_id);
 		free(key->alg_id);
@@ -1196,10 +1233,13 @@ sc_pkcs15_free_pubkey(struct sc_pkcs15_pubkey *key)
 void
 sc_pkcs15_free_pubkey_info(sc_pkcs15_pubkey_info_t *info)
 {
-	if (info->subject.value)
+	if (info) {
 		free(info->subject.value);
-	sc_pkcs15_free_key_params(&info->params);
-	free(info);
+		free(info->direct.spki.value);
+		free(info->direct.raw.value);
+		sc_pkcs15_free_key_params(&info->params);
+		free(info);
+	}
 }
 
 
@@ -1240,16 +1280,16 @@ sc_pkcs15_read_der_file(sc_context_t *ctx, char * filename,
 
 	body = tagbuf;
 	r = sc_asn1_read_tag(&body, len, &cla_out, &tag_out, &bodylen);
-	if (r != SC_SUCCESS)
+	if (r != SC_SUCCESS && r != SC_ERROR_ASN1_END_OF_CONTENTS)
 		goto out;
 
-	if (tag_out == SC_ASN1_TAG_EOC || body == NULL)   {
+	if (body == NULL)   {
 		r = SC_SUCCESS;
 		goto out;
 	}
 
 	offs = body - tagbuf;
-	if (offs > len || offs < 2)   {
+	if (offs > len || offs < 2 || offs > offs + bodylen)   {
 		r = SC_ERROR_INVALID_ASN1_OBJECT;
 		goto out;
 	}
@@ -1298,7 +1338,9 @@ sc_pkcs15_pubkey_from_spki_fields(struct sc_context *ctx, struct sc_pkcs15_pubke
 	unsigned char *tmp_buf = NULL;
 	int r;
 
-	sc_log(ctx, "sc_pkcs15_pubkey_from_spki_fields() called: %p:%d\n%s", buf, buflen, sc_dump_hex(buf, buflen));
+	sc_log(ctx,
+	       "sc_pkcs15_pubkey_from_spki_fields() called: %p:%"SC_FORMAT_LEN_SIZE_T"u\n%s",
+	       buf, buflen, sc_dump_hex(buf, buflen));
 
 	tmp_buf = malloc(buflen);
 	if (!tmp_buf) {
@@ -1316,7 +1358,6 @@ sc_pkcs15_pubkey_from_spki_fields(struct sc_context *ctx, struct sc_pkcs15_pubke
 		r = SC_ERROR_OUT_OF_MEMORY;
 		LOG_TEST_GOTO_ERR(ctx, r, "");
 	}
-	*outpubkey = pubkey;
 
 	sc_copy_asn1_entry(c_asn1_pkinfo, asn1_pkinfo);
 
@@ -1324,7 +1365,10 @@ sc_pkcs15_pubkey_from_spki_fields(struct sc_context *ctx, struct sc_pkcs15_pubke
 	sc_format_asn1_entry(asn1_pkinfo + 1, &pk.value, &pk.len, 0);
 
 	r = sc_asn1_decode(ctx, asn1_pkinfo, tmp_buf, buflen, NULL, NULL);
-	LOG_TEST_GOTO_ERR(ctx, r, "ASN.1 parsing of subjectPubkeyInfo failed");
+	if (r != SC_SUCCESS) {
+		sc_asn1_clear_algorithm_id(&pk_alg);
+		LOG_TEST_GOTO_ERR(ctx, r, "ASN.1 parsing of subjectPubkeyInfo failed");
+	}
 
 	pubkey->alg_id = calloc(1, sizeof(struct sc_algorithm_id));
 	if (pubkey->alg_id == NULL) {
@@ -1363,8 +1407,10 @@ sc_pkcs15_pubkey_from_spki_fields(struct sc_context *ctx, struct sc_pkcs15_pubke
 		}
 
 		pubkey->u.ec.ecpointQ.value = malloc(pk.len);
-		if (pubkey->u.ec.ecpointQ.value == NULL)
-			LOG_FUNC_RETURN(ctx, SC_ERROR_OUT_OF_MEMORY);
+		if (pubkey->u.ec.ecpointQ.value == NULL) {
+			r = SC_ERROR_OUT_OF_MEMORY;
+			LOG_TEST_GOTO_ERR(ctx, r, "failed to malloc() memory");
+		}
 		memcpy(pubkey->u.ec.ecpointQ.value, pk.value, pk.len);
 		pubkey->u.ec.ecpointQ.len = pk.len;
 	}
@@ -1374,7 +1420,12 @@ sc_pkcs15_pubkey_from_spki_fields(struct sc_context *ctx, struct sc_pkcs15_pubke
 		LOG_TEST_GOTO_ERR(ctx, r, "ASN.1 parsing of subjectPubkeyInfo failed");
 	}
 
+	*outpubkey = pubkey;
+	pubkey = NULL;
+
 err:
+	if (pubkey)
+		sc_pkcs15_free_pubkey(pubkey);
 	if (pk.value)
 		free(pk.value);
 	if (tmp_buf)
@@ -1399,8 +1450,11 @@ sc_pkcs15_pubkey_from_spki_sequence(struct sc_context *ctx, const unsigned char
 	r = sc_asn1_decode(ctx, asn1_spki, buf, buflen, NULL, NULL);
 	LOG_TEST_RET(ctx, r, "ASN.1 cannot parse subjectPublicKeyInfo");
 
-	if(outpubkey)
+	if(outpubkey) {
+		free(*outpubkey);
 		*outpubkey = pubkey;
+	} else
+		free(pubkey);
 
 	LOG_FUNC_RETURN(ctx, r);
 }
@@ -1458,6 +1512,8 @@ static struct ec_curve_info {
 		{"brainpoolP224r1",	"1.3.36.3.3.2.8.1.1.5", "06092B2403030208010105", 224},
 		{"brainpoolP256r1",	"1.3.36.3.3.2.8.1.1.7", "06092B2403030208010107", 256},
 		{"brainpoolP320r1",	"1.3.36.3.3.2.8.1.1.9", "06092B2403030208010109", 320},
+		{"brainpoolP384r1",	"1.3.36.3.3.2.8.1.1.11", "06092B240303020801010B", 384},
+		{"brainpoolP512r1",	"1.3.36.3.3.2.8.1.1.13", "06092B240303020801010D", 512},
 
 		{"secp192k1",		"1.3.132.0.31", "06052B8104001F", 192},
 		{"secp256k1",		"1.3.132.0.10", "06052B8104000A", 256},
@@ -1507,7 +1563,8 @@ sc_pkcs15_fix_ec_parameters(struct sc_context *ctx, struct sc_ec_parameters *ecp
 			sc_format_oid(&ecparams->id, ec_curve_infos[ii].oid_str);
 
 		ecparams->field_length = ec_curve_infos[ii].size;
-		sc_log(ctx, "Curve length %i", ecparams->field_length);
+		sc_log(ctx, "Curve length %"SC_FORMAT_LEN_SIZE_T"u",
+		       ecparams->field_length);
 	}
 	else if (ecparams->named_curve)   {	/* it can be name of curve or OID in ASCII form */
 		for (ii=0; ec_curve_infos[ii].name; ii++)   {
@@ -1531,9 +1588,8 @@ sc_pkcs15_fix_ec_parameters(struct sc_context *ctx, struct sc_ec_parameters *ecp
 			LOG_TEST_RET(ctx, rv, "Cannot encode object ID");
 		}
 	}
-	else if (sc_valid_oid(&ecparams->id))  {
+	else
 		LOG_TEST_RET(ctx, SC_ERROR_NOT_IMPLEMENTED, "EC parameters has to be presented as a named curve or explicit data");
-	}
 
 	LOG_FUNC_RETURN(ctx, SC_SUCCESS);
 }
@@ -1544,14 +1600,19 @@ sc_pkcs15_convert_pubkey(struct sc_pkcs15_pubkey *pkcs15_key, void *evp_key)
 {
 #ifdef ENABLE_OPENSSL
 	EVP_PKEY *pk = (EVP_PKEY *)evp_key;
+	int pk_type;
+	pk_type = EVP_PKEY_base_id(pk);
 
-	switch (pk->type) {
+	switch (pk_type) {
 	case EVP_PKEY_RSA: {
 		struct sc_pkcs15_pubkey_rsa *dst = &pkcs15_key->u.rsa;
 		RSA *src = EVP_PKEY_get1_RSA(pk);
+		const BIGNUM *src_n, *src_e;
+
+		RSA_get0_key(src, &src_n, &src_e, NULL);
 
 		pkcs15_key->algorithm = SC_ALGORITHM_RSA;
-		if (!sc_pkcs15_convert_bignum(&dst->modulus, src->n) || !sc_pkcs15_convert_bignum(&dst->exponent, src->e))
+		if (!sc_pkcs15_convert_bignum(&dst->modulus, src_n) || !sc_pkcs15_convert_bignum(&dst->exponent, src_e))
 			return SC_ERROR_INVALID_DATA;
 		RSA_free(src);
 		break;
@@ -1559,16 +1620,20 @@ sc_pkcs15_convert_pubkey(struct sc_pkcs15_pubkey *pkcs15_key, void *evp_key)
 	case EVP_PKEY_DSA: {
 		struct sc_pkcs15_pubkey_dsa *dst = &pkcs15_key->u.dsa;
 		DSA *src = EVP_PKEY_get1_DSA(pk);
+		const BIGNUM *src_pub_key, *src_priv_key, *src_p, *src_q, *src_g;
+
+		DSA_get0_key(src, &src_pub_key, &src_priv_key);
+		DSA_get0_pqg(src, &src_p, &src_q, &src_g);
 
 		pkcs15_key->algorithm = SC_ALGORITHM_DSA;
-		sc_pkcs15_convert_bignum(&dst->pub, src->pub_key);
-		sc_pkcs15_convert_bignum(&dst->p, src->p);
-		sc_pkcs15_convert_bignum(&dst->q, src->q);
-		sc_pkcs15_convert_bignum(&dst->g, src->g);
+		sc_pkcs15_convert_bignum(&dst->pub, src_pub_key);
+		sc_pkcs15_convert_bignum(&dst->p, src_p);
+		sc_pkcs15_convert_bignum(&dst->q, src_q);
+		sc_pkcs15_convert_bignum(&dst->g, src_g);
 		DSA_free(src);
 		break;
 	}
-#if OPENSSL_VERSION_NUMBER >= 0x10000000L && !defined(OPENSSL_NO_EC)
+#if !defined(OPENSSL_NO_EC)
 	case NID_id_GostR3410_2001: {
 		struct sc_pkcs15_pubkey_gostr3410 *dst = &pkcs15_key->u.gostr3410;
 		EC_KEY *eckey = EVP_PKEY_get0(pk);
@@ -1583,7 +1648,7 @@ sc_pkcs15_convert_pubkey(struct sc_pkcs15_pubkey *pkcs15_key, void *evp_key)
 		X = BN_new();
 		Y = BN_new();
 		if (X && Y && EC_KEY_get0_group(eckey))
-			r = EC_POINT_get_affine_coordinates_GFp(EC_KEY_get0_group(eckey),
+					r = EC_POINT_get_affine_coordinates_GFp(EC_KEY_get0_group(eckey),
 					point, X, Y, NULL);
 		if (r == 1) {
 			dst->xy.len = BN_num_bytes(X) + BN_num_bytes(Y);
@@ -1630,7 +1695,7 @@ sc_pkcs15_convert_pubkey(struct sc_pkcs15_pubkey *pkcs15_key, void *evp_key)
 		nid = EC_GROUP_get_curve_name(grp);
 		if(nid != 0) {
 			const char *name = OBJ_nid2sn(nid);
-			if(sizeof(name) > 0)
+			if (name)
 				dst->params.named_curve = strdup(name);
 		}
 
@@ -1649,7 +1714,7 @@ sc_pkcs15_convert_pubkey(struct sc_pkcs15_pubkey *pkcs15_key, void *evp_key)
 
 		break;
 	}
-#endif /* OPENSSL_VERSION_NUMBER >= 0x10000000L && !defined(OPENSSL_NO_EC) */
+#endif /* !defined(OPENSSL_NO_EC) */
 	default:
 		return SC_ERROR_NOT_SUPPORTED;
 	}
