@@ -99,9 +99,10 @@ static int insert_key(
 	int               key_length,
 	unsigned char     auth_id,
 	const char       *label
-){
-	sc_card_t *card=p15card->card;
-	sc_context_t *ctx=p15card->card->ctx;
+)
+{
+	sc_card_t *card = p15card->card;
+	sc_context_t *ctx = p15card->card->ctx;
 	sc_file_t *f;
 	struct sc_pkcs15_prkey_info prkey_info;
 	struct sc_pkcs15_object prkey_obj;
@@ -121,13 +122,14 @@ static int insert_key(
 	prkey_obj.auth_id.len      = 1;
 	prkey_obj.auth_id.value[0] = auth_id;
 
-	can_sign=can_crypt=0;
-	if(card->type==SC_CARD_TYPE_TCOS_V3){
+	can_sign = can_crypt = 0;
+	if (card->type == SC_CARD_TYPE_TCOS_V3) {
 		unsigned char buf[256];
-		int i, rec_no=0;
-		if(prkey_info.path.len>=2) prkey_info.path.len-=2;
+		int i, rec_no = 0;
+		if (prkey_info.path.len >= 2)
+			prkey_info.path.len -= 2;
 		sc_append_file_id(&prkey_info.path, 0x5349);
-		if(sc_select_file(card, &prkey_info.path, NULL)!=SC_SUCCESS){
+		if (sc_select_file(card, &prkey_info.path, NULL) != SC_SUCCESS) {
 			sc_log(ctx, 
 				"Select(%s) failed\n",
 				sc_print_path(&prkey_info.path));
@@ -135,40 +137,50 @@ static int insert_key(
 		}
 		sc_log(ctx, 
 			"Searching for Key-Ref %02X\n", key_reference);
-		while((r=sc_read_record(card, ++rec_no, buf, sizeof(buf), SC_RECORD_BY_REC_NR))>0){
-			int found=0;
-			if(buf[0]!=0xA0) continue;
-			for(i=2;i<buf[1]+2;i+=2+buf[i+1]){
-				if(buf[i]==0x83 && buf[i+1]==1 && buf[i+2]==key_reference) ++found;
+		while ((r = sc_read_record(card, ++rec_no, buf, sizeof(buf), SC_RECORD_BY_REC_NR)) > 0) {
+			int found = 0;
+			if (buf[0] != 0xA0 || r < 2)
+				continue;
+			for (i = 2; i < buf[1] + 2 && i < r - 2; i += 2 + buf[i + 1]) {
+				if (buf[i] == 0x83 && buf[i + 1] == 1 && buf[i + 2] == key_reference)
+					++found;
 			}
-			if(found) break;
+			if (found)
+				break;
 		}
-		if(r<=0){
+		if (r <= 0) {
 			sc_log(ctx, "No EF_KEYD-Record found\n");
 			return 1;
 		}
-		for(i=0;i<r;i+=2+buf[i+1]){
-			if(buf[i]==0xB6) can_sign++;
-			if(buf[i]==0xB8) can_crypt++;
+		for (i = 0; i < r; i += 2 + buf[i + 1]) {
+			if (buf[i] == 0xB6)
+				can_sign++;
+			if (buf[i] == 0xB8)
+				can_crypt++;
 		}
 	} else {
-		if(sc_select_file(card, &prkey_info.path, &f)!=SC_SUCCESS
+		if (sc_select_file(card, &prkey_info.path, &f) != SC_SUCCESS
 			   	|| !f->prop_attr || f->prop_attr_len < 2){
 			sc_log(ctx, 
 				"Select(%s) failed\n",
 				sc_print_path(&prkey_info.path));
+			sc_file_free(f);
 			return 1;
 		}
-		if (f->prop_attr[1] & 0x04) can_crypt=1;
-		if (f->prop_attr[1] & 0x08) can_sign=1;
+		if (f->prop_attr[1] & 0x04)
+			can_crypt = 1;
+		if (f->prop_attr[1] & 0x08)
+			can_sign = 1;
 		sc_file_free(f);
 	}
-	prkey_info.usage= SC_PKCS15_PRKEY_USAGE_SIGN;
-	if(can_crypt) prkey_info.usage |= SC_PKCS15_PRKEY_USAGE_ENCRYPT|SC_PKCS15_PRKEY_USAGE_DECRYPT;
-	if(can_sign) prkey_info.usage |= SC_PKCS15_PRKEY_USAGE_NONREPUDIATION;
+	prkey_info.usage = SC_PKCS15_PRKEY_USAGE_SIGN;
+	if (can_crypt)
+		prkey_info.usage |= SC_PKCS15_PRKEY_USAGE_ENCRYPT | SC_PKCS15_PRKEY_USAGE_DECRYPT;
+	if (can_sign)
+		prkey_info.usage |= SC_PKCS15_PRKEY_USAGE_NONREPUDIATION;
 
-	r=sc_pkcs15emu_add_rsa_prkey(p15card, &prkey_obj, &prkey_info);
-	if(r!=SC_SUCCESS){
+	r = sc_pkcs15emu_add_rsa_prkey(p15card, &prkey_obj, &prkey_info);
+	if(r != SC_SUCCESS) {
 		sc_log(ctx,  "sc_pkcs15emu_add_rsa_prkey(%s) failed\n", path);
 		return 4;
 	}
@@ -188,7 +200,7 @@ static int insert_pin(
 ){
 	sc_card_t *card=p15card->card;
 	sc_context_t *ctx=p15card->card->ctx;
-	sc_file_t *f;
+	sc_file_t *f = NULL;
 	struct sc_pkcs15_auth_info pin_info;
 	struct sc_pkcs15_object pin_obj;
 	int r;
@@ -241,9 +253,10 @@ static int insert_pin(
 			return 1;
 		}
 	} else {
-		if(sc_select_file(card, &pin_info.path, &f)!=SC_SUCCESS
+		if (sc_select_file(card, &pin_info.path, &f) != SC_SUCCESS
 			   	|| !f->prop_attr || f->prop_attr_len < 4){
 			sc_log(ctx, "Select(%s) failed\n", path);
+			sc_file_free(f);
 			return 1;
 		}
 		pin_info.tries_left=f->prop_attr[3];
@@ -285,8 +298,8 @@ static int detect_netkey(
 	sprintf(dir,"%04X", f->id);
 	sc_file_free(f);
 
-	p15card->tokeninfo->manufacturer_id = strdup("TeleSec GmbH");
-	p15card->tokeninfo->label = strdup(card->type==SC_CARD_TYPE_TCOS_V3 ? "NetKey V3 Card" : "NetKey Card");
+	set_string(&p15card->tokeninfo->manufacturer_id, "TeleSec GmbH");
+	set_string(&p15card->tokeninfo->label, card->type==SC_CARD_TYPE_TCOS_V3 ? "NetKey V3 Card" : "NetKey Card");
 	keylen= card->type==SC_CARD_TYPE_TCOS_V3 ? 2048 : 1024;
 	c_auth= card->type==SC_CARD_TYPE_TCOS_V3 ? "C500" : "C100";
 
@@ -373,13 +386,15 @@ static int detect_idkey(
 	memcpy(p.value, "\xD2\x76\x00\x00\x03\x0C\x01", p.len=7);
 	if (sc_select_file(card,&p,NULL)!=SC_SUCCESS) return 1;
 
-	p15card->tokeninfo->manufacturer_id = strdup("TeleSec GmbH");
-	p15card->tokeninfo->label = strdup("IDKey Card");
+	set_string(&p15card->tokeninfo->manufacturer_id, "TeleSec GmbH");
+	set_string(&p15card->tokeninfo->label, "IDKey Card");
 
 	insert_cert(p15card, "DF074331", 0x45, 1, "Signatur Zertifikat 1");
 	insert_cert(p15card, "DF074332", 0x46, 1, "Signatur Zertifikat 2");
 	insert_cert(p15card, "DF074333", 0x47, 1, "Signatur Zertifikat 3");
 	insert_cert(p15card, "DF084331", 0x4B, 1, "Verschluesselungs Zertifikat 1");
+	insert_cert(p15card, "DF084332", 0x4C, 1, "Verschluesselungs Zertifikat 2");
+	insert_cert(p15card, "DF084333", 0x4D, 1, "Verschluesselungs Zertifikat 3");
 	/* TODO should others come here too? */
 
 	insert_key(p15card, "DF074E03", 0x45, 0x84, 2048, 1, "IDKey1");
@@ -389,6 +404,8 @@ static int detect_idkey(
 	insert_key(p15card, "DF074E07", 0x49, 0x88, 2048, 1, "IDKey5");
 	insert_key(p15card, "DF074E08", 0x4A, 0x89, 2048, 1, "IDKey6");
 	insert_key(p15card, "DF084E01", 0x4B, 0x81, 2048, 1, "IDKey7");
+	insert_key(p15card, "DF084E02", 0x4C, 0x82, 2048, 1, "IDKey8");
+	insert_key(p15card, "DF084E03", 0x4D, 0x83, 2048, 1, "IDKey9");
 
 	insert_pin(p15card, "5000", 1, 2, 0x00, 6, "PIN",
 		SC_PKCS15_PIN_FLAG_CASE_SENSITIVE | SC_PKCS15_PIN_FLAG_INITIALIZED
@@ -405,8 +422,8 @@ static int detect_signtrust(
 	sc_pkcs15_card_t *p15card
 ){
 	if(insert_cert(p15card,"8000DF01C000", 0x45, 1, "Signatur Zertifikat")) return 1;
-	p15card->tokeninfo->manufacturer_id = strdup("Deutsche Post");
-	p15card->tokeninfo->label = strdup("SignTrust Card");
+	set_string(&p15card->tokeninfo->manufacturer_id, "Deutsche Post");
+	set_string(&p15card->tokeninfo->label, "SignTrust Card");
 
 	insert_cert(p15card,"800082008220", 0x46, 1, "Verschluesselungs Zertifikat");
 	insert_cert(p15card,"800083008320", 0x47, 1, "Authentifizierungs Zertifikat");
@@ -435,8 +452,8 @@ static int detect_datev(
 	sc_pkcs15_card_t *p15card
 ){
 	if(insert_cert(p15card,"3000C500", 0x45, 0, "Signatur Zertifikat")) return 1;
-	p15card->tokeninfo->manufacturer_id = strdup("DATEV");
-	p15card->tokeninfo->label = strdup("DATEV Classic");
+	set_string(&p15card->tokeninfo->manufacturer_id, "DATEV");
+	set_string(&p15card->tokeninfo->label, "DATEV Classic");
 
 	insert_cert(p15card,"DF02C200", 0x46, 0, "Verschluesselungs Zertifikat");
 	insert_cert(p15card,"DF02C500", 0x47, 0, "Authentifizierungs Zertifikat");
@@ -456,8 +473,8 @@ static int detect_unicard(
 	sc_pkcs15_card_t *p15card
 ){
 	if(!insert_cert(p15card,"41004352", 0x45, 1, "Zertifikat 1")){
-		p15card->tokeninfo->manufacturer_id = strdup("JLU Giessen");
-		p15card->tokeninfo->label = strdup("JLU Giessen Card");
+		set_string(&p15card->tokeninfo->manufacturer_id, "JLU Giessen");
+		set_string(&p15card->tokeninfo->label, "JLU Giessen Card");
 
 		insert_cert(p15card,"41004353", 0x46, 1, "Zertifikat 2");
 		insert_cert(p15card,"41004354", 0x47, 1, "Zertifikat 3");
@@ -466,8 +483,8 @@ static int detect_unicard(
 		insert_key(p15card,"41005105", 0x47, 0x85, 1024, 1, "Schluessel 3");
 
 	} else if(!insert_cert(p15card,"41014352", 0x45, 1, "Zertifikat 1")){
-		p15card->tokeninfo->manufacturer_id = strdup("TU Darmstadt");
-		p15card->tokeninfo->label = strdup("TUD Card");
+		set_string(&p15card->tokeninfo->manufacturer_id, "TU Darmstadt");
+		set_string(&p15card->tokeninfo->label, "TUD Card");
 
 		insert_cert(p15card,"41014353", 0x46, 1, "Zertifikat 2");
 		insert_cert(p15card,"41014354", 0x47, 1, "Zertifikat 3");
@@ -509,7 +526,7 @@ int sc_pkcs15emu_tcos_init_ex(
 	}
 	sc_bin_to_hex(serialnr.value, serialnr.len , serial, sizeof(serial), 0);
 	serial[19] = '\0';
-	p15card->tokeninfo->serial_number = strdup(serial);
+	set_string(&p15card->tokeninfo->serial_number, serial);
 
 	if(!detect_netkey(p15card)) return SC_SUCCESS;
 	if(!detect_idkey(p15card)) return SC_SUCCESS;
