@@ -50,17 +50,31 @@ static const struct sc_asn1_entry c_asn1_twlabel[] = {
 static const struct sc_asn1_entry c_asn1_algorithm_info[7] = {
 	{ "reference",		SC_ASN1_INTEGER,	SC_ASN1_TAG_INTEGER,	0, NULL, NULL },
 	{ "algorithmPKCS#11",	SC_ASN1_INTEGER,	SC_ASN1_TAG_INTEGER,	0, NULL, NULL },
-	{ "parameters",		SC_ASN1_NULL,		SC_ASN1_TAG_NULL,	0, NULL, NULL },
+	{ "parameters",		SC_ASN1_CHOICE,		0,			0, NULL, NULL },
 	{ "supportedOperations",SC_ASN1_BIT_FIELD,	SC_ASN1_TAG_BIT_STRING,	0, NULL, NULL },
 	{ "objId",		SC_ASN1_OBJECT,		SC_ASN1_TAG_OBJECT,	SC_ASN1_OPTIONAL, NULL, NULL },
 	{ "algRef",		SC_ASN1_INTEGER,	SC_ASN1_TAG_INTEGER,	SC_ASN1_OPTIONAL, NULL, NULL },
 	{ NULL, 0, 0, 0, NULL, NULL }
 };
 
+static const struct sc_asn1_entry c_asn1_algorithm_info_parameters[3] = {
+	{ "PKCS15RSAParameters",SC_ASN1_NULL,		SC_ASN1_TAG_NULL,	0, NULL, NULL },
+	{ "PKCS15ECParameters",	SC_ASN1_OBJECT,		SC_ASN1_TAG_OBJECT,	0, NULL, NULL },
+	{ NULL, 0, 0, 0, NULL, NULL }
+};
+
 /*
- * in src/libopensc/types.h SC_MAX_SUPPORTED_ALGORITHMS  defined as 8
+ * in src/libopensc/types.h SC_MAX_SUPPORTED_ALGORITHMS  defined as 16
  */
 static const struct sc_asn1_entry c_asn1_supported_algorithms[SC_MAX_SUPPORTED_ALGORITHMS + 1] = {
+	{ "algorithmInfo", SC_ASN1_STRUCT, SC_ASN1_TAG_SEQUENCE | SC_ASN1_CONS, SC_ASN1_OPTIONAL, NULL, NULL },
+	{ "algorithmInfo", SC_ASN1_STRUCT, SC_ASN1_TAG_SEQUENCE | SC_ASN1_CONS, SC_ASN1_OPTIONAL, NULL, NULL },
+	{ "algorithmInfo", SC_ASN1_STRUCT, SC_ASN1_TAG_SEQUENCE | SC_ASN1_CONS, SC_ASN1_OPTIONAL, NULL, NULL },
+	{ "algorithmInfo", SC_ASN1_STRUCT, SC_ASN1_TAG_SEQUENCE | SC_ASN1_CONS, SC_ASN1_OPTIONAL, NULL, NULL },
+	{ "algorithmInfo", SC_ASN1_STRUCT, SC_ASN1_TAG_SEQUENCE | SC_ASN1_CONS, SC_ASN1_OPTIONAL, NULL, NULL },
+	{ "algorithmInfo", SC_ASN1_STRUCT, SC_ASN1_TAG_SEQUENCE | SC_ASN1_CONS, SC_ASN1_OPTIONAL, NULL, NULL },
+	{ "algorithmInfo", SC_ASN1_STRUCT, SC_ASN1_TAG_SEQUENCE | SC_ASN1_CONS, SC_ASN1_OPTIONAL, NULL, NULL },
+	{ "algorithmInfo", SC_ASN1_STRUCT, SC_ASN1_TAG_SEQUENCE | SC_ASN1_CONS, SC_ASN1_OPTIONAL, NULL, NULL },
 	{ "algorithmInfo", SC_ASN1_STRUCT, SC_ASN1_TAG_SEQUENCE | SC_ASN1_CONS, SC_ASN1_OPTIONAL, NULL, NULL },
 	{ "algorithmInfo", SC_ASN1_STRUCT, SC_ASN1_TAG_SEQUENCE | SC_ASN1_CONS, SC_ASN1_OPTIONAL, NULL, NULL },
 	{ "algorithmInfo", SC_ASN1_STRUCT, SC_ASN1_TAG_SEQUENCE | SC_ASN1_CONS, SC_ASN1_OPTIONAL, NULL, NULL },
@@ -125,18 +139,20 @@ int sc_pkcs15_parse_tokeninfo(sc_context_t *ctx,
 	u8 serial[128];
 	size_t serial_len = sizeof(serial);
 	u8 mnfid[SC_PKCS15_MAX_LABEL_SIZE];
-	size_t mnfid_len  = sizeof(mnfid);
+	size_t mnfid_len  = sizeof(mnfid) - 1;
 	u8 label[SC_PKCS15_MAX_LABEL_SIZE];
-	size_t label_len = sizeof(label);
+	size_t label_len = sizeof(label) - 1;
 	u8 last_update[32], profile_indication[SC_PKCS15_MAX_LABEL_SIZE];
 	size_t lupdate_len = sizeof(last_update) - 1, pi_len = sizeof(profile_indication) - 1;
 	size_t flags_len   = sizeof(ti->flags);
 	u8 preferred_language[3];
 	size_t lang_length = sizeof(preferred_language);
 	struct sc_asn1_entry asn1_supported_algorithms[SC_MAX_SUPPORTED_ALGORITHMS + 1],
-			asn1_algo_infos[SC_MAX_SUPPORTED_ALGORITHMS][7];
+			asn1_algo_infos[SC_MAX_SUPPORTED_ALGORITHMS][7],
+			asn1_algo_infos_parameters[SC_MAX_SUPPORTED_ALGORITHMS][3];
 	size_t reference_len = sizeof(ti->supported_algos[0].reference);
 	size_t mechanism_len = sizeof(ti->supported_algos[0].mechanism);
+	size_t parameter_len = sizeof(ti->supported_algos[0].parameters);
 	size_t operations_len = sizeof(ti->supported_algos[0].operations);
 	size_t algo_ref_len = sizeof(ti->supported_algos[0].algo_ref);
 
@@ -145,6 +161,10 @@ int sc_pkcs15_parse_tokeninfo(sc_context_t *ctx,
 	struct sc_asn1_entry asn1_toki_attrs[C_ASN1_TOKI_ATTRS_SIZE], asn1_tokeninfo[3], asn1_twlabel[3];
 
 	memset(last_update, 0, sizeof(last_update));
+	memset(label, 0, sizeof(label));
+	memset(profile_indication, 0, sizeof(profile_indication));
+	memset(mnfid, 0, sizeof(mnfid));
+
 	sc_copy_asn1_entry(c_asn1_twlabel, asn1_twlabel);
 	sc_copy_asn1_entry(c_asn1_toki_attrs, asn1_toki_attrs);
 	sc_copy_asn1_entry(c_asn1_tokeninfo, asn1_tokeninfo);
@@ -152,14 +172,22 @@ int sc_pkcs15_parse_tokeninfo(sc_context_t *ctx,
 	sc_format_asn1_entry(asn1_twlabel, label, &label_len, 0);
 	sc_copy_asn1_entry(c_asn1_profile_indication, asn1_profile_indication);
 
-	for (ii=0; ii<SC_MAX_SUPPORTED_ALGORITHMS; ii++)
+	for (ii=0; ii<SC_MAX_SUPPORTED_ALGORITHMS; ii++) {
 		sc_copy_asn1_entry(c_asn1_algorithm_info, asn1_algo_infos[ii]);
+		sc_copy_asn1_entry(c_asn1_algorithm_info_parameters,
+			asn1_algo_infos_parameters[ii]);
+	}
 	sc_copy_asn1_entry(c_asn1_supported_algorithms, asn1_supported_algorithms);
 
 	for (ii=0; ii<SC_MAX_SUPPORTED_ALGORITHMS; ii++)   {
 		sc_format_asn1_entry(asn1_algo_infos[ii] + 0, &ti->supported_algos[ii].reference, &reference_len, 0);
 		sc_format_asn1_entry(asn1_algo_infos[ii] + 1, &ti->supported_algos[ii].mechanism, &mechanism_len, 0);
-		sc_format_asn1_entry(asn1_algo_infos[ii] + 2, NULL, NULL, 0);
+		sc_format_asn1_entry(asn1_algo_infos[ii] + 2,
+			asn1_algo_infos_parameters[ii], NULL, 0);
+		sc_format_asn1_entry(asn1_algo_infos_parameters[ii] + 0,
+			NULL, NULL, 0);
+		sc_format_asn1_entry(asn1_algo_infos_parameters[ii] + 1,
+			&ti->supported_algos[ii].parameters, &parameter_len, 0);
 		sc_format_asn1_entry(asn1_algo_infos[ii] + 3, &ti->supported_algos[ii].operations, &operations_len, 0);
 		sc_format_asn1_entry(asn1_algo_infos[ii] + 4, &ti->supported_algos[ii].algo_id, NULL, 1);
 		sc_format_asn1_entry(asn1_algo_infos[ii] + 5, &ti->supported_algos[ii].algo_ref, &algo_ref_len, 0);
@@ -191,18 +219,12 @@ int sc_pkcs15_parse_tokeninfo(sc_context_t *ctx,
 	r = sc_asn1_decode(ctx, asn1_tokeninfo, buf, blen, NULL, NULL);
 	LOG_TEST_RET(ctx, r, "ASN.1 parsing of EF(TokenInfo) failed");
 
-	if (asn1_toki_attrs[1].flags & SC_ASN1_PRESENT)   {
+	if (asn1_toki_attrs[1].flags & SC_ASN1_PRESENT && serial_len > 0)   {
+		free(ti->serial_number);
 		ti->serial_number = malloc(serial_len * 2 + 1);
 		if (ti->serial_number == NULL)
 			return SC_ERROR_OUT_OF_MEMORY;
-
-		ti->serial_number[0] = 0;
-		for (ii = 0; ii < serial_len; ii++) {
-			char byte[3];
-
-			sprintf(byte, "%02X", serial[ii]);
-			strcat(ti->serial_number, byte);
-		}
+		sc_bin_to_hex(serial, serial_len, ti->serial_number, serial_len * 2 + 1, 0);
 		sc_log(ctx, "TokenInfo.serialNunmber '%s'", ti->serial_number);
 	}
 
@@ -270,9 +292,11 @@ sc_pkcs15_encode_tokeninfo(sc_context_t *ctx, sc_pkcs15_tokeninfo_t *ti,
 	struct sc_asn1_entry asn1_toki_attrs[C_ASN1_TOKI_ATTRS_SIZE];
 	struct sc_asn1_entry asn1_tokeninfo[2];
 	struct sc_asn1_entry asn1_supported_algorithms[SC_MAX_SUPPORTED_ALGORITHMS + 1],
-			asn1_algo_infos[SC_MAX_SUPPORTED_ALGORITHMS][7];
+			asn1_algo_infos[SC_MAX_SUPPORTED_ALGORITHMS][7],
+			asn1_algo_infos_parameters[SC_MAX_SUPPORTED_ALGORITHMS][3];
 	size_t reference_len = sizeof(ti->supported_algos[0].reference);
 	size_t mechanism_len = sizeof(ti->supported_algos[0].mechanism);
+	size_t parameter_len = sizeof(ti->supported_algos[0].parameters);
 	size_t operations_len = sizeof(ti->supported_algos[0].operations);
 	size_t algo_ref_len = sizeof(ti->supported_algos[0].algo_ref);
 	struct sc_asn1_entry asn1_last_update[C_ASN1_LAST_UPDATE_SIZE];
@@ -283,14 +307,26 @@ sc_pkcs15_encode_tokeninfo(sc_context_t *ctx, sc_pkcs15_tokeninfo_t *ti,
 	sc_copy_asn1_entry(c_asn1_last_update, asn1_last_update);
 	sc_copy_asn1_entry(c_asn1_profile_indication, asn1_profile_indication);
 
-	for (ii=0; ii<SC_MAX_SUPPORTED_ALGORITHMS && ti->supported_algos[ii].reference; ii++)
+	for (ii=0; ii<SC_MAX_SUPPORTED_ALGORITHMS && ti->supported_algos[ii].reference; ii++) {
 		sc_copy_asn1_entry(c_asn1_algorithm_info, asn1_algo_infos[ii]);
+		sc_copy_asn1_entry(c_asn1_algorithm_info_parameters,
+			asn1_algo_infos_parameters[ii]);
+	}
 	sc_copy_asn1_entry(c_asn1_supported_algorithms, asn1_supported_algorithms);
 
 	for (ii=0; ii<SC_MAX_SUPPORTED_ALGORITHMS && ti->supported_algos[ii].reference; ii++)   {
 		sc_format_asn1_entry(asn1_algo_infos[ii] + 0, &ti->supported_algos[ii].reference, &reference_len, 1);
 		sc_format_asn1_entry(asn1_algo_infos[ii] + 1, &ti->supported_algos[ii].mechanism, &mechanism_len, 1);
-		sc_format_asn1_entry(asn1_algo_infos[ii] + 2, NULL, NULL, 0);
+		sc_format_asn1_entry(asn1_algo_infos[ii] + 2,
+			asn1_algo_infos_parameters[ii], NULL, 1);
+		if (!sc_valid_oid(&ti->supported_algos[ii].parameters)) {
+			sc_format_asn1_entry(asn1_algo_infos_parameters[ii] + 0,
+				NULL, NULL, 1);
+		}
+		else {
+			sc_format_asn1_entry(asn1_algo_infos_parameters[ii] + 1,
+				&ti->supported_algos[ii].parameters, &parameter_len, 0);
+		}
 		sc_format_asn1_entry(asn1_algo_infos[ii] + 3, &ti->supported_algos[ii].operations, &operations_len, 1);
 		sc_format_asn1_entry(asn1_algo_infos[ii] + 4, &ti->supported_algos[ii].algo_id, NULL, 1);
 		sc_format_asn1_entry(asn1_algo_infos[ii] + 5, &ti->supported_algos[ii].algo_ref, &algo_ref_len, 1);
@@ -413,52 +449,6 @@ fix_authentic_ddo(struct sc_pkcs15_card *p15card)
 	}
 }
 
-
-static void
-fix_starcos_pkcs15_card(struct sc_pkcs15_card *p15card)
-{
-	struct sc_context *ctx = p15card->card->ctx;
-
-	/* set special flags based on card meta data */
-	if (strcmp(p15card->card->driver->short_name,"cardos") == 0
-			&& p15card->tokeninfo && p15card->tokeninfo->label) {
-
-		/* D-Trust cards (D-TRUST, D-SIGN) */
-		if (strstr(p15card->tokeninfo->label,"D-TRUST") != NULL
-				|| strstr(p15card->tokeninfo->label,"D-SIGN") != NULL) {
-
-			/* D-TRUST Card 2.0 2cc (standard cards, which always add
-			 * SHA1 prefix itself */
-			if (strstr(p15card->tokeninfo->label, "2cc") != NULL) {
-				p15card->card->caps |= SC_CARD_CAP_ONLY_RAW_HASH_STRIPPED;
-				sc_log(ctx, "D-TRUST 2cc card detected, only SHA1 works with this card");
-				/* XXX: add detection when other hash than SHA1 is used with
-				 *      such a card, as this produces invalid signatures.
-				 */
-			}
-
-			/* D-SIGN multicard 2.0 2ca (cards working with all types of hashes
-			 * and no addition of prefix) */
-			else if (strstr(p15card->tokeninfo->label, "2ca") != NULL) {
-				p15card->card->caps |= SC_CARD_CAP_ONLY_RAW_HASH;
-				sc_log(ctx, "D-TRUST 2ca card detected");
-			}
-
-			/* D-TRUST card 2.4 2ce (cards working with all types of hashes
-			 * and no addition of prefix) */
-			else if (strstr(p15card->tokeninfo->label, "2ce") != NULL) {
-				p15card->card->caps |= SC_CARD_CAP_ONLY_RAW_HASH;
-				sc_log(ctx, "D-TRUST 2ce card detected");
-			}
-
-			/* XXX: probably there are more D-Trust card in the wild,
-			 *      which also need these flags to produce valid signatures
-			 */
-		}
-	}
-}
-
-
 static int
 parse_ddo(struct sc_pkcs15_card *p15card, const u8 * buf, size_t buflen)
 {
@@ -491,6 +481,8 @@ parse_ddo(struct sc_pkcs15_card *p15card, const u8 * buf, size_t buflen)
 		p15card->file_odf->path = odf_path;
 	}
 	if (asn1_ddo[2].flags & SC_ASN1_PRESENT) {
+		if (p15card->file_tokeninfo)
+			sc_file_free(p15card->file_tokeninfo);
 		p15card->file_tokeninfo = sc_file_new();
 		if (p15card->file_tokeninfo == NULL)
 			goto mem_err;
@@ -624,12 +616,14 @@ parse_odf(const unsigned char * buf, size_t buflen, struct sc_pkcs15_card *p15ca
 		if (r < 0)
 			return r;
 		type = r;
-		r = sc_pkcs15_make_absolute_path(&p15card->file_app->path, &path);
-		if (r < 0)
-			return r;
-		r = sc_pkcs15_add_df(p15card, odf_indexes[type], &path);
-		if (r)
-			return r;
+		if (p15card->file_app) {
+			r = sc_pkcs15_make_absolute_path(&p15card->file_app->path, &path);
+			if (r < 0)
+				return r;
+			r = sc_pkcs15_add_df(p15card, odf_indexes[type], &path);
+			if (r)
+				return r;
+		}
 	}
 	return 0;
 }
@@ -712,55 +706,58 @@ sc_pkcs15_card_new(void)
 		return NULL;
 	}
 
-	sc_init_oid(&p15card->tokeninfo->profile_indication.oid);
-
 	p15card->magic = SC_PKCS15_CARD_MAGIC;
 	return p15card;
 }
 
 
-void
-sc_pkcs15_free_tokeninfo(struct sc_pkcs15_card *p15card)
+struct sc_pkcs15_tokeninfo *
+sc_pkcs15_tokeninfo_new(void)
 {
-	if (!p15card || !p15card->tokeninfo)
+	struct sc_pkcs15_tokeninfo *tokeninfo;
+
+	tokeninfo = calloc(1, sizeof(struct sc_pkcs15_tokeninfo));
+	if (tokeninfo == NULL) {
+		return NULL;
+	}
+
+	sc_init_oid(&tokeninfo->profile_indication.oid);
+
+	return tokeninfo;
+}
+
+
+void
+sc_pkcs15_free_tokeninfo(struct sc_pkcs15_tokeninfo *tokeninfo)
+{
+	if (!tokeninfo)
 		return;
 
-	if (p15card->tokeninfo->label != NULL)
-		free(p15card->tokeninfo->label);
-	if (p15card->tokeninfo->serial_number != NULL)
-		free(p15card->tokeninfo->serial_number);
-	if (p15card->tokeninfo->manufacturer_id != NULL)
-		free(p15card->tokeninfo->manufacturer_id);
-	if (p15card->tokeninfo->last_update.gtime != NULL)
-		free(p15card->tokeninfo->last_update.gtime);
-	if (p15card->tokeninfo->preferred_language != NULL)
-		free(p15card->tokeninfo->preferred_language);
-	if (p15card->tokeninfo->profile_indication.name != NULL)
-		free(p15card->tokeninfo->profile_indication.name);
-	if (p15card->tokeninfo->seInfo != NULL) {
+	free(tokeninfo->label);
+	free(tokeninfo->serial_number);
+	free(tokeninfo->manufacturer_id);
+	free(tokeninfo->last_update.gtime);
+	free(tokeninfo->preferred_language);
+	free(tokeninfo->profile_indication.name);
+	if (tokeninfo->seInfo != NULL) {
 		unsigned i;
-		for (i = 0; i < p15card->tokeninfo->num_seInfo; i++)
-			free(p15card->tokeninfo->seInfo[i]);
-		free(p15card->tokeninfo->seInfo);
+		for (i = 0; i < tokeninfo->num_seInfo; i++)
+			free(tokeninfo->seInfo[i]);
+		free(tokeninfo->seInfo);
 	}
-	free(p15card->tokeninfo);
-
-	p15card->tokeninfo = NULL;
+	free(tokeninfo);
 }
 
 
 void
 sc_pkcs15_free_app(struct sc_pkcs15_card *p15card)
 {
-	if (!p15card || !p15card->app)
-		return;
-
-	if (p15card->app->label)
+	if (p15card && p15card->app) {
 		free(p15card->app->label);
-	if (p15card->app->ddo.value)
 		free(p15card->app->ddo.value);
-	free(p15card->app);
-	p15card->app = NULL;
+		free(p15card->app);
+		p15card->app = NULL;
+	}
 }
 
 
@@ -778,6 +775,7 @@ sc_pkcs15_card_free(struct sc_pkcs15_card *p15card)
 	if (p15card->md_data)
 		free(p15card->md_data);
 
+	sc_pkcs15_free_app(p15card);
 	sc_pkcs15_remove_objects(p15card);
 	sc_pkcs15_remove_dfs(p15card);
 	sc_pkcs15_free_unusedspace(p15card);
@@ -789,7 +787,7 @@ sc_pkcs15_card_free(struct sc_pkcs15_card *p15card)
 	sc_file_free(p15card->file_unusedspace);
 
 	p15card->magic = 0;
-	sc_pkcs15_free_tokeninfo(p15card);
+	sc_pkcs15_free_tokeninfo(p15card->tokeninfo);
 	sc_pkcs15_free_app(p15card);
 	free(p15card);
 }
@@ -820,30 +818,19 @@ sc_pkcs15_card_clear(struct sc_pkcs15_card *p15card)
 	p15card->file_odf = NULL;
 	sc_file_free(p15card->file_unusedspace);
 	p15card->file_unusedspace = NULL;
-	if (p15card->tokeninfo->label != NULL) {
-		free(p15card->tokeninfo->label);
-		p15card->tokeninfo->label = NULL;
-	}
-	if (p15card->tokeninfo->serial_number != NULL) {
-		free(p15card->tokeninfo->serial_number);
-		p15card->tokeninfo->serial_number = NULL;
-	}
-	if (p15card->tokeninfo->manufacturer_id != NULL) {
-		free(p15card->tokeninfo->manufacturer_id);
-		p15card->tokeninfo->manufacturer_id = NULL;
-	}
-	if (p15card->tokeninfo->last_update.gtime != NULL) {
-		free(p15card->tokeninfo->last_update.gtime);
-		p15card->tokeninfo->last_update.gtime = NULL;
-	}
-	if (p15card->tokeninfo->preferred_language != NULL) {
-		free(p15card->tokeninfo->preferred_language);
-		p15card->tokeninfo->preferred_language = NULL;
-	}
-	if (p15card->tokeninfo->profile_indication.name != NULL)   {
-		free(p15card->tokeninfo->profile_indication.name);
-		p15card->tokeninfo->profile_indication.name = NULL;
-	}
+
+	free(p15card->tokeninfo->label);
+	p15card->tokeninfo->label = NULL;
+	free(p15card->tokeninfo->serial_number);
+	p15card->tokeninfo->serial_number = NULL;
+	free(p15card->tokeninfo->manufacturer_id);
+	p15card->tokeninfo->manufacturer_id = NULL;
+	free(p15card->tokeninfo->last_update.gtime);
+	p15card->tokeninfo->last_update.gtime = NULL;
+	free(p15card->tokeninfo->preferred_language);
+	p15card->tokeninfo->preferred_language = NULL;
+	free(p15card->tokeninfo->profile_indication.name);
+	p15card->tokeninfo->profile_indication.name = NULL;
 	if (p15card->tokeninfo->seInfo != NULL) {
 		size_t i;
 		for (i = 0; i < p15card->tokeninfo->num_seInfo; i++)
@@ -972,6 +959,7 @@ sc_pkcs15_bind_internal(struct sc_pkcs15_card *p15card, struct sc_aid *aid)
 		if (err != SC_SUCCESS)
 			sc_log(ctx, "unable to enumerate apps: %s", sc_strerror(err));
 	}
+	sc_file_free(p15card->file_app);
 	p15card->file_app = sc_file_new();
 	if (p15card->file_app == NULL) {
 		err = SC_ERROR_OUT_OF_MEMORY;
@@ -983,6 +971,7 @@ sc_pkcs15_bind_internal(struct sc_pkcs15_card *p15card, struct sc_aid *aid)
 	info = sc_find_app(card, aid);
 	if (info)   {
 		sc_log(ctx, "bind to application('%s',aid:'%s')", info->label, sc_dump_hex(info->aid.value, info->aid.len));
+		sc_pkcs15_free_app(p15card);
 		p15card->app = sc_dup_app_info(info);
 		if (!p15card->app)   {
 			err = SC_ERROR_OUT_OF_MEMORY;
@@ -1047,9 +1036,15 @@ sc_pkcs15_bind_internal(struct sc_pkcs15_card *p15card, struct sc_aid *aid)
 		sc_log(ctx, "EF(ODF) is empty");
 		goto end;
 	}
+	if (len > MAX_FILE_SIZE) {
+		sc_log(ctx, "EF(ODF) too large");
+		goto end;
+	}
 	buf = malloc(len);
-	if(buf == NULL)
-		LOG_FUNC_RETURN(ctx, SC_ERROR_OUT_OF_MEMORY);
+	if(buf == NULL) {
+		err = SC_ERROR_OUT_OF_MEMORY;
+		goto end;
+	}
 
 	err = -1; /* file state: not in cache */
 	if (p15card->opts.use_file_cache) {
@@ -1059,15 +1054,17 @@ sc_pkcs15_bind_internal(struct sc_pkcs15_card *p15card, struct sc_aid *aid)
 	}
 	if (err < 0) {
 		err = sc_read_binary(card, 0, buf, len, 0);
-		if (err < 0)   {
-			sc_log(ctx, "read EF(ODF) file error: %s", sc_strerror(err));
-			goto end;
-		}
-		else if (err < 2) {
-			err = SC_ERROR_PKCS15_APP_NOT_FOUND;
-			sc_log(ctx, "Invalid content of EF(ODF): %s", sc_strerror(err));
+		if (err < 2) {
+			if (err < 0) {
+				sc_log(ctx, "read EF(ODF) file error: %s", sc_strerror(err));
+			} else {
+				err = SC_ERROR_PKCS15_APP_NOT_FOUND;
+				sc_log(ctx, "Invalid content of EF(ODF): %s", sc_strerror(err));
+			}
 			goto end;
 		}
+		/* sc_read_binary may return less than requested */
+		len = err;
 
 		if (p15card->opts.use_file_cache) {
 			sc_pkcs15_cache_file(p15card, &tmppath, buf, len);
@@ -1113,9 +1110,15 @@ sc_pkcs15_bind_internal(struct sc_pkcs15_card *p15card, struct sc_aid *aid)
 		sc_log(ctx, "EF(TokenInfo) is empty");
 		goto end;
 	}
+	if (len > MAX_FILE_SIZE) {
+		sc_log(ctx, "EF(TokenInfo) too large");
+		goto end;
+	}
 	buf = malloc(len);
-	if(buf == NULL)
-		LOG_FUNC_RETURN(ctx, SC_ERROR_OUT_OF_MEMORY);
+	if(buf == NULL) {
+		err = SC_ERROR_OUT_OF_MEMORY;
+		goto end;
+	}
 
 	err = -1; /* file state: not in cache */
 	if (p15card->opts.use_file_cache) {
@@ -1125,20 +1128,21 @@ sc_pkcs15_bind_internal(struct sc_pkcs15_card *p15card, struct sc_aid *aid)
 	}
 	if (err < 0) {
 		err = sc_read_binary(card, 0, buf, len, 0);
-		if (err < 0)   {
-			sc_log(ctx, "read EF(TokenInfo) file error: %s", sc_strerror(err));
-			goto end;
-		}
 		if (err <= 2) {
-			err = SC_ERROR_PKCS15_APP_NOT_FOUND;
-			sc_log(ctx, "Invalid content of EF(TokenInfo): %s", sc_strerror(err));
+			if (err < 0)   {
+				sc_log(ctx, "read EF(TokenInfo) file error: %s", sc_strerror(err));
+			} else {
+				err = SC_ERROR_PKCS15_APP_NOT_FOUND;
+				sc_log(ctx, "Invalid content of EF(TokenInfo): %s", sc_strerror(err));
+			}
 			goto end;
 		}
+		/* sc_read_binary may return less than requested */
+		len = err;
 
 		if (p15card->opts.use_file_cache) {
 			sc_pkcs15_cache_file(p15card, &tmppath, buf, len);
 		}
-		err = len;
 	}
 
 	memset(&tokeninfo, 0, sizeof(tokeninfo));
@@ -1150,6 +1154,10 @@ sc_pkcs15_bind_internal(struct sc_pkcs15_card *p15card, struct sc_aid *aid)
 
 	*(p15card->tokeninfo) = tokeninfo;
 
+	if (!p15card->tokeninfo->serial_number && 0 == card->serialnr.len) {
+		sc_card_ctl(p15card->card, SC_CARDCTL_GET_SERIALNR, &card->serialnr);
+	}
+
 	if (!p15card->tokeninfo->serial_number && card->serialnr.len)   {
 		char *serial = calloc(1, card->serialnr.len*2 + 1);
 		size_t ii;
@@ -1185,16 +1193,19 @@ sc_pkcs15_bind(struct sc_card *card, struct sc_aid *aid,
 		struct sc_pkcs15_card **p15card_out)
 {
 	struct sc_pkcs15_card *p15card = NULL;
-	struct sc_context *ctx = card->ctx;
+	struct sc_context *ctx;
 	scconf_block *conf_block = NULL;
 	int r, emu_first, enable_emu;
+	const char *private_certificate;
+
+	if (card == NULL || p15card_out == NULL) {
+		return SC_ERROR_INVALID_ARGUMENTS;
+	}
+	ctx = card->ctx;
 
 	LOG_FUNC_CALLED(ctx);
 	sc_log(ctx, "application(aid:'%s')", aid ? sc_dump_hex(aid->value, aid->len) : "empty");
 
-	if (p15card_out == NULL) {
-		return SC_ERROR_INVALID_ARGUMENTS;
-	}
 	p15card = sc_pkcs15_card_new();
 	if (p15card == NULL)
 		LOG_FUNC_RETURN(ctx, SC_ERROR_OUT_OF_MEMORY);
@@ -1204,19 +1215,33 @@ sc_pkcs15_bind(struct sc_card *card, struct sc_aid *aid,
 	p15card->opts.use_pin_cache = 1;
 	p15card->opts.pin_cache_counter = 10;
 	p15card->opts.pin_cache_ignore_user_consent = 0;
+	if(0 == strcmp(ctx->app_name, "tokend")) {
+		private_certificate = "ignore";
+		p15card->opts.private_certificate = SC_PKCS15_CARD_OPTS_PRIV_CERT_IGNORE;
+	} else {
+		private_certificate = "protect";
+		p15card->opts.private_certificate = SC_PKCS15_CARD_OPTS_PRIV_CERT_PROTECT;
+	}
 
 	conf_block = sc_get_conf_block(ctx, "framework", "pkcs15", 1);
-
 	if (conf_block) {
 		p15card->opts.use_file_cache = scconf_get_bool(conf_block, "use_file_caching", p15card->opts.use_file_cache);
 		p15card->opts.use_pin_cache = scconf_get_bool(conf_block, "use_pin_caching", p15card->opts.use_pin_cache);
 		p15card->opts.pin_cache_counter = scconf_get_int(conf_block, "pin_cache_counter", p15card->opts.pin_cache_counter);
-		p15card->opts.pin_cache_ignore_user_consent =  scconf_get_bool(conf_block, "pin_cache_ignore_user_consent",
+		p15card->opts.pin_cache_ignore_user_consent = scconf_get_bool(conf_block, "pin_cache_ignore_user_consent",
 				p15card->opts.pin_cache_ignore_user_consent);
+		private_certificate = scconf_get_str(conf_block, "private_certificate", private_certificate);
 	}
-	sc_log(ctx, "PKCS#15 options: use_file_cache=%d use_pin_cache=%d pin_cache_counter=%d pin_cache_ignore_user_consent=%d",
+	if (0 == strcmp(private_certificate, "protect")) {
+		p15card->opts.private_certificate = SC_PKCS15_CARD_OPTS_PRIV_CERT_PROTECT;
+	} else if (0 == strcmp(private_certificate, "ignore")) {
+		p15card->opts.private_certificate = SC_PKCS15_CARD_OPTS_PRIV_CERT_IGNORE;
+	} else if (0 == strcmp(private_certificate, "declassify")) {
+		p15card->opts.private_certificate = SC_PKCS15_CARD_OPTS_PRIV_CERT_DECLASSIFY;
+	}
+	sc_log(ctx, "PKCS#15 options: use_file_cache=%d use_pin_cache=%d pin_cache_counter=%d pin_cache_ignore_user_consent=%d private_certificate=%d",
 			p15card->opts.use_file_cache, p15card->opts.use_pin_cache,p15card->opts.pin_cache_counter,
-			p15card->opts.pin_cache_ignore_user_consent);
+			p15card->opts.pin_cache_ignore_user_consent, p15card->opts.private_certificate);
 
 	r = sc_lock(card);
 	if (r) {
@@ -1251,8 +1276,6 @@ sc_pkcs15_bind(struct sc_card *card, struct sc_aid *aid,
 			goto error;
 	}
 done:
-	fix_starcos_pkcs15_card(p15card);
-
 	*p15card_out = p15card;
 	sc_unlock(card);
 	LOG_FUNC_RETURN(ctx, SC_SUCCESS);
@@ -1295,12 +1318,13 @@ __sc_pkcs15_search_objects(struct sc_pkcs15_card *p15card, unsigned int class_ma
 
 	/* Make sure the class mask we have makes sense */
 	if (class_mask == 0
-	 || (class_mask & ~(SC_PKCS15_SEARCH_CLASS_PRKEY |
-			    SC_PKCS15_SEARCH_CLASS_PUBKEY |
-			    SC_PKCS15_SEARCH_CLASS_SKEY |
-			    SC_PKCS15_SEARCH_CLASS_CERT |
-			    SC_PKCS15_SEARCH_CLASS_DATA |
-			    SC_PKCS15_SEARCH_CLASS_AUTH))) {
+			|| (class_mask & ~(
+					SC_PKCS15_SEARCH_CLASS_PRKEY |
+					SC_PKCS15_SEARCH_CLASS_PUBKEY |
+					SC_PKCS15_SEARCH_CLASS_SKEY |
+					SC_PKCS15_SEARCH_CLASS_CERT |
+					SC_PKCS15_SEARCH_CLASS_DATA |
+					SC_PKCS15_SEARCH_CLASS_AUTH))) {
 		LOG_FUNC_RETURN(p15card->card->ctx, SC_ERROR_INVALID_ARGUMENTS);
 	}
 
@@ -1869,6 +1893,11 @@ sc_pkcs15_free_object(struct sc_pkcs15_object *obj)
 		sc_pkcs15_free_prkey_info((sc_pkcs15_prkey_info_t *)obj->data);
 		break;
 	case SC_PKCS15_TYPE_PUBKEY:
+		/* This is normally passed to framework-pkcs15,
+		 * but if something fails on the way, it would not get freed */
+		if (obj->emulated) {
+			sc_pkcs15_free_pubkey(obj->emulated);
+		}
 		sc_pkcs15_free_pubkey_info((sc_pkcs15_pubkey_info_t *)obj->data);
 		break;
 	case SC_PKCS15_TYPE_CERT:
@@ -1985,6 +2014,8 @@ sc_pkcs15_encode_df(struct sc_context *ctx, struct sc_pkcs15_card *p15card, stru
 			free(buf);
 			return r;
 		}
+		if (!tmpsize)
+			continue;
 		p = (u8 *) realloc(buf, bufsize + tmpsize);
 		if (!p) {
 			free(tmp);
@@ -2240,7 +2271,7 @@ sc_pkcs15_parse_unusedspace(const unsigned char *buf, size_t buflen, struct sc_p
 	const unsigned char *p = buf;
 	size_t left = buflen;
 	int r;
-	struct sc_path path, dummy_path;
+	struct sc_path path;
 	struct sc_pkcs15_id auth_id;
 	struct sc_asn1_entry asn1_unusedspace[] = {
 		{ "UnusedSpace", SC_ASN1_STRUCT, SC_ASN1_TAG_SEQUENCE | SC_ASN1_CONS, 0, NULL, NULL },
@@ -2255,9 +2286,6 @@ sc_pkcs15_parse_unusedspace(const unsigned char *buf, size_t buflen, struct sc_p
 	/* Clean the list if already present */
 	sc_pkcs15_free_unusedspace(p15card);
 
-	sc_format_path("3F00", &dummy_path);
-	dummy_path.index = dummy_path.count = 0;
-
 	sc_format_asn1_entry(asn1_unusedspace, asn1_unusedspace_values, NULL, 1);
 	sc_format_asn1_entry(asn1_unusedspace_values, &path, NULL, 1);
 	sc_format_asn1_entry(asn1_unusedspace_values+1, &auth_id, NULL, 0);
@@ -2272,7 +2300,7 @@ sc_pkcs15_parse_unusedspace(const unsigned char *buf, size_t buflen, struct sc_p
 		/* If the path length is 0, it's a dummy path then don't add it.
 		 * If the path length isn't included (-1) then it's against the standard
 		 *   but we'll just ignore it instead of returning an error. */
-		if (path.count > 0) {
+		if (path.count > 0 && p15card->file_app) {
 			r = sc_pkcs15_make_absolute_path(&p15card->file_app->path, &path);
 			if (r < 0)
 				return r;
@@ -2321,7 +2349,8 @@ sc_pkcs15_read_file(struct sc_pkcs15_card *p15card, const struct sc_path *in_pat
 
 	if (r) {
 		r = sc_lock(p15card->card);
-		LOG_TEST_RET(ctx, r, "sc_lock() failed");
+		if (r)
+			goto fail;
 		r = sc_select_file(p15card->card, in_path, &file);
 		if (r)
 			goto fail_unlock;
@@ -2392,7 +2421,7 @@ sc_pkcs15_read_file(struct sc_pkcs15_card *p15card, const struct sc_path *in_pat
 
 		sc_file_free(file);
 
-		if (p15card->opts.use_file_cache) {
+		if (len && p15card->opts.use_file_cache) {
 			sc_pkcs15_cache_file(p15card, in_path, data, len);
 		}
 	}
@@ -2401,9 +2430,10 @@ sc_pkcs15_read_file(struct sc_pkcs15_card *p15card, const struct sc_path *in_pat
 	LOG_FUNC_RETURN(ctx, SC_SUCCESS);
 
 fail_unlock:
+	sc_unlock(p15card->card);
+fail:
 	free(data);
 	sc_file_free(file);
-	sc_unlock(p15card->card);
 	LOG_FUNC_RETURN(ctx, r);
 }
 
@@ -2500,8 +2530,15 @@ sc_pkcs15_make_absolute_path(const struct sc_path *parent, struct sc_path *child
 void sc_pkcs15_free_object_content(struct sc_pkcs15_object *obj)
 {
 	if (obj->content.value && obj->content.len)   {
-		sc_mem_clear(obj->content.value, obj->content.len);
-		free(obj->content.value);
+		if (SC_PKCS15_TYPE_AUTH & obj->type
+			|| SC_PKCS15_TYPE_SKEY & obj->type
+			|| SC_PKCS15_TYPE_PRKEY & obj->type) {
+			/* clean everything that potentially contains a secret */
+			sc_mem_clear(obj->content.value, obj->content.len);
+			sc_mem_secure_free(obj->content.value, obj->content.len);
+		} else {
+			free(obj->content.value);
+		}
 	}
 	obj->content.value = NULL;
 	obj->content.len = 0;
@@ -2525,7 +2562,13 @@ sc_pkcs15_allocate_object_content(struct sc_context *ctx, struct sc_pkcs15_objec
 	/* Need to pass by temporary variable,
 	 * because 'value' and 'content.value' pointers can be the sames.
 	 */
-	tmp_buf = (unsigned char *)sc_mem_alloc_secure(ctx, len);
+	if (SC_PKCS15_TYPE_AUTH & obj->type
+			|| SC_PKCS15_TYPE_SKEY & obj->type
+			|| SC_PKCS15_TYPE_PRKEY & obj->type) {
+		tmp_buf = sc_mem_secure_alloc(len);
+	} else {
+		tmp_buf = malloc(len);
+	}
 	if (!tmp_buf)
 		return SC_ERROR_OUT_OF_MEMORY;
 
@@ -2561,13 +2604,38 @@ sc_pkcs15_get_supported_algo(struct sc_pkcs15_card *p15card, unsigned operation,
 	return info;
 }
 
+struct sc_supported_algo_info *
+sc_pkcs15_get_specific_supported_algo(struct sc_pkcs15_card *p15card, unsigned operation, unsigned mechanism, const struct sc_object_id *algo_oid)
+{
+	struct sc_context *ctx = p15card->card->ctx;
+	struct sc_supported_algo_info *info = NULL;
+	int ii;
+
+	if (algo_oid == NULL)
+		return NULL;
+
+	for (ii=0;ii<SC_MAX_SUPPORTED_ALGORITHMS && p15card->tokeninfo->supported_algos[ii].reference; ii++)
+		if ((p15card->tokeninfo->supported_algos[ii].operations & operation)
+				&& (p15card->tokeninfo->supported_algos[ii].mechanism == mechanism)
+				&& sc_compare_oid(algo_oid, &p15card->tokeninfo->supported_algos[ii].algo_id) == 1)
+			break;
+
+	if (ii < SC_MAX_SUPPORTED_ALGORITHMS && p15card->tokeninfo->supported_algos[ii].reference)   {
+		info = &p15card->tokeninfo->supported_algos[ii];
+		sc_log(ctx, "found supported algorithm (ref:%X,mech:%X,ops:%X,algo_ref:%X)",
+				info->reference, info->mechanism, info->operations, info->algo_ref);
+	}
+
+	return info;
+}
+
 int
 sc_pkcs15_get_generalized_time(struct sc_context *ctx, char **out)
 {
 #ifdef HAVE_GETTIMEOFDAY
 	struct timeval tv;
 #endif
-	struct tm *tm_time;
+	struct tm tm;
 	time_t t;
 
 	if (!ctx || !out)
@@ -2580,16 +2648,21 @@ sc_pkcs15_get_generalized_time(struct sc_context *ctx, char **out)
 #else
 	t = time(NULL);
 #endif
-	tm_time = gmtime(&t);
-	if (!tm_time)
-		LOG_TEST_RET(ctx, SC_ERROR_INTERNAL, "gmtime failed");
+
+#ifdef _WIN32
+	if (0 != gmtime_s(&tm, &t))
+		LOG_FUNC_RETURN(ctx, SC_ERROR_INTERNAL);
+#else
+	if (NULL == gmtime_r(&t, &tm))
+		LOG_FUNC_RETURN(ctx, SC_ERROR_INTERNAL);
+#endif
 
 	*out = calloc(1, 16);
 	if (*out == NULL)
 		LOG_TEST_RET(ctx, SC_ERROR_OUT_OF_MEMORY, "memory failure");
 
 	/* print time in generalized time format */
-	if (!strftime(*out, 16, "%Y%m%d%H%M%SZ", tm_time)) {
+	if (!strftime(*out, 16, "%Y%m%d%H%M%SZ", &tm)) {
 		free(*out);
 		LOG_TEST_RET(ctx, SC_ERROR_INTERNAL, "strftime failed");
 	}
