@@ -1,7 +1,7 @@
 /*
  * reader-escape.c: implementation related to escape commands with pseudo APDUs
  *
- * Copyright (C) 2013-2015  Frank Morgner
+ * Copyright (C) 2013-2018  Frank Morgner
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -40,23 +40,39 @@
 #ifndef _BSD_SOURCE
 #define _BSD_SOURCE             /* See feature_test_macros(7) */
 #endif
+#ifdef HAVE_SYS_ENDIAN_H
+#include <sys/endian.h>
+#endif
+#ifdef HAVE_ENDIAN_H
 #include <endian.h>
 #endif
+#endif
 
-static const u8 escape_cla                          = 0xff;
-static const u8 escape_ins                          = 0x9a;
+int get_pace_capabilities(u8 *bitmap)
+{
+	if (!bitmap)
+		return SC_ERROR_INVALID_ARGUMENTS;
 
-static const u8 escape_p1_PIN                       = 0x04;
-static const u8 escape_p2_GetReaderPACECapabilities = 0x01;
-static const u8 escape_p2_EstablishPACEChannel      = 0x02;
-/*static const u8 escape_p2_DestroyPACEChannel        = 0x03;*/
-static const u8 escape_p2_PC_to_RDR_Secure          = 0x10;
+	/* BitMap */
+	*bitmap = EAC_BITMAP_PACE|EAC_BITMAP_EID|EAC_BITMAP_ESIGN;
 
-static const u8 escape_p1_IFD                       = 0x01;
-static const u8 escape_p2_vendor                    = 0x01;
-/*static const u8 escape_p2_product                   = 0x03;*/
-static const u8 escape_p2_version_firmware          = 0x06;
-/*static const u8 escape_p2_version_driver            = 0x07;*/
+	return SC_SUCCESS;
+}
+
+const u8 escape_cla                          = 0xff;
+const u8 escape_ins                          = 0x9a;
+
+const u8 escape_p1_PIN                       = 0x04;
+const u8 escape_p2_GetReaderPACECapabilities = 0x01;
+const u8 escape_p2_EstablishPACEChannel      = 0x02;
+/*const u8 escape_p2_DestroyPACEChannel        = 0x03;*/
+const u8 escape_p2_PC_to_RDR_Secure          = 0x10;
+
+const u8 escape_p1_IFD                       = 0x01;
+const u8 escape_p2_vendor                    = 0x01;
+/*const u8 escape_p2_product                   = 0x03;*/
+const u8 escape_p2_version_firmware          = 0x06;
+/*const u8 escape_p2_version_driver            = 0x07;*/
 
 struct sc_asn1_entry g_boolean[] = {
 	{ "boolean",
@@ -79,7 +95,7 @@ struct sc_asn1_entry g_numeric_string_as_octet_string[] = {
 	{ NULL , 0 , 0 , 0 , NULL , NULL }
 };
 
-static const struct sc_asn1_entry g_EstablishPACEChannelInput_data[] = {
+const struct sc_asn1_entry g_EstablishPACEChannelInput_data[] = {
 	{ "passwordID",
 		/* use an OCTET STRING to avoid a conversion to int */
 		SC_ASN1_STRUCT, SC_ASN1_CTX|0x01|SC_ASN1_CONS, 0, NULL, NULL },
@@ -94,7 +110,7 @@ static const struct sc_asn1_entry g_EstablishPACEChannelInput_data[] = {
 		SC_ASN1_STRUCT, SC_ASN1_CTX|0x05|SC_ASN1_CONS, SC_ASN1_OPTIONAL|SC_ASN1_ALLOC, NULL, NULL },
 	{ NULL , 0 , 0 , 0 , NULL , NULL }
 };
-static const struct sc_asn1_entry g_EstablishPACEChannelOutput_data[] = {
+const struct sc_asn1_entry g_EstablishPACEChannelOutput_data[] = {
 	{ "errorCode",
 		SC_ASN1_STRUCT, SC_ASN1_CTX|0x01|SC_ASN1_CONS, 0, NULL, NULL },
 	{ "statusMSESetAT",
@@ -109,7 +125,7 @@ static const struct sc_asn1_entry g_EstablishPACEChannelOutput_data[] = {
 		SC_ASN1_STRUCT, SC_ASN1_CTX|0x06|SC_ASN1_CONS, SC_ASN1_OPTIONAL|SC_ASN1_ALLOC, NULL, NULL },
 	{ NULL , 0 , 0 , 0 , NULL , NULL }
 };
-static const struct sc_asn1_entry g_EstablishPACEChannel[] = {
+const struct sc_asn1_entry g_EstablishPACEChannel[] = {
 	{ "EstablishPACEChannel",
 		SC_ASN1_STRUCT, SC_ASN1_TAG_SEQUENCE|SC_ASN1_CONS, 0, NULL, NULL },
 	{ NULL , 0 , 0 , 0 , NULL , NULL }
@@ -494,7 +510,7 @@ static int escape_pin_cmd_to_buf(sc_context_t *ctx,
 		/* GLP PIN length is encoded in 4 bits and block size is always 8 bytes */
 		bmPINBlockString = 0x40 | 0x08;
 	} else if (pin_ref->encoding == SC_PIN_ENCODING_ASCII && data->flags & SC_PIN_CMD_NEED_PADDING) {
-		bmPINBlockString = pin_ref->pad_length;
+		bmPINBlockString = (uint8_t) pin_ref->pad_length;
 	} else {
 		bmPINBlockString = 0x00;
 	}
@@ -566,11 +582,11 @@ static int escape_pin_cmd_to_buf(sc_context_t *ctx,
 			modify->bmPINLengthFormat = bmPINLengthFormat;
 			if (!(data->flags & SC_PIN_CMD_IMPLICIT_CHANGE)
 					&& data->pin1.offset) {
-				modify->bInsertionOffsetOld = data->pin1.offset - 5;
+				modify->bInsertionOffsetOld = (uint8_t) data->pin1.offset - 5;
 			} else {
 				modify->bInsertionOffsetOld = 0;
 			}
-			modify->bInsertionOffsetNew = data->pin2.offset ? data->pin2.offset - 5 : 0;
+			modify->bInsertionOffsetNew = data->pin2.offset ? (uint8_t) data->pin2.offset - 5 : 0;
 			modify->wPINMaxExtraDigit = wPINMaxExtraDigit;
 			modify->bConfirmPIN = CCID_PIN_CONFIRM_NEW
 				| (data->flags & SC_PIN_CMD_IMPLICIT_CHANGE ? 0 : CCID_PIN_INSERT_OLD);
@@ -653,23 +669,21 @@ static int escape_perform_verify(struct sc_reader *reader,
 	r = escape_pin_cmd_to_buf(reader->ctx, data,
 			(unsigned char **) &apdu.data, &apdu.datalen);
 	if (r < 0) {
-		sc_debug(reader->ctx, SC_LOG_DEBUG_NORMAL,
+		sc_log(reader->ctx, 
 				"Error encoding PC_to_RDR_Secure");
 		goto err;
 	}
 	apdu.lc = apdu.datalen;
 
-	r = SC_SUCCESS;
-
 	r = reader->ops->transmit(reader, &apdu);
 	if (r < 0) {
-		sc_debug(reader->ctx, SC_LOG_DEBUG_NORMAL,
+		sc_log(reader->ctx, 
 				"Error performing PC_to_RDR_Secure");
 		goto err;
 	}
 
 	if (apdu.sw1 != 0x90 && apdu.sw2 != 0x00) {
-		sc_debug(reader->ctx, SC_LOG_DEBUG_NORMAL,
+		sc_log(reader->ctx, 
 				"Error decoding PC_to_RDR_Secure");
 		r = SC_ERROR_NOT_SUPPORTED;
 		goto err;
@@ -714,7 +728,7 @@ static int escape_perform_pace(struct sc_reader *reader,
 	r = escape_pace_input_to_buf(reader->ctx, input,
 			(unsigned char **) &apdu.data, &apdu.datalen);
 	if (r < 0) {
-		sc_debug(reader->ctx, SC_LOG_DEBUG_NORMAL,
+		sc_log(reader->ctx, 
 				"Error encoding EstablishPACEChannel");
 		goto err;
 	}
@@ -722,13 +736,13 @@ static int escape_perform_pace(struct sc_reader *reader,
 
 	r = reader->ops->transmit(reader, &apdu);
 	if (r < 0) {
-		sc_debug(reader->ctx, SC_LOG_DEBUG_NORMAL,
+		sc_log(reader->ctx, 
 				"Error performing EstablishPACEChannel");
 		goto err;
 	}
 
 	if (apdu.sw1 != 0x90 && apdu.sw2 != 0x00) {
-		sc_debug(reader->ctx, SC_LOG_DEBUG_NORMAL,
+		sc_log(reader->ctx, 
 				"Error decoding EstablishPACEChannel");
 		r = SC_ERROR_NOT_SUPPORTED;
 		goto err;
@@ -936,7 +950,7 @@ void sc_detect_escape_cmds(sc_reader_t *reader)
 					&& !(reader->capabilities & SC_READER_CAP_PIN_PAD)) {
 				((struct sc_reader_operations *) reader->ops)->perform_verify =
 					escape_perform_verify;
-				sc_debug(reader->ctx, SC_LOG_DEBUG_NORMAL,
+				sc_log(reader->ctx, 
 						"Added escape command wrappers for PIN verification/modification to '%s'", reader->name);
 			}
 
@@ -944,14 +958,14 @@ void sc_detect_escape_cmds(sc_reader_t *reader)
 					&& !(reader->capabilities & SC_READER_CAP_PACE_GENERIC)) {
 				((struct sc_reader_operations *) reader->ops)->perform_pace =
 					escape_perform_pace;
-				sc_debug(reader->ctx, SC_LOG_DEBUG_NORMAL,
+				sc_log(reader->ctx, 
 						"Added escape command wrappers for PACE to '%s'", reader->name);
 			}
 
 			reader->capabilities |= capabilities;
 		} else {
 			error++;
-			sc_debug(reader->ctx, SC_LOG_DEBUG_NORMAL,
+			sc_log(reader->ctx, 
 					"%s does not support escape commands", reader->name);
 		}
 
@@ -988,7 +1002,7 @@ void sc_detect_escape_cmds(sc_reader_t *reader)
 	}
 
 	if (error && reader) {
-		sc_debug(reader->ctx, SC_LOG_DEBUG_NORMAL,
+		sc_log(reader->ctx, 
 				"%d escape command%s failed, need to reset the card",
 				error, error == 1 ? "" : "s");
 		if (reader->ops && reader->ops->transmit) {
