@@ -35,6 +35,12 @@
 #endif /* PKCS11_THREAD_LOCKING */
 
 #include "sc-pkcs11.h"
+#include "ui/notify.h"
+
+#include "libopensc/sc-ossl-compat.h"
+#ifdef ENABLE_OPENPACE
+#include <eac/eac.h>
+#endif
 
 #ifndef MODULE_APP_NAME
 #define MODULE_APP_NAME "opensc-pkcs11"
@@ -209,7 +215,52 @@ static int slot_list_seeker(const void *el, const void *key) {
 	return 0;
 }
 
+#ifndef _WIN32
+__attribute__((constructor))
+#endif
+int module_init()
+{
+	sc_notify_init();
+	return 1;
+}
 
+#ifndef _WIN32
+__attribute__((destructor))
+#endif
+int module_close()
+{
+	sc_notify_close();
+#if defined(ENABLE_OPENSSL) && defined(OPENSSL_SECURE_MALLOC_SIZE)
+	CRYPTO_secure_malloc_done();
+#endif
+#ifdef ENABLE_OPENPACE
+	EAC_cleanup();
+#endif
+	return 1;
+}
+
+#ifdef _WIN32
+BOOL APIENTRY DllMain( HINSTANCE hinstDLL,
+	DWORD  ul_reason_for_call,
+	LPVOID lpReserved
+)
+{
+	switch (ul_reason_for_call)
+	{
+	case DLL_PROCESS_ATTACH:
+		if (!module_init())
+			return FALSE;
+		break;
+	case DLL_PROCESS_DETACH:
+		if (lpReserved == NULL) {
+			if (!module_close())
+				return FALSE;
+		}
+		break;
+	}
+	return TRUE;
+}
+#endif
 
 CK_RV C_Initialize(CK_VOID_PTR pInitArgs)
 {
@@ -218,7 +269,6 @@ CK_RV C_Initialize(CK_VOID_PTR pInitArgs)
 	pid_t current_pid = getpid();
 #endif
 	int rc;
-	unsigned int i;
 	sc_context_param_t ctx_opts;
 
 #if !defined(_WIN32)
@@ -264,16 +314,13 @@ CK_RV C_Initialize(CK_VOID_PTR pInitArgs)
 	list_attributes_seeker(&sessions, session_list_seeker);
 
 	/* List of slots */
-	list_init(&virtual_slots);
 	if (0 != list_init(&virtual_slots)) {
 		rv = CKR_HOST_MEMORY;
 		goto out;
 	}
 	list_attributes_seeker(&virtual_slots, slot_list_seeker);
 
-	/* Create slots for readers found on initialization, only if in 2.11 mode */
-	for (i=0; i<sc_ctx_get_reader_count(context); i++)
-			initialize_reader(sc_ctx_get_reader(context, i));
+	card_detect_all();
 
 out:
 	if (context != NULL)
@@ -301,6 +348,10 @@ CK_RV C_Finalize(CK_VOID_PTR pReserved)
 	if (pReserved != NULL_PTR)
 		return CKR_ARGUMENTS_BAD;
 
+#if !defined(_WIN32)
+	sc_notify_close();
+#endif
+
 	if (context == NULL)
 		return CKR_CRYPTOKI_NOT_INITIALIZED;
 
@@ -395,13 +446,23 @@ CK_RV C_GetSlotList(CK_BBOOL       tokenPresent,  /* only slots with token prese
 
 	sc_log(context, "C_GetSlotList(token=%d, %s)", tokenPresent,
 			pSlotList==NULL_PTR? "plug-n-play":"refresh");
+	DEBUG_VSS(NULL, "C_GetSlotList before ctx_detect_detect");
 
 	/* Slot list can only change in v2.20 */
 	if (pSlotList == NULL_PTR)
 		sc_ctx_detect_readers(context);
 
+	DEBUG_VSS(NULL, "C_GetSlotList after ctx_detect_readers");
+
 	card_detect_all();
 
+	if (list_empty(&virtual_slots)) {
+		sc_log(context, "returned 0 slots\n");
+		*pulCount = 0;
+		rv = CKR_OK;
+		goto out;
+	}
+
 	found = calloc(list_size(&virtual_slots), sizeof(CK_SLOT_ID));
 
 	if (found == NULL) {
@@ -414,13 +475,11 @@ CK_RV C_GetSlotList(CK_BBOOL       tokenPresent,  /* only slots with token prese
 	for (i=0; i<list_size(&virtual_slots); i++) {
 		slot = (sc_pkcs11_slot_t *) list_get_at(&virtual_slots, i);
 		/* the list of available slots contains:
-		 * - if present, virtual hotplug slot;
+		 * - without token(s), at least one empty slot per reader;
 		 * - any slot with token;
-		 * - without token(s), one empty slot per reader;
 		 * - any slot that has already been seen;
 		 */
-		if ((!tokenPresent && !slot->reader)
-				|| (!tokenPresent && slot->reader != prev_reader)
+		if ((!tokenPresent && slot->reader != prev_reader)
 				|| (slot->slot_info.flags & CKF_TOKEN_PRESENT)
 				|| (slot->flags & SC_PKCS11_SLOT_FLAG_SEEN)) {
 			found[numMatches++] = slot->id;
@@ -428,6 +487,7 @@ CK_RV C_GetSlotList(CK_BBOOL       tokenPresent,  /* only slots with token prese
 		}
 		prev_reader = slot->reader;
 	}
+	DEBUG_VSS(NULL, "C_GetSlotList after card_detect_all");
 
 	if (pSlotList == NULL_PTR) {
 		sc_log(context, "was only a size inquiry (%lu)\n", numMatches);
@@ -435,6 +495,7 @@ CK_RV C_GetSlotList(CK_BBOOL       tokenPresent,  /* only slots with token prese
 		rv = CKR_OK;
 		goto out;
 	}
+	DEBUG_VSS(NULL, "C_GetSlotList after slot->id reassigned");
 
 	if (*pulCount < numMatches) {
 		sc_log(context, "buffer was too small (needed %lu)\n", numMatches);
@@ -448,12 +509,10 @@ CK_RV C_GetSlotList(CK_BBOOL       tokenPresent,  /* only slots with token prese
 	rv = CKR_OK;
 
 	sc_log(context, "returned %lu slots\n", numMatches);
+	DEBUG_VSS(NULL, "Returning a new slot list");
 
 out:
-	if (found != NULL) {
-		free (found);
-		found = NULL;
-	}
+	free (found);
 	sc_pkcs11_unlock();
 	return rv;
 }
@@ -487,7 +546,7 @@ static sc_timestamp_t get_current_time(void)
 
 CK_RV C_GetSlotInfo(CK_SLOT_ID slotID, CK_SLOT_INFO_PTR pInfo)
 {
-	struct sc_pkcs11_slot *slot;
+	struct sc_pkcs11_slot *slot = NULL;
 	sc_timestamp_t now;
 	CK_RV rv;
 
@@ -501,7 +560,7 @@ CK_RV C_GetSlotInfo(CK_SLOT_ID slotID, CK_SLOT_INFO_PTR pInfo)
 	sc_log(context, "C_GetSlotInfo(0x%lx)", slotID);
 
 	if (sc_pkcs11_conf.init_sloppy) {
-		/* Most likely virtual_slots only contains the hotplug slot and has not
+		/* Most likely virtual_slots is empty and has not
 		 * been initialized because the caller has *not* called C_GetSlotList
 		 * before C_GetSlotInfo, as required by PKCS#11.  Initialize
 		 * virtual_slots to make things work and hope the caller knows what
@@ -510,12 +569,12 @@ CK_RV C_GetSlotInfo(CK_SLOT_ID slotID, CK_SLOT_INFO_PTR pInfo)
 	}
 
 	rv = slot_get_slot(slotID, &slot);
-	sc_log(context, "C_GetSlotInfo() get slot rv %lu", rv);
-	if (rv == CKR_OK)   {
-		if (slot->reader == NULL)   {
+	DEBUG_VSS(slot, "C_GetSlotInfo found");
+	sc_log(context, "C_GetSlotInfo() get slot rv %s", lookup_enum( RV_T, rv));
+	if (rv == CKR_OK) {
+		if (slot->reader == NULL) {
 			rv = CKR_TOKEN_NOT_PRESENT;
-		}
-		else {
+		} else {
 			now = get_current_time();
 			if (now >= slot->slot_state_expires || now == 0) {
 				/* Update slot status */
@@ -671,13 +730,6 @@ again:
 	sc_log(context, "C_WaitForSlotEvent() reader_states:%p", reader_states);
 	sc_pkcs11_unlock();
 	r = sc_wait_for_event(context, mask, &found, &events, -1, &reader_states);
-	if (events & SC_EVENT_READER_ATTACHED) {
-		rv = sc_pkcs11_lock();
-		if (rv != CKR_OK)
-			return rv;
-
-		goto out;
-	}
 	/* Was C_Finalize called ? */
 	if (in_finalize == 1)
 		return CKR_CRYPTOKI_NOT_INITIALIZED;
@@ -753,7 +805,7 @@ sc_pkcs11_init_lock(CK_C_INITIALIZE_ARGS_PTR args)
 		/* Shall be used in threaded environment, must use operating system locking */
 		global_locking = default_mutex_funcs;
 	} else if (applock && !oslock) {
-		/* Shall be used in threaded envirnoment, must use app provided locking */
+		/* Shall be used in threaded environment, must use app provided locking */
 		global_locking = args;
 	} else if (!applock && !oslock) {
 		/* Shall not be used in threaded environment, use operating system locking */
