@@ -33,12 +33,10 @@
 
 #define CRYPTOKI_EXPORTS
 #include "pkcs11-display.h"
+#include "common/libpkcs11.h"
 
 #define __PASTE(x,y)      x##y
 
-extern void *C_LoadModule(const char *name, CK_FUNCTION_LIST_PTR_PTR);
-extern CK_RV C_UnloadModule(void *module);
-
 /* Declare all spy_* Cryptoki function */
 
 /* Spy Module Function List */
@@ -50,12 +48,12 @@ static void *modhandle = NULL;
 /* Spy module output */
 static FILE *spy_output = NULL;
 
-/* Inits the spy. If successfull, po != NULL */
+/* Inits the spy. If successful, po != NULL */
 static CK_RV
 init_spy(void)
 {
 	const char *output, *module;
-	int rv = CKR_OK;
+	CK_RV rv = CKR_OK;
 #ifdef _WIN32
         char temp_path[PATH_MAX], expanded_path[PATH_MAX];
         DWORD temp_len, expanded_len;
@@ -153,7 +151,7 @@ init_spy(void)
 
 #ifdef _WIN32
 	if (!spy_output) {
-		/* try for the machine version first, as we may be runing
+		/* try for the machine version first, as we may be running
 		 * without a user during login
 		 */
 		rc = RegOpenKeyEx( HKEY_LOCAL_MACHINE, "Software\\OpenSC Project\\PKCS11-Spy", 0, KEY_QUERY_VALUE, &hKey );
@@ -172,7 +170,7 @@ init_spy(void)
 				}
 			}
 
-	                if( (rc == ERROR_SUCCESS) && (temp_len < PATH_MAX) )
+			if( (rc == ERROR_SUCCESS) && (temp_len < PATH_MAX) )
 				output = temp_path;
 			RegCloseKey( hKey );
 		}
@@ -188,7 +186,7 @@ init_spy(void)
 	module = getenv("PKCS11SPY");
 #ifdef _WIN32
 	if (!module) {
-		/* try for the machine version first, as we may be runing
+		/* try for the machine version first, as we may be running
 		 * without a user during login
 		 */
 		rc = RegOpenKeyEx( HKEY_LOCAL_MACHINE, "Software\\OpenSC Project\\PKCS11-Spy",
@@ -786,6 +784,26 @@ C_EncryptInit(CK_SESSION_HANDLE hSession, CK_MECHANISM_PTR pMechanism, CK_OBJECT
 	enter("C_EncryptInit");
 	spy_dump_ulong_in("hSession", hSession);
 	fprintf(spy_output, "pMechanism->type=%s\n", lookup_enum(MEC_T, pMechanism->mechanism));
+	switch (pMechanism->mechanism) {
+	case CKM_AES_GCM:
+		if (pMechanism->pParameter != NULL) {
+			CK_GCM_PARAMS *param =
+				(CK_GCM_PARAMS *) pMechanism->pParameter;
+			spy_dump_string_in("pIv[ulIvLen]",
+				param->pIv, param->ulIvLen);
+			spy_dump_ulong_in("ulIvBits", param->ulIvBits);
+			spy_dump_string_in("pAAD[ulAADLen]",
+				param->pAAD, param->ulAADLen);
+			fprintf(spy_output, "pMechanism->pParameter->ulTagBits=%lu\n", param->ulTagBits);
+		} else {
+			fprintf(spy_output, "Parameters block for %s is empty...\n",
+				lookup_enum(MEC_T, pMechanism->mechanism));
+		}
+		break;
+	default:
+		spy_dump_string_in("pParameter[ulParameterLen]", pMechanism->pParameter, pMechanism->ulParameterLen);
+		break;
+	}
 	spy_dump_ulong_in("hKey", hKey);
 	rv = po->C_EncryptInit(hSession, pMechanism, hKey);
 	return retne(rv);
@@ -844,6 +862,27 @@ C_DecryptInit(CK_SESSION_HANDLE hSession, CK_MECHANISM_PTR pMechanism, CK_OBJECT
 	enter("C_DecryptInit");
 	spy_dump_ulong_in("hSession", hSession);
 	fprintf(spy_output, "pMechanism->type=%s\n", lookup_enum(MEC_T, pMechanism->mechanism));
+	switch (pMechanism->mechanism) {
+	case CKM_RSA_PKCS_OAEP:
+		if (pMechanism->pParameter != NULL) {
+ 			CK_RSA_PKCS_OAEP_PARAMS *param =
+				(CK_RSA_PKCS_OAEP_PARAMS *) pMechanism->pParameter;
+			fprintf(spy_output, "pMechanism->pParameter->hashAlg=%s\n",
+				lookup_enum(MEC_T, param->hashAlg));
+			fprintf(spy_output, "pMechanism->pParameter->mgf=%s\n",
+				lookup_enum(MGF_T, param->mgf));
+			fprintf(spy_output, "pMechanism->pParameter->source=%lu\n", param->source);
+			spy_dump_string_out("pSourceData[ulSourceDalaLen]", 
+				param->pSourceData, param->ulSourceDataLen);
+		} else {
+			fprintf(spy_output, "Parameters block for %s is empty...\n",
+				lookup_enum(MEC_T, pMechanism->mechanism));
+		}
+		break;
+	default:
+		spy_dump_string_in("pParameter[ulParameterLen]", pMechanism->pParameter, pMechanism->ulParameterLen);
+		break;
+	}
 	spy_dump_ulong_in("hKey", hKey);
 	rv = po->C_DecryptInit(hSession, pMechanism, hKey);
 	return retne(rv);
@@ -969,6 +1008,27 @@ C_SignInit(CK_SESSION_HANDLE hSession, CK_MECHANISM_PTR pMechanism, CK_OBJECT_HA
 	enter("C_SignInit");
 	spy_dump_ulong_in("hSession", hSession);
 	fprintf(spy_output, "pMechanism->type=%s\n", lookup_enum(MEC_T, pMechanism->mechanism));
+	switch (pMechanism->mechanism) {
+	case CKM_RSA_PKCS_PSS:
+	case CKM_SHA1_RSA_PKCS_PSS:
+	case CKM_SHA256_RSA_PKCS_PSS:
+	case CKM_SHA384_RSA_PKCS_PSS:
+	case CKM_SHA512_RSA_PKCS_PSS:
+		if (pMechanism->pParameter != NULL) {
+			CK_RSA_PKCS_PSS_PARAMS *param =
+				(CK_RSA_PKCS_PSS_PARAMS *) pMechanism->pParameter;
+			fprintf(spy_output, "pMechanism->pParameter->hashAlg=%s\n",
+				lookup_enum(MEC_T, param->hashAlg));
+			fprintf(spy_output, "pMechanism->pParameter->mgf=%s\n",
+				lookup_enum(MGF_T, param->mgf));
+			fprintf(spy_output, "pMechanism->pParameter->sLen=%lu\n",
+				param->sLen);
+		} else {
+			fprintf(spy_output, "Parameters block for %s is empty...\n",
+				lookup_enum(MEC_T, pMechanism->mechanism));
+		}
+		break;
+	}
 	spy_dump_ulong_in("hKey", hKey);
 	rv = po->C_SignInit(hSession, pMechanism, hKey);
 	return retne(rv);
@@ -1053,6 +1113,27 @@ C_VerifyInit(CK_SESSION_HANDLE hSession, CK_MECHANISM_PTR pMechanism, CK_OBJECT_
 	enter("C_VerifyInit");
 	spy_dump_ulong_in("hSession", hSession);
 	fprintf(spy_output, "pMechanism->type=%s\n", lookup_enum(MEC_T, pMechanism->mechanism));
+	switch (pMechanism->mechanism) {
+	case CKM_RSA_PKCS_PSS:
+	case CKM_SHA1_RSA_PKCS_PSS:
+	case CKM_SHA256_RSA_PKCS_PSS:
+	case CKM_SHA384_RSA_PKCS_PSS:
+	case CKM_SHA512_RSA_PKCS_PSS:
+		if (pMechanism->pParameter != NULL) {
+			CK_RSA_PKCS_PSS_PARAMS *param =
+				(CK_RSA_PKCS_PSS_PARAMS *) pMechanism->pParameter;
+			fprintf(spy_output, "pMechanism->pParameter->hashAlg=%s\n",
+				lookup_enum(MEC_T, param->hashAlg));
+			fprintf(spy_output, "pMechanism->pParameter->mgf=%s\n",
+				lookup_enum(MGF_T, param->mgf));
+			fprintf(spy_output, "pMechanism->pParameter->sLen=%lu\n",
+				param->sLen);
+		} else {
+			fprintf(spy_output, "Parameters block for %s is empty...\n",
+				lookup_enum(MEC_T, pMechanism->mechanism));
+		}
+		break;
+	}
 	spy_dump_ulong_in("hKey", hKey);
 	rv = po->C_VerifyInit(hSession, pMechanism, hKey);
 	return retne(rv);
@@ -1279,7 +1360,52 @@ C_DeriveKey(CK_SESSION_HANDLE hSession, CK_MECHANISM_PTR pMechanism, CK_OBJECT_H
 
 	enter("C_DeriveKey");
 	spy_dump_ulong_in("hSession", hSession);
-	fprintf(spy_output, "pMechanism->type=%s\n", lookup_enum(MEC_T, pMechanism->mechanism));
+	fprintf(spy_output, "[in] pMechanism->type=%s\n",
+		lookup_enum(MEC_T, pMechanism->mechanism));
+	switch (pMechanism->mechanism) {
+	case CKM_ECDH1_DERIVE:
+	case CKM_ECDH1_COFACTOR_DERIVE:
+		if (pMechanism->pParameter == NULL) {
+			fprintf(spy_output, "[in] pMechanism->pParameter = NULL\n");
+			break;
+		}
+		CK_ECDH1_DERIVE_PARAMS *param =
+			(CK_ECDH1_DERIVE_PARAMS *) pMechanism->pParameter;
+		fprintf(spy_output, "[in] pMechanism->pParameter = {\n\tkdf=%s\n",
+			lookup_enum(CKD_T, param->kdf));
+		fprintf(spy_output, "\tpSharedData[ulSharedDataLen] = ");
+		print_generic(spy_output, 0, param->pSharedData,
+			param->ulSharedDataLen, NULL);
+		fprintf(spy_output, "\tpPublicData[ulPublicDataLen] = ");
+		print_generic(spy_output, 0, param->pPublicData,
+			param->ulPublicDataLen, NULL);
+		fprintf(spy_output, "}\n");
+		break;
+	case CKM_ECMQV_DERIVE:
+		if (pMechanism->pParameter == NULL) {
+			fprintf(spy_output, "[in] pMechanism->pParameter = NULL\n");
+			break;
+		}
+		CK_ECMQV_DERIVE_PARAMS *param2 =
+			(CK_ECMQV_DERIVE_PARAMS *) pMechanism->pParameter;
+		fprintf(spy_output, "[in] pMechanism->pParameter = {\n\tkdf=%s\n",
+			lookup_enum(CKD_T, param2->kdf));
+		fprintf(spy_output, "\tpSharedData[ulSharedDataLen] =");
+		print_generic(spy_output, 0, param2->pSharedData,
+			param2->ulSharedDataLen, NULL);
+		fprintf(spy_output, "\tpPublicData[ulPublicDataLen] = ");
+		print_generic(spy_output, 0, param2->pPublicData,
+			param2->ulPublicDataLen, NULL);
+		fprintf(spy_output, "\tulPrivateDataLen = %lu",
+			param2->ulPrivateDataLen);
+		fprintf(spy_output, "\thPrivateData = %lu", param2->hPrivateData);
+		fprintf(spy_output, "\tpPublicData2[ulPublicDataLen2] = ");
+		print_generic(spy_output, 0, param2->pPublicData2,
+			param2->ulPublicDataLen2, NULL);
+		fprintf(spy_output, "\tpublicKey = %lu", param2->publicKey);
+		fprintf(spy_output, "}\n");
+		break;
+	}
 	spy_dump_ulong_in("hBaseKey", hBaseKey);
 	spy_attribute_list_in("pTemplate", pTemplate, ulAttributeCount);
 	rv = po->C_DeriveKey(hSession, pMechanism, hBaseKey, pTemplate, ulAttributeCount, phKey);
@@ -1343,6 +1469,9 @@ C_WaitForSlotEvent(CK_FLAGS flags, CK_SLOT_ID_PTR pSlot, CK_VOID_PTR pRserved)
 
 	enter("C_WaitForSlotEvent");
 	spy_dump_ulong_in("flags", flags);
+	if (pSlot != NULL) {
+		spy_dump_ulong_in("pSlot", *pSlot);
+	}
 	rv = po->C_WaitForSlotEvent(flags, pSlot, pRserved);
 	return retne(rv);
 }
