@@ -36,11 +36,25 @@
 #undef KEEP_AC_NONE_FOR_INIT_APPLET
 
 #define MYEID_MAX_PINS   14
-#define MYEID_MAX_RSA_KEY_LEN 2048
+#define MYEID_MAX_RSA_KEY_LEN 4096
 
 unsigned char MYEID_DEFAULT_PUBKEY[] = {0x01, 0x00, 0x01};
 #define MYEID_DEFAULT_PUBKEY_LEN       sizeof(MYEID_DEFAULT_PUBKEY)
 
+#define MYEID_PROP_INFO_2_EXCTRACTABLE			0x08;
+#define MYEID_PROP_INFO_1_TRUSTED				0x04;
+#define MYEID_PROP_INFO_1_WRAP_WITH_TRUSTED		0x08;
+#define MYEID_PROP_INFO_2_SESSION_OBJECT		0x01;
+
+static const struct sc_object_id id_aes128_ecb = { { 2, 16, 840, 1, 101, 3, 4, 1, 1, -1 } };
+static const struct sc_object_id id_aes128_cbc = { { 2, 16, 840, 1, 101, 3, 4, 1, 2, -1 } };
+static const struct sc_object_id id_aes256_ecb = { { 2, 16, 840, 1, 101, 3, 4, 1, 41, -1 } };
+static const struct sc_object_id id_aes256_cbc = { { 2, 16, 840, 1, 101, 3, 4, 1, 42, -1 } };
+
+static void
+_add_supported_algo(struct sc_profile *profile, struct sc_pkcs15_card *p15card, struct sc_pkcs15_object *object,
+		    unsigned operations, unsigned mechanism, const struct sc_object_id *oid);
+
 /* For Myeid, all objects are files that can be deleted in any order */
 static int
 myeid_delete_object(struct sc_profile *profile, struct sc_pkcs15_card *p15card,
@@ -173,17 +187,22 @@ myeid_init_card(sc_profile_t *profile,
 		sc_pkcs15_card_t *p15card) {
 	struct sc_path path;
 	struct sc_file *file = NULL;
-        u8 rbuf[256];
+	u8 rbuf[256];
 	int r;
 
 	LOG_FUNC_CALLED(p15card->card->ctx);
 
 	p15card->tokeninfo->flags = SC_PKCS15_TOKEN_PRN_GENERATION | SC_PKCS15_TOKEN_EID_COMPLIANT;
 
-        r = sc_card_ctl(p15card->card, SC_CARDCTL_GET_SERIALNR, &rbuf);
+	_add_supported_algo(profile, p15card, NULL, SC_PKCS15_ALGO_OP_DECIPHER|SC_PKCS15_ALGO_OP_ENCIPHER, CKM_AES_ECB, &id_aes128_ecb);
+	_add_supported_algo(profile, p15card, NULL, SC_PKCS15_ALGO_OP_DECIPHER|SC_PKCS15_ALGO_OP_ENCIPHER, CKM_AES_CBC, &id_aes128_cbc);
+	_add_supported_algo(profile, p15card, NULL, SC_PKCS15_ALGO_OP_DECIPHER|SC_PKCS15_ALGO_OP_ENCIPHER, CKM_AES_ECB, &id_aes256_ecb);
+	_add_supported_algo(profile, p15card, NULL, SC_PKCS15_ALGO_OP_DECIPHER|SC_PKCS15_ALGO_OP_ENCIPHER, CKM_AES_CBC, &id_aes256_cbc);
+
+	r = sc_card_ctl(p15card->card, SC_CARDCTL_GET_SERIALNR, &rbuf);
 	LOG_TEST_RET(p15card->card->ctx, r,  "Get applet info failed");
 
-        sc_format_path("3F00", &path);
+	sc_format_path("3F00", &path);
 	r = sc_select_file(p15card->card, &path, &file);
 
 	sc_file_free(file);
@@ -232,6 +251,7 @@ myeid_create_dir(sc_profile_t *profile, sc_pkcs15_card_t *p15card, sc_file_t *df
 		for (ii = 0; create_dfs[ii]; ii++) {
 			sc_log(ctx, "Create '%s'", create_dfs[ii]);
 
+			file = NULL;
 			r = sc_profile_get_file(profile, create_dfs[ii], &file);
 			sc_file_free(file);
 			if (r) {
@@ -427,13 +447,23 @@ myeid_encode_public_key(sc_profile_t *profile, sc_card_t *card,
 }
 
 /*
+ * Add AlgorithmInfo of a supported algorithm to supportedAlgorithms field in tokenInfo. If object != NULL,
+ * add reference to the algorithmInfo to the passed object.
  */
 static void
 _add_supported_algo(struct sc_profile *profile, struct sc_pkcs15_card *p15card, struct sc_pkcs15_object *object,
 		    unsigned operations, unsigned mechanism, const struct sc_object_id *oid)
 {
 	struct sc_supported_algo_info *algo;
-	algo = sc_pkcs15_get_supported_algo(p15card, operations, mechanism);
+	struct sc_context *ctx = p15card->card->ctx;
+	if (oid == NULL) {
+		sc_log(ctx, "Failed to add algorithms refs - invalid arguments.");
+		return;
+	}
+	algo = sc_pkcs15_get_specific_supported_algo(p15card, operations, mechanism, oid);
+	int rv;
+
+	LOG_FUNC_CALLED(ctx);
 	if (!algo) {
 		unsigned ref = 1, ii;
 
@@ -451,16 +481,19 @@ _add_supported_algo(struct sc_profile *profile, struct sc_pkcs15_card *p15card,
 		}
 
 	}
-	sc_pkcs15_add_supported_algo_ref(object, algo);
+	if (object != NULL)
+		rv = sc_pkcs15_add_supported_algo_ref(object, algo);
+	else
+		rv = SC_SUCCESS;
+
+	if (rv != SC_SUCCESS) {
+		sc_log(ctx, "Failed to add algorithms refs");
+	}
 }
 
 static void
 myeid_fixup_supported_algos(struct sc_profile *profile, struct sc_pkcs15_card *p15card, struct sc_pkcs15_object *object)
 {
-	static const struct sc_object_id id_aes128_ecb = { { 2, 16, 840, 1, 101, 3, 4, 1, 1, -1 } };
-	static const struct sc_object_id id_aes128_cbc = { { 2, 16, 840, 1, 101, 3, 4, 1, 2, -1 } };
-	static const struct sc_object_id id_aes256_ecb = { { 2, 16, 840, 1, 101, 3, 4, 1, 41, -1 } };
-	static const struct sc_object_id id_aes256_cbc = { { 2, 16, 840, 1, 101, 3, 4, 1, 42, -1 } };
 	struct sc_context *ctx = p15card->card->ctx;
 	struct sc_pkcs15_skey_info *skey_info = (struct sc_pkcs15_skey_info *) object->data;
 
@@ -468,11 +501,11 @@ myeid_fixup_supported_algos(struct sc_profile *profile, struct sc_pkcs15_card *p
 	switch (object->type) {
 	case SC_PKCS15_TYPE_SKEY_GENERIC:
 		switch (skey_info->key_type | (skey_info->value_len << 16)) {
-		case CKM_AES_ECB | (128 << 16):
+		case CKK_AES | (128 << 16):
 			_add_supported_algo(profile, p15card, object, SC_PKCS15_ALGO_OP_DECIPHER|SC_PKCS15_ALGO_OP_ENCIPHER, CKM_AES_ECB, &id_aes128_ecb);
 			_add_supported_algo(profile, p15card, object, SC_PKCS15_ALGO_OP_DECIPHER|SC_PKCS15_ALGO_OP_ENCIPHER, CKM_AES_CBC, &id_aes128_cbc);
 			break;
-		case CKM_AES_ECB | (256 << 16):
+		case CKK_AES | (256 << 16):
 			_add_supported_algo(profile, p15card, object, SC_PKCS15_ALGO_OP_DECIPHER|SC_PKCS15_ALGO_OP_ENCIPHER, CKM_AES_ECB, &id_aes256_ecb);
 			_add_supported_algo(profile, p15card, object, SC_PKCS15_ALGO_OP_DECIPHER|SC_PKCS15_ALGO_OP_ENCIPHER, CKM_AES_CBC, &id_aes256_cbc);
 			break;
@@ -500,6 +533,8 @@ myeid_create_key(struct sc_profile *profile, struct sc_pkcs15_card *p15card,
 	struct sc_pkcs15_auth_info *pkcs15_auth_info = NULL;
 	unsigned char sec_attrs[] = {0xFF, 0xFF, 0xFF};
 	int r, ef_structure = 0, keybits = 0, pin_reference = -1;
+	unsigned char prop_info[] = {0x00, 0x00};
+	int extractable = FALSE;
 
 	LOG_FUNC_CALLED(card->ctx);
 
@@ -516,16 +551,24 @@ myeid_create_key(struct sc_profile *profile, struct sc_pkcs15_card *p15card,
 		case SC_PKCS15_TYPE_SKEY_3DES:
 			ef_structure = SC_CARDCTL_MYEID_KEY_DES;
 			keybits = skey_info->value_len;
+			if ((skey_info->access_flags & SC_PKCS15_PRKEY_ACCESS_EXTRACTABLE) == SC_PKCS15_PRKEY_ACCESS_EXTRACTABLE)
+				extractable = TRUE;
 			break;
 		case SC_PKCS15_TYPE_SKEY_GENERIC:
 			keybits = skey_info->value_len;
+			if ((skey_info->access_flags & SC_PKCS15_PRKEY_ACCESS_EXTRACTABLE) == SC_PKCS15_PRKEY_ACCESS_EXTRACTABLE)
+				extractable = TRUE;
 			switch (skey_info->key_type) {
-			case CKM_AES_ECB:
+			case CKK_AES:
 				ef_structure = SC_CARDCTL_MYEID_KEY_AES;
 				break;
-			case CKM_DES_ECB:
+			case CKK_DES:
 				ef_structure = SC_CARDCTL_MYEID_KEY_DES;
 				break;
+			default:
+				if (object->type == SC_PKCS15_TYPE_SKEY_GENERIC)
+					ef_structure = SC_CARDCTL_MYEID_KEY_GENERIC_SECRET;
+				break;
 			}
 			break;
 	}
@@ -597,6 +640,22 @@ myeid_create_key(struct sc_profile *profile, struct sc_pkcs15_card *p15card,
 		LOG_TEST_RET(ctx, SC_ERROR_INVALID_ARGUMENTS, "Invalid AuthID value for a private key.");
 	}
 
+	/* TODO: fill all proprietary attributes here based on the object */
+
+	if (object->user_consent != 0 && pin_reference >= 1)
+	    prop_info[0] |= (pin_reference << 4);
+
+	if (extractable)
+	    prop_info[1] |= MYEID_PROP_INFO_2_EXCTRACTABLE;
+
+	if (object->session_object != 0) /* Object will be removed during next reset. */
+		prop_info[1] |= MYEID_PROP_INFO_2_SESSION_OBJECT;
+
+	/* TODO: add other flags, like CKA_TRUSTED and CKA_WRAP_WITH_TRUSTED */
+
+	r = sc_file_set_prop_attr(file, prop_info, 2);
+	LOG_TEST_RET(ctx, r, "Cannot create MyEID key file");
+
 	/* Now create the key file */
 	r = sc_pkcs15init_create_file(profile, p15card, file);
 	sc_file_free(file);
@@ -742,7 +801,6 @@ myeid_generate_key(struct sc_profile *profile, struct sc_pkcs15_card *p15card,
 			break;
 		default:
 			LOG_TEST_RET(ctx, SC_ERROR_INVALID_ARGUMENTS, "Unsupported key type");
-			break;
 	}
 
 	sc_log(ctx, "Generate key with ID:%s and path:%s",
@@ -770,7 +828,7 @@ myeid_generate_key(struct sc_profile *profile, struct sc_pkcs15_card *p15card,
 	r = sc_card_ctl(card, SC_CARDCTL_MYEID_GENERATE_STORE_KEY, &args);
 	LOG_TEST_RET(ctx, r, "Card control 'MYEID_GENERATE_STORE_KEY' failed");
 
-	/* Keypair generation -> collect public key info */
+	/* Key pair generation -> collect public key info */
 	if (pubkey != NULL) {
 		struct sc_cardctl_myeid_data_obj data_obj;
 
@@ -821,6 +879,8 @@ myeid_generate_key(struct sc_profile *profile, struct sc_pkcs15_card *p15card,
 
 			dataptr = data_obj.Data;
 			r = sc_asn1_read_tag((const u8 **)&dataptr, data_obj.DataLen, &cla, &tag, &taglen);
+			if (dataptr == NULL)
+				r = SC_ERROR_ASN1_OBJECT_NOT_FOUND;
 			LOG_TEST_RET(ctx, r, "Invalid EC public key data. Cannot parse DER structure.");
 
 			if (taglen == 0)
