@@ -26,6 +26,7 @@
 
 #include "libopensc/opensc.h"
 #include "libopensc/cardctl.h"
+#include "libopensc/internal.h"
 #include "libopensc/log.h"
 #include "libopensc/cards.h"
 #include "libopensc/asn1.h"
@@ -104,7 +105,7 @@ static int openpgp_create_key(sc_profile_t *profile, sc_pkcs15_card_t *p15card,
 }
 
 /**
- * Stores an external (RSA) on the card.
+ * Stores an external key on the card.
  * @param  profile  profile information for this card
  * @param  card     sc_card_t object to use
  * @param  obj      sc_pkcs15_object_t object with pkcs15 information
@@ -116,44 +117,72 @@ static int openpgp_store_key(sc_profile_t *profile, sc_pkcs15_card_t *p15card,
 {
 	sc_card_t *card = p15card->card;
 	sc_pkcs15_prkey_info_t *kinfo = (sc_pkcs15_prkey_info_t *) obj->data;
-	struct sc_pkcs15_prkey_rsa *rsa = &(key->u.rsa);
 	sc_cardctl_openpgp_keystore_info_t key_info;
 	int r;
+	unsigned int i;
 
 	LOG_FUNC_CALLED(card->ctx);
 
-	if (obj->type != SC_PKCS15_TYPE_PRKEY_RSA) {
-		sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "only RSA is currently supported");
-		return SC_ERROR_NOT_SUPPORTED;
-	}
+	switch(obj->type)
+	{
+	case SC_PKCS15_TYPE_PRKEY_RSA:
+		memset(&key_info, 0, sizeof(sc_cardctl_openpgp_keystore_info_t));
+		key_info.algorithm = SC_OPENPGP_KEYALGO_RSA;
+		key_info.key_id = kinfo->id.value[0];
+		key_info.u.rsa.e = key->u.rsa.exponent.data;
+		key_info.u.rsa.e_len = key->u.rsa.exponent.len * 8; /* use bits instead of bytes */
+		key_info.u.rsa.p = key->u.rsa.p.data;
+		key_info.u.rsa.p_len = key->u.rsa.p.len;
+		key_info.u.rsa.q = key->u.rsa.q.data;
+		key_info.u.rsa.q_len = key->u.rsa.q.len;
+		key_info.u.rsa.n = key->u.rsa.modulus.data;
+		key_info.u.rsa.n_len = key->u.rsa.modulus.len * 8; /* use bits instead of bytes */
+		r = sc_card_ctl(card, SC_CARDCTL_OPENPGP_STORE_KEY, &key_info);
+		break;
+	case SC_PKCS15_TYPE_PRKEY_EC:
+		if (card->type < SC_CARD_TYPE_OPENPGP_V3) {
+			sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "only RSA is supported on this card");
+			LOG_FUNC_RETURN(card->ctx, SC_ERROR_NOT_SUPPORTED);
+		}
+		memset(&key_info, 0, sizeof(sc_cardctl_openpgp_keystore_info_t));
+		key_info.algorithm = (kinfo->id.value[0] == SC_OPENPGP_KEY_ENCR)
+				   ? SC_OPENPGP_KEYALGO_ECDH /* ECDH for slot 2 only */
+				   : SC_OPENPGP_KEYALGO_ECDSA; /* ECDSA for slot 1 and 3 */
+		key_info.key_id = kinfo->id.value[0];
+		key_info.u.ec.privateD = key->u.ec.privateD.data;
+		key_info.u.ec.privateD_len = key->u.ec.privateD.len;
+		key_info.u.ec.ecpointQ = key->u.ec.ecpointQ.value;
+		key_info.u.ec.ecpointQ_len = key->u.ec.ecpointQ.len;
+		/* extract oid the way we need to import it to OpenPGP Card */
+		if (key->u.ec.params.der.len > 2)
+			key_info.u.ec.oid_len = key->u.ec.params.der.value[1];
+		else
+			LOG_FUNC_RETURN(card->ctx, SC_ERROR_INVALID_ARGUMENTS);
 
-	memset(&key_info, 0, sizeof(sc_cardctl_openpgp_keystore_info_t));
-	key_info.keytype = kinfo->id.value[0];
-	key_info.e = rsa->exponent.data;
-	key_info.e_len = rsa->exponent.len;
-	key_info.p = rsa->p.data;
-	key_info.p_len = rsa->p.len;
-	key_info.q = rsa->q.data;
-	key_info.q_len = rsa->q.len;
-	key_info.n = rsa->modulus.data;
-	key_info.n_len = rsa->modulus.len;
-	r = sc_card_ctl(card, SC_CARDCTL_OPENPGP_STORE_KEY, &key_info);
+		for (i=0; (i < key_info.u.ec.oid_len) && (i+2 < key->u.ec.params.der.len); i++){
+			key_info.u.ec.oid.value[i] = key->u.ec.params.der.value[i+2];
+		}
+		key_info.u.ec.oid.value[key_info.u.ec.oid_len] = -1;
+		r = sc_card_ctl(card, SC_CARDCTL_OPENPGP_STORE_KEY, &key_info);
+		break;
+	default:
+		r = SC_ERROR_NOT_SUPPORTED;
+		sc_log(card->ctx, "%s: Key generation failed: Unknown/unsupported key type.", strerror(r));
+	}
 
 	LOG_FUNC_RETURN(card->ctx, r);
 }
 
 /**
- * Generates a new (RSA) key pair using an existing key file.
- * @param  profile  IN profile information for this card
+ * Generates a new RSA key pair on card.
  * @param  card     IN sc_card_t object to use
  * @param  obj      IN sc_pkcs15_object_t object with pkcs15 information
  * @param  pukkey   OUT the newly created public key
  * @return SC_SUCCESS on success and an error code otherwise
  **/
-static int openpgp_generate_key(sc_profile_t *profile, sc_pkcs15_card_t *p15card,
-	sc_pkcs15_object_t *obj, sc_pkcs15_pubkey_t *pubkey)
+static int openpgp_generate_key_rsa(sc_card_t *card, sc_pkcs15_object_t *obj,
+	sc_pkcs15_pubkey_t *pubkey)
 {
-	sc_card_t *card = p15card->card;
 	sc_context_t *ctx = card->ctx;
 	sc_cardctl_openpgp_keygen_info_t key_info;
 	sc_pkcs15_prkey_info_t *required = (sc_pkcs15_prkey_info_t *)obj->data;
@@ -169,53 +198,180 @@ static int openpgp_generate_key(sc_profile_t *profile, sc_pkcs15_card_t *p15card
 		/* Default key is authentication key. We choose this because the common use
 		 * is to generate from PKCS#11 (Firefox/Thunderbird) */
 		sc_log(ctx, "Authentication key is to be generated.");
-		key_info.keytype = 3;
+		key_info.key_id = 3;
 	}
-	if (!key_info.keytype && (kid->len > 1 || kid->value[0] > 3)) {
+	if (!key_info.key_id && (kid->len > 1 || kid->value[0] > 3)) {
 		sc_log(ctx, "Key ID must be 1, 2 or 3!");
 		LOG_FUNC_RETURN(ctx, SC_ERROR_INVALID_ARGUMENTS);
 	}
 
-	if (!key_info.keytype)
-		key_info.keytype = kid->value[0];
+	if (!key_info.key_id)
+		key_info.key_id = kid->value[0];
+
+	if (obj->type != SC_PKCS15_TYPE_PRKEY_RSA) {
+		sc_log(card->ctx,  "only RSA is currently supported");
+		return SC_ERROR_NOT_SUPPORTED;
+	}
+
+
+	key_info.algorithm = SC_OPENPGP_KEYALGO_RSA;
 
 	/* Prepare buffer */
-	key_info.modulus_len = required->modulus_length;
-	key_info.modulus = calloc(required->modulus_length >> 3, 1);
-	if (key_info.modulus == NULL)
+	key_info.u.rsa.modulus_len = required->modulus_length;
+	key_info.u.rsa.modulus = calloc(required->modulus_length >> 3, 1);
+	if (key_info.u.rsa.modulus == NULL)
 		LOG_FUNC_RETURN(ctx, SC_ERROR_NOT_ENOUGH_MEMORY);
 
 	/* The OpenPGP supports only 32-bit exponent. */
-	key_info.exponent_len = 32;
-	key_info.exponent = calloc(key_info.exponent_len>>3, 1); /* 1/8 */
-	if (key_info.exponent == NULL) {
-		free(key_info.modulus);
+	key_info.u.rsa.exponent_len = 32;
+	key_info.u.rsa.exponent = calloc(BYTES4BITS(key_info.u.rsa.exponent_len), 1);
+	if (key_info.u.rsa.exponent == NULL) {
+		free(key_info.u.rsa.modulus);
 		LOG_FUNC_RETURN(ctx, SC_ERROR_NOT_ENOUGH_MEMORY);
 	}
 
 	r = sc_card_ctl(card, SC_CARDCTL_OPENPGP_GENERATE_KEY, &key_info);
-	if (r < 0)
-		goto out;
+	LOG_TEST_GOTO_ERR(card->ctx, r, "on-card EC key generation failed");
 
+	pubkey->algorithm = SC_ALGORITHM_RSA;
 	sc_log(ctx, "Set output modulus info");
-	pubkey->u.rsa.modulus.len = key_info.modulus_len;
-	pubkey->u.rsa.modulus.data = calloc(key_info.modulus_len, 1);
+	pubkey->u.rsa.modulus.len = key_info.u.rsa.modulus_len;
+	pubkey->u.rsa.modulus.data = calloc(key_info.u.rsa.modulus_len, 1);
 	if (pubkey->u.rsa.modulus.data == NULL)
-		goto out;
-	memcpy(pubkey->u.rsa.modulus.data, key_info.modulus, key_info.modulus_len);
+		goto err;
+	memcpy(pubkey->u.rsa.modulus.data, key_info.u.rsa.modulus, key_info.u.rsa.modulus_len);
 
 	sc_log(ctx, "Set output exponent info");
-	pubkey->u.rsa.exponent.len = key_info.exponent_len;
-	pubkey->u.rsa.exponent.data = calloc(key_info.exponent_len>>3, 1); /* 1/8 */
+	pubkey->u.rsa.exponent.len = key_info.u.rsa.exponent_len;
+	pubkey->u.rsa.exponent.data = calloc(BYTES4BITS(key_info.u.rsa.exponent_len), 1);
 	if (pubkey->u.rsa.exponent.data == NULL)
-		goto out;
-	memcpy(pubkey->u.rsa.exponent.data, key_info.exponent, key_info.exponent_len>>3); /* 1/8 */
+		goto err;
+	memcpy(pubkey->u.rsa.exponent.data, key_info.u.rsa.exponent, BYTES4BITS(key_info.u.rsa.exponent_len));
+
+err:
+	free(key_info.u.rsa.modulus);
+	free(key_info.u.rsa.exponent);
+	LOG_FUNC_RETURN(ctx, r);
+}
+
+
+/**
+ * Generates a new ECC key pair on card.
+ * @param  card     IN sc_card_t object to use
+ * @param  obj      IN sc_pkcs15_object_t object with pkcs15 information
+ * @param  pukkey   OUT the newly created public key
+ * @return SC_SUCCESS on success and an error code otherwise
+ **/
+static int openpgp_generate_key_ec(sc_card_t *card, sc_pkcs15_object_t *obj,
+									sc_pkcs15_pubkey_t *pubkey)
+{
+	sc_context_t *ctx = card->ctx;
+	sc_cardctl_openpgp_keygen_info_t key_info;
+	sc_pkcs15_prkey_info_t *required = (sc_pkcs15_prkey_info_t *)obj->data;
+	sc_pkcs15_id_t *kid = &(required->id);
+	const struct sc_ec_parameters *info_ec =
+	    (struct sc_ec_parameters *) required->params.data;
+	unsigned int i;
+	int r;
+
+	LOG_FUNC_CALLED(ctx);
+	memset(&key_info, 0, sizeof(key_info));
+
+	sc_log(ctx, "Key ID to be generated: %s", sc_dump_hex(kid->value, kid->len));
+
+	/* Accept KeyID = 45, which is default value set by pkcs15init */
+	if (kid->len == 1 && kid->value[0] == 0x45) {
+		/* Default key is authentication key. We choose this because the common use
+		 * is to generate from PKCS#11 (Firefox/Thunderbird) */
+		sc_log(ctx, "Authentication key is to be generated.");
+		key_info.key_id = 3;
+	}
+	if (!key_info.key_id && (kid->len > 1 || kid->value[0] > 3)) {
+		sc_log(ctx, "Key ID must be 1, 2 or 3!");
+		LOG_FUNC_RETURN(ctx, SC_ERROR_INVALID_ARGUMENTS);
+	}
+
+	if (!key_info.key_id)
+		key_info.key_id = kid->value[0];
+
+
+	/* set algorithm id based on key reference */
+	key_info.algorithm = (key_info.key_id == SC_OPENPGP_KEY_ENCR)
+			   ? SC_OPENPGP_KEYALGO_ECDH /* ECDH for slot 2 only */
+			   : SC_OPENPGP_KEYALGO_ECDSA; /* ECDSA for slot 1 and 3 */
+
+	/* extract oid the way we need to import it to OpenPGP Card */
+	if (info_ec->der.len > 2)
+		key_info.u.ec.oid_len = info_ec->der.value[1];
+	else
+		LOG_FUNC_RETURN(ctx, SC_ERROR_INVALID_ARGUMENTS);
+
+	for (i=0; (i < key_info.u.ec.oid_len) && (i+2 < info_ec->der.len); i++){
+		key_info.u.ec.oid.value[i] = info_ec->der.value[i+2];
+	}
+	key_info.u.ec.oid.value[key_info.u.ec.oid_len] = -1;
+
+	/* Prepare buffer */
+	key_info.u.ec.ecpoint_len = required->field_length;
+	key_info.u.ec.ecpoint = malloc(key_info.u.ec.ecpoint_len);
+	if (key_info.u.ec.ecpoint == NULL)
+		LOG_FUNC_RETURN(ctx, SC_ERROR_NOT_ENOUGH_MEMORY);
+
+	/* generate key on card */
+	r = sc_card_ctl(card, SC_CARDCTL_OPENPGP_GENERATE_KEY, &key_info);
+	LOG_TEST_GOTO_ERR(card->ctx, r, "on-card EC key generation failed");
+
+	/* set pubkey according to response of card */
+	sc_log(ctx, "Set output ecpoint info");
+	pubkey->algorithm = SC_ALGORITHM_EC;
+	pubkey->u.ec.ecpointQ.len = key_info.u.ec.ecpoint_len;
+	pubkey->u.ec.ecpointQ.value = malloc(key_info.u.ec.ecpoint_len);
+	if (pubkey->u.ec.ecpointQ.value == NULL)
+		goto err;
+	memcpy(pubkey->u.ec.ecpointQ.value, key_info.u.ec.ecpoint, key_info.u.ec.ecpoint_len);
+
+err:
+	if (key_info.u.ec.ecpoint)
+		free(key_info.u.ec.ecpoint);
+
+	LOG_FUNC_RETURN(ctx, r);
+}
+
+
+/**
+ * Generates a new key pair using an existing key file.
+ * @param  profile  IN profile information for this card
+ * @param  card     IN sc_card_t object to use
+ * @param  obj      IN sc_pkcs15_object_t object with pkcs15 information
+ * @param  pukkey   OUT the newly created public key
+ * @return SC_SUCCESS on success and an error code otherwise
+ **/
+static int openpgp_generate_key(sc_profile_t *profile, sc_pkcs15_card_t *p15card,
+	sc_pkcs15_object_t *obj, sc_pkcs15_pubkey_t *pubkey)
+{
+	sc_card_t *card = p15card->card;
+	sc_context_t *ctx = card->ctx;
+	int r;
+
+	LOG_FUNC_CALLED(ctx);
+
+	switch(obj->type)
+	{
+	case SC_PKCS15_TYPE_PRKEY_RSA:
+		r = openpgp_generate_key_rsa(card, obj, pubkey);
+		break;
+	case SC_PKCS15_TYPE_PRKEY_EC:
+		if (card->type < SC_CARD_TYPE_OPENPGP_V3) {
+			sc_debug(card->ctx, SC_LOG_DEBUG_NORMAL, "only RSA is supported on this card");
+			return SC_ERROR_NOT_SUPPORTED;
+		}
+		r = openpgp_generate_key_ec(card, obj, pubkey);
+		break;
+	default:
+		r = SC_ERROR_NOT_SUPPORTED;
+		sc_log(card->ctx, "%s: Key generation failed: Unknown/unsupported key type.", strerror(r));
+	}
 
-out:
-	if (key_info.modulus)
-		free(key_info.modulus);
-	if (key_info.exponent)
-		free(key_info.exponent);
 	LOG_FUNC_RETURN(ctx, r);
 }
 
