@@ -437,6 +437,7 @@ void sc_notify_id(struct sc_context *ctx, struct sc_atr *atr,
 #elif defined(ENABLE_NOTIFY) && defined(ENABLE_GIO2)
 
 #include <gio/gio.h>
+#include "libopensc/log.h"
 
 static GApplication *application = NULL;
 
@@ -481,6 +482,14 @@ static void notify_gio(struct sc_context *ctx,
 			}
 		}
 
+		if (ctx) {
+			sc_log(ctx, "%s %s %s %s",
+					title ? title : "",
+					text ? text : "",
+					icon ? icon : "",
+					group ? group : "");
+		}
+
 		g_application_send_notification(application, group, notification);
 
 		if (gicon) {
