@@ -9,6 +9,10 @@ if [[ ! -f $P11LIB ]]; then
     echo "WARNINIG: The SoftHSM is not installed. Can not run this test"
     exit 77;
 fi
+
+# The Ubuntu has old softhsm version not supporting this feature
+grep "Ubuntu 18.04" /etc/issue && echo "WARNING: Not supported on Ubuntu 18.04" && exit 77
+
 card_setup
 
 echo "======================================================="
